/**
 * @file
 * 
 * @brief This file declares the AABB struct that models 
 *        an axis aligned bounding box.
 * 
 * @author Jan-Luca Hansel
 */

#ifndef AABB_H
#define AABB_H

#include "Polygon.h"

namespace Pathfinder {

	/**
	 * @brief This struct models an AABB.
	 * @details An AABB is defined to be the volume between the two corners
	 *          min and max.
	struct AABB {

	class AABB : public Geometry::BoundingBox {
	public:
	*/
	struct AABB {
		Geometry::Vec3 min; /**< The smallest corner of the AABB.*/
		Geometry::Vec3 max; /**< The biggest corner of the AABB.*/


		/**
		 * @brief A constructor that creates an object with dummy attributes.
		 * @details The two data members are initialized with two empty Vec3 
		 *          objects.
		 */
		AABB();

		
		/**
		 * @brief A constructor that creates an AABB out of two vectors.
		 * @details This constructor initializes min and max with the given
		 *          parameters of the same name.
		 *
		 * @param min The smallest corner of the AABB.
		 * @param max The biggest corner of the AABB.
		 */
		AABB(
			const Geometry::Vec3& min, 
			const Geometry::Vec3& max
		);

	
		/**
		 * @brief Check whether a polygon definetely lies outside of this AABB.
		 * @details This is the case if the vertices p1, p2 and p3 of p all 
		 *          share a component which is smaller/greater than the 
		 *          corresponding component of min/max.
		 * 
		 * @param p         The polygon that shall be compared.
		 * @param component The component which is to be examined. 
		 *                  x = 0, y = 1, z = 2.
		 *
		 * @return Return true if such a component exists, false otherwise.
		 */
		bool 
		triNotInAABB(
			const Polygon* p, 
			const short& component
		);


		/**
		 * @brief Check whether the polygon intersects with this AABB.
		 * @details If triNotInAABB returns true for any component, this 
		 *          function will be false.
		 *
		 * @param p The polygon which is to be checked for intersection.
		 *
		 * @return Return true if the polygon has the possibility of 
		 *         intersecting with the AABB. Return false if there is no 
		 *         chance of intersection.
		 */
		bool 
		triInAABB(
			const Polygon* p
		); 


		/**
		 * @brief Check whether the point lies inside this AABB.
		 *
		 * @param p The point that shall be checked.
		 *
		 * @return Return true if p lies inside the AABB, false otherwise.
		 */
		bool
		pointInAABB(
			Geometry::Vec3 p
		);

	};


	/**
	 * @brief Check if two AABBs intersect.
	 * @details The involved check is presented in the thesis. 
	 *
	 * @param a The first AABB.
	 * @param b The second AABB.
	 *
	 * @return Return true if the AABBs intersect, false otherwise.
	 */	
	bool 
	AABBintersection(
		const AABB& a, 
		const AABB& b
	);

}

#endif
