/** 
 * @file
 *
 * @brief This file implements all functionality declared in Analysis.h.
 *
 * @author Jan-Luca Hansel
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <sys/stat.h>
#include <sys/types.h>
#include <bits/stdc++.h>
#include <ctime>
#include <memory>
#include <chrono>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include <boost/graph/graph_utility.hpp>
#include <boost/graph/connected_components.hpp>


#include "Analysis.h"
#include "OBJ.h"
#include "Intersection.h"
#include "BuildGraph.h"
#include "BuildGroups.h"
#include "Merge.h"
#include "Output.h"

namespace Pathfinder {

	std::string createOutputDirs() {
		// Create a new folder that bears the current time as a name
		std::time_t time = std::time(0);
		std::stringstream currentStream;
		currentStream << "../Output/" << time;
		std::string topLevelDirStr = currentStream.str();
		char* topLevelDir = (char*) topLevelDirStr.c_str();
		mkdir(topLevelDir, 0777);

		if(objOutput) { 
			// Create a subfolder for the OBJ files created from meshes
			std::string pathOBJMeshesStr = topLevelDirStr + "/OBJMeshes";
			char* pathOBJMeshes = (char*) pathOBJMeshesStr.c_str();
			mkdir(pathOBJMeshes, 0777);

			// Create a subfolder for the OBJ files created from graphs
			std::string pathOBJGraphsStr = topLevelDirStr + "/OBJGraphs";
			char* pathOBJGraphs = (char*) pathOBJGraphsStr.c_str();
			mkdir(pathOBJGraphs, 0777);

			// Create a subfolder for the OBJ files created from the merged 
			// graphs
			std::string pathOBJMergedGraphsStr = topLevelDirStr + "/OBJMerged";
			char* pathOBJMergedGraphs = (char*) pathOBJMergedGraphsStr.c_str();
			mkdir(pathOBJMergedGraphs, 0777);
		}

		return topLevelDirStr;
	}

	//Check if graph is connected
	//@author Vineet
	bool isGraphConnected(const Graph& graph) {
		std::vector<int> component(num_vertices(graph));
		int num = boost::connected_components(graph, &component[0]);
		return num == 1;
	}

	std::vector<Graph*> sceneAnalysis(const MinSG::ListNode& root) {
		// Start the timer for this function
		spdlog::stopwatch sw;

		std::cout << yellow << "Start Analysis" << reset << std::endl;

		// Create the output directories
		std::string dir = createOutputDirs();

		// Create logger objects
		createLoggers(dir);

		// Get all GeometryNodes of the scene
		std::vector<MinSG::GeometryNode*> geometryNodes = 
			extractGeometryNodes(root);

		// Create OBJ files out of the meshes
		if(objOutput) { 
			for(auto n : geometryNodes) {
				std::stringstream pathStream;
				pathStream << dir << "/OBJMeshes/" << n << ".obj";
				std::string path = pathStream.str();
				writeOBJFromMesh(n, path);
			}
		}

		// Create MeshAccessor objects
		std::vector<MeshAccessor*> meshAccessors = 
			createMeshAccessors(geometryNodes);

		// Build the graph list
		std::vector<Graph*> graphList = 
			buildGraphList(geometryNodes, meshAccessors);

		// Check if each graph is connected
		for (const auto& graph : graphList) {
			if (!isGraphConnected(*graph)) {
				std::cerr << "Graph " << graph << " is not connected!" << std::endl;
			}
		}

		// Construct an octree
		Octree* octree = createOctree(graphList);
		//printOctree(octree, "");

		// Create OBJ files out of the graphs
		if(objOutput) {
			for(auto g : graphList) {
				std::stringstream pathStream;
				pathStream << dir << "/OBJGraphs/" << g << ".obj";
				std::string path = pathStream.str();
				writeOBJFromGraph(g, path);
			}
		}

		// Build the intersection groups
		std::vector<std::vector<Graph*>> groups = 
			buildGroups(geometryNodes, graphList);

		// Create OBJ files for every intersection
		if(objOutput) {
			// Look at each group
			for(size_t i = 0; i < groups.size(); ++i) {
				for(size_t j = 0; j < groups[i].size(); ++j) {
					Graph* tj = groups[i][j];
					for(size_t k = 0; k < groups[i].size(); ++k) {
						Graph* tk = groups[i][k];
						// Look at each pair of graphs in the group
						if(j >= k) {
							continue;
						} else {
							// Create a subfolder for the current Group
							std::string pathGroupStr = dir + "/Group" + std::to_string(i);
							char* pathGroup = (char*) pathGroupStr.c_str();
							mkdir(pathGroup, 0777);
							// Create a subfolder for the pointclouds
							std::string pathPointcloudsStr = dir + "/Group" + std::to_string(i) + "/pointclouds/";
							char* pathPointclouds = (char*) pathPointcloudsStr.c_str();
							mkdir(pathPointclouds, 0777);
							// Create a subfolder for the convex hulls
							std::string pathConvexHullStr = dir + "/Group" + std::to_string(i) + "/convexHulls/";
							char* pathConvexHull = (char*) pathConvexHullStr.c_str();
							mkdir(pathConvexHull, 0777);
							// Write the intersections to an OBJ file
							std::vector<std::tuple<Graph*, Graph*, std::vector<Geometry::Vec3>>> intersections = intersection(tj, tk);
							std::stringstream pathStream;
							pathStream << dir << "/Group" << i << "/";
							std::string path = pathStream.str();
							writeIntersectionOBJFiles(intersections, path); 
						}
					}
				}
			}
		}

		// Merge all graphs in each group
		std::vector<Graph*> mergedGraphs;
		for(size_t i = 0; i < groups.size(); ++i) {
			mergedGraphs.push_back(merge(groups[i], octree));
		}

		std::cout << green << "Merging finished\n" << reset;
				
		// Create OBJ files for every merged graph
		if(objOutput) {
			for(auto g : mergedGraphs) {
				std::stringstream pathStream;
				pathStream << dir << "/OBJMerged/" << g << ".obj";
				std::string path = pathStream.str();
				writeOBJFromGraph(g, path);
			}
		}

		// Free all memory allocated by octree
		octree->deleteOctree(true);
		delete octree;

/* TODO
		// Free all memory allocated by MeshAccessor objects
		for(size_t i = 0; i < graphList.size(); ++i) {
			std::get<1>(graphList[i]).deletePolygons();
		}
*/

		spdlog::get("time_log")->info("sceneAnalysis: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));

		return mergedGraphs;
	}

}
