/**
 * @file
 *
 * @brief This file implements all functionality declared in Astar.h.
 *
 * @author Jan-Luca Hansel
 */

#include <algorithm>
#include <stack>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include <boost/optional/optional_io.hpp>

#include "Astar.h"
#include "Vec.h"

namespace Pathfinder {

	boost::optional<Vertex> nextSuccessor(
		const Graph* graph, const Vertex& v, const Vertex& goal, 
		const short& agent, const std::vector<Vertex>& considered
	) {
		// Get all candidate successors
		std::vector<Vertex> candidates;
		auto incidentEIts = boost::out_edges(v, *graph);
		for(auto eit = incidentEIts.first; eit != incidentEIts.second; ++eit) {
			// Edge must be traversable for the agent
			if(agent >= (*graph)[*eit].drivability) {
				candidates.push_back(boost::target(*eit, *graph));
			}
		}

		// Do not return a vertex that is already on the stack
		for(auto v : considered) {
			candidates.erase(
				std::remove(candidates.begin(), candidates.end(), v), 
				candidates.end()
			);
		}

		if(candidates.size() != 0) {
			// Sucessors exist. Get the one with the smallest euclidean 
			// distance to the goal
			float minDistance = 
				distance3((*graph)[candidates[0]].pos, (*graph)[goal].pos);
			Vertex chosen = candidates[0];
			for(size_t i = 1; i < candidates.size(); ++i) {
				float current = 
					distance3((*graph)[candidates[i]].pos, (*graph)[goal].pos);
				if(current < minDistance) {
					minDistance = current;
					chosen = candidates[i];
				}	
			}
			return chosen;
		}
		return boost::none;
	}


	float cost(const Graph* graph, const Vertex& ni, const Vertex& nj) {
		// Get the edge between ni and nj
		auto e = boost::edge(ni, nj, *graph);

		// Check if the edge exists
		if(e.second) {
			return (*graph)[e.first].distance;
		}
		spdlog::get("warn_log")->info("cost returns std::numeric_limits<float>::max().");
		return std::numeric_limits<float>::max();
	}	


	std::vector<Vertex> astar(
		const Graph* graph, const Vertex& start, 
		const Vertex& goal, const short& agent
	) {
		// Start timer for this function
		spdlog::stopwatch sw;

		std::cout << yellow << "Start A*" << reset << '\n';

		std::stack<Vertex> s;
		std::vector<Vertex> considered;
		bool found = false;
	
		// g and h values for every vertex
		std::vector<float> 
			g(boost::num_vertices(*graph), std::numeric_limits<float>::max());
		std::vector<float> 
			h(boost::num_vertices(*graph), std::numeric_limits<float>::max());

		// Set g and h for the start node
		g[start] = 0;
		h[start] = distance3((*graph)[start].pos, (*graph)[goal].pos);

		// Initial threshold
		float B = h[start];

		// Try to find a path
		while(!found && B < std::numeric_limits<float>::max()) {
			s.push(start);
			considered = {start};
			float Bnew = std::numeric_limits<float>::max();
			while(!found && s.size() != 0) {
				// Check if all nodes lie on the stack
				if(s.size() == boost::num_vertices(*graph)) {
					return {};
				}
				Vertex ni = s.top();
				if(ni != goal) {
					boost::optional<Vertex> nj = 
						nextSuccessor(graph, ni, goal, agent, considered);
					if(nj != boost::none) {
						considered.push_back(*nj);
						g[*nj] = 
							std::min(g[*nj], g[ni] + cost(graph, ni, *nj));
						h[*nj] = 
							distance3((*graph)[*nj].pos, (*graph)[goal].pos);
						if((g[*nj] + h[*nj]) > B) {
							Bnew = std::min(Bnew, (g[*nj] + h[*nj]));
						} else {
							s.push(*nj);
						}
					} else {
						s.pop();
					}
				} else {
					std::cout << "Found a path!\n";
					found = true;
					std::vector<Vertex> p;
					while(s.size() != 0) {
						p.push_back(s.top());
						s.pop();
					}
					std::reverse(p.begin(), p.end());
					spdlog::get("time_log")->info("astar: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
					return p;
				}
			}
			B = Bnew;
			spdlog::get("node_log")->info("[{}] {}", B, considered.size());
		}
		spdlog::get("time_log")->info("astar: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
		return {};
	}

}
