/**
 * @file
 *
 * @brief This file declares all functionality that is related to the A*
 *        pathfinding algorithm.
 *
 * @author Jan-Luca Hansel
 */

#ifndef ASTAR_H
#define ASTAR_H

#include <boost/optional.hpp>

#include "Using.h"

namespace Pathfinder {

	/**
	 * @brief This function yields the next vertex that shall be processed by
	 *        A*.
	 * @details The function looks at all incident edges of v. Only those edges
	 *          that are passable by agent are considered. The function then
	 *          selects a candidate vertex that was not considered before.
	 *
	 * @param graph      A pointer to the graph that is worked on.
	 * @param v          The vertex whose successors shall be found.
	 * @param goal       The goal node of the search.
	 * @param agent      Only edges that are passable by agent shall be 
	 *                   considered.
	 * @param considered A vector containing all already considered vertices.
	 *
	 * @return Returns a possible successor of v that is reachable by agent.
	 *         If such a successor does not exist, boost::none is returned.
	 */
	boost::optional<Vertex> 
	nextSuccessor(
		const Graph* graph, 
		const Vertex& v, 
		const Vertex& goal, 
		const short& agent, 
		const std::vector<Vertex>& considered
	); 


	/**
	 * @brief Return the edge weight of the edge {ni, nj} in graph.
	 *
	 * @param graph A pointer to the graph that is worked on.
	 * @param ni    The first endpoint of the edge.
	 * @param nj    The second endpoint of the edge.
	 *
	 * @return Returns the weight associated to the edge {ni, nj}.
	 */
	float
	cost(
		const Graph* graph,
		const Vertex& ni,
		const Vertex& nj
	);


	/**
	 * @brief Find a path from start to goal using A*. This path must be 
	 *        passable by agent.
	 * @details This is the IDA* algorithm presented in 4.4.
	 *
	 * @param graph A pointer to the graph that is worked on.
	 * @param start The vertex description of the starting point.
	 * @param goal  The vertex description of the goal point.
	 * @param agent The agent that shall traverse the graph.
	 *
	 * @return Returns a vector contaning all vertices on the shortest path 
	 *         from start to goal that is passable by agent.
	 */
	std::vector<Vertex>
	astar(
		const Graph* graph,
		const Vertex& start,
		const Vertex& goal,
		const short& agent
	);

}

#endif
