/**
 * @file
 *
 * @brief This file implements all functionality declared in AstarAlternativeHeuristic.h.
 *
 * @author Vineet
 */

#include <algorithm>
#include <stack>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include <boost/optional/optional_io.hpp>

#include "AstarAlternativeHeuristic.h"
#include <queue>

#include "Vec.h"

namespace Pathfinder {
    typedef float (*distanceFunction)(const Geometry::_Vec3<float> &v1, const Geometry::_Vec3<float> &v2);

    distanceFunction distance = //distanceDiagonal;       // Diagonal distance
            distance3; //Euclidean distance
     //distanceChebyshev;      //Chebyshev distance
    //distanceManhatten;      //Manhattan distance


    boost::optional<Vertex> nextSuccessor(
        const Graph *graph, const Vertex &v, const Vertex &goal,
        const short &agent, const std::vector<Vertex> &considered
    ) {
        // Get all candidate successors
        std::vector<Vertex> candidates;
        auto incidentEIts = boost::out_edges(v, *graph);
        for (auto eit = incidentEIts.first; eit != incidentEIts.second; ++eit) {
            // Edge must be traversable for the agent
            if (agent >= (*graph)[*eit].drivability) {
                // std::cout << "\tdrivability: "<<(*graph)[*eit].drivability <<"\tagent" <<agent;
                candidates.push_back(boost::target(*eit, *graph));
            }
        }

        // Do not return a vertex that is already on the stack
        for (auto v: considered) {
            candidates.erase(
                std::remove(candidates.begin(), candidates.end(), v),
                candidates.end()
            );
        }

        if (candidates.size() != 0) {
            // Successors exist. Get the one with the smallest Euclidean
            // distance to the goal
            float minDistance =
                    //distance3((*graph)[candidates[0]].pos, (*graph)[goal].pos);
                    distance((*graph)[candidates[0]].pos, (*graph)[goal].pos);
            Vertex chosen = candidates[0];
            for (size_t i = 1; i < candidates.size(); ++i) {
                float current =

                        //distance3((*graph)[candidates[i]].pos, (*graph)[goal].pos);
                        distance((*graph)[candidates[i]].pos, (*graph)[goal].pos);
                if (current < minDistance) {
                    minDistance = current;
                    chosen = candidates[i];
                }
            }
            //		std::cout << " v val "<<v<<'\n';
            //		std::cout << " considered val "<<&considered<<'\n';


            //		for (Vertex i: considered)
            //			std::cout << i << ' ';

            //		std::cout <<'\n';
            //		std::cout << " chosen "<<chosen<<'\n';
            return chosen;
        }

        return boost::none;
    }

    float est = 0;

    float cost(const Graph *graph, const Vertex &ni, const Vertex &nj) {
        // Get the edge between ni and nj
        auto e = boost::edge(ni, nj, *graph);

        // @author: Vineet
        //when determining next vertex, A star uses edge weight as one of the decision factor
        // adding edge weight along with distance between two edges

        // Check if the edge exists
        if (e.second) {
            if ((*graph)[e.first].edge_weight == 0) {
                //             std::cout << "\nedge distance: " << (*graph)[e.first].distance;

                return (*graph)[e.first].distance;
            } else {
                est = (*graph)[e.first].distance - 0.125 * (*graph)[e.first].distance + 0.125 * (*graph)[e.first].
                      distance * (*graph)[e.first].edge_weight;
                //                std::cout << "\nedge distance: " << (*graph)[e.first].distance;
                // std::cout << "\t edge weight: " << (*graph)[e.first].edge_weight;
                 //               std::cout << "\t total estimate: " << est;
                //                std::cout << "\t type: " << (*graph)[e.first].edge_type;
                return est;
            }
        }
        spdlog::get("warn_log")->info("cost returns std::numeric_limits<float>::max().");
        return std::numeric_limits<float>::max();
    }

    std::vector<Vertex> AstarAlternativeHeuristicORI( //Original Astar code
        const Graph *graph, const Vertex &start,
        const Vertex &goal, const short &agent
    ) {
        // Start timer for this function
        spdlog::stopwatch sw;

        std::cout << yellow << "Start A*" << reset << '\n';

        std::stack<Vertex> s;
        std::vector<Vertex> considered;
        bool found = false;

        // g and h values for every vertex
        std::vector<float>
                g(boost::num_vertices(*graph), std::numeric_limits<float>::max());
        std::vector<float>
                h(boost::num_vertices(*graph), std::numeric_limits<float>::max());

        // Set g and h for the start node
        g[start] = 0;
        h[start] = distance3((*graph)[start].pos, (*graph)[goal].pos);

        // Initial threshold
        float B = h[start];
        int PathCounter = 0;
        int NodesCounter = 0;

        // Try to find a path
        while (!found && B < std::numeric_limits<float>::max()) {
            s.push(start);
            considered = {start};
            float Bnew = std::numeric_limits<float>::max();
            while (!found && s.size() != 0) {
                // Check if all nodes lie on the stack
                if (s.size() == boost::num_vertices(*graph)) {
                    return {};
                }
                Vertex ni = s.top();
                if (ni != goal) {
                    boost::optional<Vertex> nj =
                            nextSuccessor(graph, ni, goal, agent, considered);
                    if (nj != boost::none) {
                        considered.push_back(*nj);
                        g[*nj] =
                                std::min(g[*nj], g[ni] + cost(graph, ni, *nj));
                        h[*nj] =
                                distance3((*graph)[*nj].pos, (*graph)[goal].pos);
                        if ((g[*nj] + h[*nj]) > B) {
                            Bnew = std::min(Bnew, (g[*nj] + h[*nj]));
                        } else {
                            s.push(*nj);
                        }
                    } else {
                        s.pop();
                    }
                } else {
                    std::cout << "Found a path!\n";
                    found = true;
                    std::vector<Vertex> p;
                    while (s.size() != 0) {
                        p.push_back(s.top());
                        s.pop();
                    }
                    std::reverse(p.begin(), p.end());
                    spdlog::get("time_log")->info("astar: {}",
                                                  std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
                    spdlog::get("time_log")->info("PathCounter: {}", PathCounter);
                    spdlog::get("time_log")->info("NodesCounter: {}", NodesCounter);
                    std::cout << " PathCounter: " << PathCounter << " NodesCounter: " << NodesCounter;

                    return p;
                }
                ++PathCounter;
            }
            B = Bnew;
            ++NodesCounter;
            spdlog::get("node_log")->info("[{}] {}", B, considered.size());
        }
        spdlog::get("time_log")->info("astar: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
        return {};
    }

    void printPath(const PathTaken *goal) {
        const PathTaken *current = goal;
        while (current) {
            current->print();
            current = current->predecessor;
        }
    }


    // Function to determine if three points form a zig-zag pattern
    bool isZigZag(const Vertex& prev, const Vertex& curr, const Vertex& next, const Graph* graph) {
        const auto& posPrev = (*graph)[prev].pos;
        const auto& posCurr = (*graph)[curr].pos;
        const auto& posNext = (*graph)[next].pos;

        auto vector1 = posCurr - posPrev;
        auto vector2 = posNext - posCurr;

        // Use the cross product in 2D (x and z) to check for collinearity
        float crossProduct = vector1.x() * vector2.z() - vector1.z() * vector2.x();

        // Calculate the dot product
        float dotProduct = vector1.x() * vector2.x() + vector1.z() * vector2.z();

        // Calculate magnitudes of vectors
        float magnitude1 = std::sqrt(vector1.x() * vector1.x() + vector1.z() * vector1.z());
        float magnitude2 = std::sqrt(vector2.x() * vector2.x() + vector2.z() * vector2.z());

        // Calculate the cosine of the angle
        float cosAngle = dotProduct / (magnitude1 * magnitude2);

        // Calculate the angle in degrees
        float angle = std::acos(cosAngle) * 180.0 / M_PI;

        // If the angle is between 60 and 90 degrees, do not smooth
        bool isBetweenSixtyAndNinety = angle > 60.0 && angle < 90.0;

        // If it's a zig-zag i.e. not collinear and not between 60 and 90 degrees
        return crossProduct != 0 && !isBetweenSixtyAndNinety;

    }
    // Function to print Edge Type
    void printEdgeType(const Vertex &v1, const Vertex &v2, const Graph *graph) {
        // Retrieve the edge between the vertices
        auto edge = boost::edge(v1, v2, *graph);

        if (edge.second) { // If the edge exists
            char edgeType = (*graph)[edge.first].edge_type;  // Retrieve edge type
            // std::cout << "Edge from Vertex " << v1 << " to Vertex " << v2
                      // << " is of type: " << edgeType << '\n';
            std::cout << " edge type: " << edgeType << '\n';
        } else {
            std::cout << "No edge exists between Vertex " << v1 << " and Vertex " << v2 << '\n';
        }
    }

    // Function to smooth the path
    std::vector<Vertex> smoothPath(const std::vector<Vertex> &path, const Graph *graph) {
        if (path.size() < 3) return path; // No smoothing needed for paths if there are less than 3 points

        std::vector<Vertex> smoothedPath;
        smoothedPath.push_back(path[0]); // Start with the first point

        for (size_t i = 1; i < path.size() - 1; ++i) {
            const Vertex &prev = path[i - 1];
            const Vertex &curr = path[i];
            const Vertex &next = path[i + 1];

            // Print edge type between previous and current vertices
            printEdgeType(prev, curr, graph);

            // Check if curr is a zig-zag point
            if (isZigZag(prev, curr, next, graph)) {
            // If it's a zig-zag, skip the current point
                continue;
            }

            // Otherwise, add the current point to the smoothed path
            smoothedPath.push_back(curr);
        }

        smoothedPath.push_back(path.back()); // End with the last point
        return smoothedPath;
    }


    std::vector<Vertex> AstarAlternativeHeuristic( //Improved code
        const Graph *graph, const Vertex &start,
        const Vertex &goal, const short &agent
    ) {
        // Start timer for this function
        spdlog::stopwatch sw;

        std::cout << yellow << "Start A*" << reset << '\n';

        std::stack<PathTaken *> s;
        std::vector<Vertex> considered;
        bool found = false;

        // g and h values for every vertex
        std::vector<float>
                g(boost::num_vertices(*graph), std::numeric_limits<float>::max());
        std::vector<float>
                h(boost::num_vertices(*graph), std::numeric_limits<float>::max());


        // Set g and h for the start node
        g[start] = 0;
        h[start] = distance((*graph)[start].pos, (*graph)[goal].pos);

        // Initial threshold
        float B = h[start];
        int PathCounter = 0;
        int NodesCounter = 0;

        // Try to find a path
        while (!found && B < std::numeric_limits<float>::max()) {
            s.push(new PathTaken(start));
            // std::cout << "Pushed start vertex: " << start << '\n';
            considered = {start};
            float Bnew = std::numeric_limits<float>::max();
            while (!found && s.size() != 0) {
                // Check if all nodes lie on the stack
                if (s.size() == boost::num_vertices(*graph)) {
                    return {};
                }
                PathTaken *ni = s.top();
                if (ni->vertex != goal) {
                    boost::optional<Vertex> nj =
                            nextSuccessor(graph, ni->vertex, goal, agent, ni->expanded);
                    if (nj != boost::none) {
                        // considered.push_back(*nj);

                        float valg;
                        valg = g[ni->vertex] + cost(graph, ni->vertex, *nj);
                        ni->expanded.push_back(*nj);

                        //	std::cout << " nj val "<< *nj <<'\n';
                        //	std::cout << "  g[*nj] val "<<  g[*nj] <<'\n';
                        //	std::cout << "  valg "<<  valg <<'\n';

                        if (g[*nj] >= valg) {
                            g[*nj] = std::min(g[*nj], valg);
                            h[*nj] =  // Different Heuristics can change the path
                                    distance((*graph)[*nj].pos, (*graph)[goal].pos);
                            float f_value = g[*nj] + h[*nj];
                            // reduce the number of calculations of f_value @vineet
                            // if ((g[*nj] + h[*nj]) > B) {
                            //     Bnew = std::min(Bnew, (g[*nj] + h[*nj]));
                            //     //    std::cout << " Bnew "<<Bnew <<'\n';
                            // }
                            if (f_value > B) {
                                Bnew = std::min(Bnew, (f_value));
                                //    std::cout << " Bnew "<<Bnew <<'\n';
                            }

                            else {
                                s.push(new PathTaken(*nj, ni));
                                 //  std::cout << "Pushed vertex: " << *nj << '\n';
                            }
                        }
                    } else {
                       // PathTaken *popped = s.top(); // @vineet Original code
                        //  std::cout << "Popped vertex: " << ni->vertex << '\n';
                        delete s.top();
                        s.pop();
                        //delete popped;  // @vineet Original code
                    }
                } else {
                    std::cout << "\nFound a path!\n";
                    found = true;
                    std::vector<Vertex> p;
                    PathTaken *current = ni;
                    std::cout << "Complete path:\n";
                  //  printPath(ni);
                    while (current) {
                        //   std::cout << "Popped vertex (path): " << current->vertex << '\n';
                        p.push_back(current->vertex);
                        PathTaken *prev = current; // @vineet renamed popped to prev
                        current = current->predecessor;
                        delete prev;
                    }

                    std::reverse(p.begin(), p.end());

                    // Smooth the path
                    std::vector<Vertex> smoothedPath = smoothPath(p, graph);



                    // Log the number of nodes in the path
                    spdlog::get("time_log")->info("PathLength: {}", p.size());

                    spdlog::get("time_log")->info("AstarAlternativeHeuristic: {}",
                                                  std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
                    spdlog::get("time_log")->info("PathCounter: {}", PathCounter);
                    spdlog::get("time_log")->info("searchRange: {}", NodesCounter);
                    std::cout << " PathCounter: " << PathCounter << " searchRange: " << NodesCounter<< '\n';
                    // Print the number of nodes in the path
                    std::cout << "Number of nodes in the path: " << p.size() << '\n';
                    return smoothedPath;
                    //return p;
                }
                ++PathCounter;
            }
            B = Bnew;
            ++NodesCounter;
            spdlog::get("node_log")->info("[{}] {}", B, considered.size());
        }

        spdlog::get("time_log")->info("AstarAlternativeHeuristic: {}",
                                      std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));


        return {};
    }

}
