/**
 * @file
 *
 * @brief This file declares all functionality that is related to the Manhatten as heuristic for
 *        pathfinding algorithm.
 *
 * @author Vineet
 */

#ifndef PATHFINDER_ASTARALTERNATIVEHEURISTIC_H
#define PATHFINDER_ASTARALTERNATIVEHEURISTIC_H

#include <boost/optional.hpp>
#include <boost/optional/optional_io.hpp>
#include <boost/graph/adjacency_list.hpp>

#include "Using.h"

namespace Pathfinder {
	/**
	 * @brief This class represents a node in the path taken by A* algorithm.
	 * @details It contains the vertex descriptor and a pointer to its predecessor in the path.
	 */
	class PathTaken {
	public:
		Vertex vertex;  ///< The vertex descriptor
		PathTaken* predecessor;  ///< Pointer to the predecessor vertex in the path
		std::vector <Vertex> expanded;

		/**
		 * @brief Constructor for PathTaken
		 * @param v The vertex descriptor
		 * @param p Pointer to the predecessor PathTaken object (default is nullptr)
		 */
		PathTaken(Vertex v, PathTaken* p = nullptr) : vertex(v), predecessor(p),expanded(std::vector<Vertex>()) {

			//expanded = std::vector<Vertex>();
		}


		/**
		 * @brief Print the contents of this PathTaken object.
		 */
		void print() const {
			std::cout << "Vertex: " << vertex;
			if (predecessor) {
				std::cout << ", Predecessor: " << predecessor->vertex;
			} else {
				std::cout << ", Predecessor: nullptr";
			}
			std::cout << std::endl;
		}
	};
	/**
	 * @brief Print the path from start to goal using PathTaken objects.
	 * @param goal The PathTaken object representing the goal node.
	 */
	void printPath(const PathTaken* goal);

	bool isZigZag(const Vertex& prev, const Vertex& curr, const Vertex& next, const Graph* graph);

	/**bool isObstacle(const Graph& graph, const Vertex& vertex);

	bool hasLineOfSight(const Graph* graph, const Vertex& start, const Vertex& end);
	*/
	std::vector<Vertex> smoothenPath(const Graph* graph, const std::vector<PathTaken*>& path);

	/**
	 * @brief This function yields the next vertex that shall be processed by
	 *        AstarAlternativeHeuristic.
	 * @details The function looks at all incident edges of v. Only those edges
	 *          that are passable by agent are considered. The function then
	 *          selects a candidate vertex that was not considered before.
	 *
	 * @param graph      A pointer to the graph that is worked on.
	 * @param v          The vertex whose successors shall be found.
	 * @param goal       The goal node of the search.
	 * @param agent      Only edges that are passable by agent shall be 
	 *                   considered.
	 * @param considered A vector containing all already considered vertices.
	 *
	 * @return Returns a possible successor of v that is reachable by agent.
	 *         If such a successor does not exist, boost::none is returned.
	 */
	boost::optional<Vertex> 
	nextSuccessor(
		const Graph* graph, 
		const Vertex& v, 
		const Vertex& goal, 
		const short& agent, 
		const std::vector<Vertex>& considered
	); 


	/**
	 * @brief Return the edge weight of the edge {ni, nj} in graph.
	 *
	 * @param graph A pointer to the graph that is worked on.
	 * @param ni    The first endpoint of the edge.
	 * @param nj    The second endpoint of the edge.
	 *
	 * @return Returns the weight associated to the edge {ni, nj}.
	 */
	float
	cost(
		const Graph* graph,
		const Vertex& ni,
		const Vertex& nj
	);


	/**
	 * @brief Find a path from start to goal using A*. This path must be 
	 *        passable by agent.
	 * @details This is the IDA* algorithm presented in 4.4.
	 *
	 * @param graph A pointer to the graph that is worked on.
	 * @param start The vertex description of the starting point.
	 * @param goal  The vertex description of the goal point.
	 * @param agent The agent that shall traverse the graph.
	 *
	 * @return Returns a vector contaning all vertices on the shortest path 
	 *         from start to goal that is passable by agent.
	 */
	std::vector<Vertex>
	AstarAlternativeHeuristic(
		const Graph* graph,
		const Vertex& start,
		const Vertex& goal,
		const short& agent
	);

}

#endif
