/** 
 * @file
 *
 * @brief This file implements all functionality declared in BuildGraph.h.
 *
 * @author Jan-Luca Hansel
 */

#include <iostream>
#include <limits>
#include <cmath>
#include <stack>
#include <map>
#include <memory>
#include <chrono>
#include <thread>
#include <future>
#include <vector>
#include <mutex>
#include <execution>
#include <algorithm>
#include <omp.h>


#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include "Vec.h"
#include "BuildGraph.h"

namespace Pathfinder {

	std::vector<MinSG::GeometryNode*> 
	extractGeometryNodes(const MinSG::ListNode& root) {
		std::cout << yellow << "Extract GeometryNodes" << reset << std::endl;

		std::vector<MinSG::GeometryNode*> geometryNodes; 
		std::stack<std::tuple<const MinSG::ListNode&, size_t>> s;
		s.push(std::tuple<const MinSG::ListNode&, size_t>(root, 0));
		size_t currentChildNumber = 0;
		while(!s.empty()) {
			std::tuple<const MinSG::ListNode&, size_t> t = s.top();
			currentChildNumber = std::get<1>(t);
			s.pop();
			if(MinSG::GeometryNode* geoNodePtr = 
			   dynamic_cast<MinSG::GeometryNode*>(
					std::get<0>(t).getChild(currentChildNumber))
			) {
				// Cast was successful -> GeometryNode was found
				geometryNodes.push_back(geoNodePtr);
			} else {
				// Cast failed -> Another ListNode
				MinSG::ListNode& next = dynamic_cast<MinSG::ListNode&>(
					*std::get<0>(t).getChild(currentChildNumber)
				);
				s.push({next, 0});
			}
			if(currentChildNumber != std::get<0>(t).countChildren()-1) {
				// There are children left -> Put on stack w/ reference to next 
				// child
				s.push({std::get<0>(t), ++currentChildNumber});
			}
		}

		return geometryNodes;
	}


	std::vector<MeshAccessor*> createMeshAccessors(
		const std::vector<MinSG::GeometryNode*>& nodes
	) {
		std::cout << yellow << "Create MeshAccessors" << reset << std::endl;
		std::vector<MeshAccessor*> meshAccessors;
		for(size_t i = 0; i < nodes.size(); ++i) {
			meshAccessors.push_back(new MeshAccessor(nodes[i]));
		}
		return meshAccessors;
	}


	short determineDrivability(
		const Geometry::Vec3& v, const Geometry::Vec3& w
	) {
		float hypotenuse = distance3(v, w);
		//float deltaY = std::abs(v[y] - w[y]);
        float deltaY = std::abs(v[y] - w[y]);
		float alphaDeg = asin(deltaY / hypotenuse) * (180 / M_PI);


        float direction = v[y] - w[y];
		//std::cout <<v<<'\n';
		//std::cout << yellow << "Determine drivability:\n" << reset;
		//std::cout << magenta << "\t(" << v << ")\n";
		//std::cout << "\t(" << w << ")\n" << reset;
		//std::cout << "\tHypotenuse: " << hypotenuse << '\n';
		//std::cout << "\tDelta y: " << deltaY << '\n';
		//std::cout << "\tAngle: " << alphaDeg << '\n';
        if (direction<=0) {
          //  std::cout <<"\tUphill";
            if (alphaDeg <= 20) {
           //     std::cout << "\t-> Person 20\n";
                return person;
            } else if (alphaDeg <= 22.5) {
            //    std::cout << "\t-> Bicycle 22.5\n";
                return bicycle;
            } else if (alphaDeg <= 25) {
            //    std::cout << "\t-> Car 25\n";
                return car;
            } else if (alphaDeg <= 27.5) {
             //   std::cout << "\t-> Truck 27.5\n";
                return truck;
            } else {
            //    std::cout << "\t-> Noone\n";
                return none;
            }
        }
        else {
          //  std::cout <<"\tDownhill";
            if (alphaDeg <= 21) {
              //  std::cout << "\t-> Person 21\n";
                return person;
            } else if (alphaDeg <= 28) {
              //  std::cout << "\t-> Bicycle 28\n";
                return bicycle;
            } else if (alphaDeg <= 35) {
              //  std::cout << "\t-> Car 35\n";
                return car;
            } else if (alphaDeg <= 45) {
              //  std::cout << "\t-> Truck 45\n";
                return truck;
            } else {
             //   std::cout << "\t-> None downhill\n";
                return none;
            }
        }

	}

    const double uphill_factor=1;       //different costs for uphill and downhill movements
    const double downhill_factor=2;
    float determineEdgeWeight(const Geometry::Vec3& v1, const Geometry::Vec3& v2) {

       // float deltaY = v1[y] - v2[y];       // Y axis is the ascent or descent as per PADrend

//        std::cout <<(v1[y] - v2[y])<<"\n";
//        return (v1[y] - v2[y]);

        float hypotenuse = distance3(v1, v2);
        //float deltaY = std::abs(v[y] - w[y]);
        float deltaY = std::abs(v1[y] - v2[y]);
        float alphaDeg = asin(deltaY / hypotenuse) * (180 / M_PI);
        float direction = v1[y] - v2[y];
        float result=alphaDeg/45;
        if (result>=1)
        {
            result=1;
        }
        if (direction<=0)  //+ve uphill
        {
            return result;
        }
        else                //-ve downhill
            return -result;
//        if (v2[y] > v1[y]) {
//
//           // std::cout << "uphill: "<<(v2[y] - v1[y]) * uphill_factor;
//            return (v2[y] - v1[y]) * uphill_factor;
//        } else {
//
//          //  std::cout << "downhill: "<<(v1[y] - v2[y]) * downhill_factor;
//            return (v1[y] - v2[y]) * downhill_factor;
//
//        }

    }

 //     Determine uphill or downhill
    char determineEdgeType(const Geometry::Vec3& v1, const Geometry::Vec3& v2) {
        if (v2[y] > v1[y]) {
            return 'u';
        } else {
            return 'd';
        }
    }


    void processNodeOLD(size_t index, MeshAccessor* meshAccessor, std::vector<Graph*>& graphList, std::mutex& graphListMutex)
    {
        std::cout << index << " ";

        Graph* currentGraph = new Graph();
        MeshAccessor* currentMeshAccessor = meshAccessor;
        std::vector<Vertex> vertices;

        // Add all vertices to currentGraph and store them in vertices
        for (uint32_t vertexId = 0; vertexId < currentMeshAccessor->getVertexCount(); ++vertexId) {
            Vertex v = boost::add_vertex(
                    GraphVertex{
                            currentMeshAccessor->getVertex(vertexId),
                            currentMeshAccessor
                    },
                    *currentGraph
            );
            vertices.push_back(v);
        }
        //std::cout << '\n' << "Total number of vertices: " << vertices.size()<< std::endl;;

        // Add all edges
        for (uint32_t vertexId = 0; vertexId < vertices.size(); ++vertexId) {
            for (uint32_t neighbourId : currentMeshAccessor->getAdjacentVertexIds(vertexId)) {
                boost::add_edge(
                        vertices[vertexId],
                        vertices[neighbourId],
                        GraphEdge{
                                distance3(
                                        (*currentGraph)[vertices[vertexId]].pos,
                                        (*currentGraph)[vertices[neighbourId]].pos
                                ),
                                determineDrivability(
                                        (*currentGraph)[vertexId].pos,
                                        (*currentGraph)[neighbourId].pos
                                ),
                                determineEdgeWeight(  //@author: Vineet
                                        (*currentGraph)[vertexId].pos,
                                        (*currentGraph)[neighbourId].pos
                                )
                                ,
                                determineEdgeType(
                                        (*currentGraph)[vertexId].pos,
                                        (*currentGraph)[neighbourId].pos
                                        )

                        },
                        *currentGraph
                );
            }
        }

        // Use a mutex to safely add to graphList
        {
            std::lock_guard<std::mutex> lock(graphListMutex);
            graphList.push_back(currentGraph);
        }
    }

    void processNode(size_t index, MeshAccessor* meshAccessor, std::vector<Graph*>& graphList, std::mutex& graphListMutex)
    {
       // std::cout << index << " ";

        Graph* currentGraph = new Graph();
       // MeshAccessor* currentMeshAccessor = meshAccessor;
        std::vector<Vertex> vertices;

        // Add all vertices to currentGraph and store them in vertices
        for (uint32_t vertexId = 0; vertexId < meshAccessor->getVertexCount(); ++vertexId) {
            Vertex v = boost::add_vertex(
                    GraphVertex{
                            meshAccessor->getVertex(vertexId),
                    	meshAccessor
                    },
                    *currentGraph
            );
            vertices.push_back(v);
        }
        //std::cout << '\n' << "Total number of vertices: " << vertices.size()<< std::endl;;

        // Add all edges
        for (uint32_t vertexId = 0; vertexId < vertices.size(); ++vertexId) {
            for (uint32_t neighbourId : meshAccessor->getAdjacentVertexIds(vertexId)) {
                boost::add_edge(
                        vertices[vertexId],
                        vertices[neighbourId],
                        GraphEdge{
                                distance3(
                                        (*currentGraph)[vertices[vertexId]].pos,
                                        (*currentGraph)[vertices[neighbourId]].pos
                                ),
                                determineDrivability(
                                        (*currentGraph)[vertexId].pos,
                                        (*currentGraph)[neighbourId].pos
                                ),
                                determineEdgeWeight(  //@author: Vineet
                                        (*currentGraph)[vertexId].pos,
                                        (*currentGraph)[neighbourId].pos
                                )
                                ,
                                determineEdgeType(
                                        (*currentGraph)[vertexId].pos,
                                        (*currentGraph)[neighbourId].pos
                                        )

                        },
                        *currentGraph
                );
            }
        }

        // Use a mutex to safely add to graphList
        {
            std::lock_guard<std::mutex> lock(graphListMutex);
            graphList.push_back(currentGraph);
        }
    }



    std::vector<Graph*> buildGraphListThreads(  //Threads
        const std::vector<MinSG::GeometryNode*>& nodes,
        const std::vector<MeshAccessor*>& meshAccessors)
    {
        // Start the timer for buildGraphList function
        spdlog::stopwatch sw;

        std::cout << yellow << "Construct individual graphs: " << reset;

        std::vector<Graph*> graphList;
        std::vector<std::thread> threads;
        std::mutex graphListMutex;

        // Create one graph per GeometryNode in separate threads
        for (size_t i = 0; i < nodes.size(); ++i) {
            threads.push_back(std::thread(processNode, i, meshAccessors[i], std::ref(graphList), std::ref(graphListMutex)));
        }

        // Join all threads
        for (auto& thread : threads) {
            thread.join();
        }

       // std::cout << '\n';
       // std::cout << std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()).count() << " microseconds\n" << std::endl;

        spdlog::get("time_log")->info("buildGraphList: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
        return graphList;
    }

    std::vector<Graph*> buildGraphList( //Original
            const std::vector<MinSG::GeometryNode*>& nodes,
            const std::vector<MeshAccessor*>& meshAccessors
    ) {
        // Start the timer for this function
        spdlog::stopwatch sw;

        std::cout << yellow << "Construct individual graphs: " << reset;


        std::vector<Graph*> graphList;
        // Create one graph per GeometryNode
        for(size_t i = 0; i < nodes.size(); ++i) {
            std::cout << i << " ";

            Graph* currentGraph = new Graph();
            MeshAccessor* currentMeshAccessor = meshAccessors[i];
            std::vector<Vertex> vertices;


            // Add all vertices to currentGraph and store them in vertices
            for(uint32_t vertexId = 0;
                vertexId < currentMeshAccessor->getVertexCount(); ++vertexId
                    ) {
                Vertex v =
                        boost::add_vertex(
                                GraphVertex{
                                        currentMeshAccessor->getVertex(vertexId),
                                        currentMeshAccessor
                                },
                                *currentGraph
                        );
                vertices.push_back(v);
            }
            std::cout << '\n'<<"Total number of vertices: "<<vertices.size();

            // Add all edges
            for(uint32_t vertexId = 0; vertexId < vertices.size(); ++vertexId) {
                for(uint32_t neighbourId :
                        currentMeshAccessor->getAdjacentVertexIds(vertexId)
                        ) {
                    boost::add_edge(
                            vertices[vertexId],
                            vertices[neighbourId],
                            GraphEdge{
                                    distance3(
                                            (*currentGraph)[vertices[vertexId]].pos,
                                            (*currentGraph)[vertices[neighbourId]].pos
                                    ),
                                    determineDrivability(
                                            (*currentGraph)[vertexId].pos,
                                            (*currentGraph)[neighbourId].pos
                                    ),
                                    determineEdgeWeight(  //@author: Vineet
                                            (*currentGraph)[vertexId].pos,
                                            (*currentGraph)[neighbourId].pos
                                    )
                                ,
                                determineEdgeType(
                                        (*currentGraph)[vertexId].pos,
                                        (*currentGraph)[neighbourId].pos
                                        )

                            },
                            *currentGraph
                    );
                }
            }

            graphList.push_back(currentGraph);
        }

        std::cout << '\n';

        spdlog::get("time_log")->info("buildGraphList: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
        return graphList;
    }

std::vector<Graph*> buildGraphListPRAGMA(   //PRAGMA
    const std::vector<MinSG::GeometryNode*>& nodes,
    const std::vector<MeshAccessor*>& meshAccessors
) {
    // Start the timer for this function
    spdlog::stopwatch sw;

    std::cout << yellow << "Construct individual graphs: " << reset;

    std::vector<Graph*> graphList;
    graphList.reserve(nodes.size()); // Pre-allocate memory for graphList

    #pragma omp parallel for
    for(size_t i = 0; i < nodes.size(); ++i) {
        std::cout << i << " ";

        Graph* currentGraph = new Graph();
        MeshAccessor* currentMeshAccessor = meshAccessors[i];

        std::vector<Vertex> vertices;
        uint32_t vertexCount = currentMeshAccessor->getVertexCount();
        vertices.reserve(vertexCount); // Pre-allocate memory for vertices

        // Add all vertices to currentGraph and store them in vertices
        for(uint32_t vertexId = 0; vertexId < vertexCount; ++vertexId) {
            Vertex v = boost::add_vertex(
                GraphVertex{ currentMeshAccessor->getVertex(vertexId), currentMeshAccessor },
                *currentGraph
            );
            vertices.push_back(v);
        }

        // Add all edges
        for(uint32_t vertexId = 0; vertexId < vertexCount; ++vertexId) {
            const auto& neighbors = currentMeshAccessor->getAdjacentVertexIds(vertexId);
            for(uint32_t neighbourId : neighbors) {
                boost::add_edge(
                    vertices[vertexId],
                    vertices[neighbourId],
                    GraphEdge{
                        distance3(
                            (*currentGraph)[vertices[vertexId]].pos,
                            (*currentGraph)[vertices[neighbourId]].pos
                        ),
                        determineDrivability(
                            (*currentGraph)[vertexId].pos,
                            (*currentGraph)[neighbourId].pos
                            ),
                        determineEdgeWeight(  //@author: Vineet
                            (*currentGraph)[vertexId].pos,
                            (*currentGraph)[neighbourId].pos
                            ),
                        determineEdgeType(
                            (*currentGraph)[vertexId].pos,
                            (*currentGraph)[neighbourId].pos)
                    },
                    *currentGraph
                );
            }
        }

        #pragma omp critical
        graphList.push_back(currentGraph);
    }

    std::cout << '\n';
    //std::cout << std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()).count() << " microseconds\n" << std::endl;

    spdlog::get("time_log")->info("buildGraphList: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
    return graphList;
}



    //https://iq.opengenus.org/multithreaded-for-loop-in-cpp/
    // Function to build the graph list using multithreading
    std::vector<Graph*> buildGraphList4threads( //4 threads
        const std::vector<MinSG::GeometryNode*>& nodes,
    const std::vector<MeshAccessor*>& meshAccessors) // 4 threads
    {
        // Start the timer for buildGraphList function
        spdlog::stopwatch sw;

        std::cout << "Construct individual graphs: ";

        const int num_threads = 4;  // Number of threads
        std::vector<Graph*> graphList;
        std::mutex graphListMutex;
        std::vector<std::thread> threads;

        // Calculate the chunk size to divide work across threads
        int chunk_size = nodes.size() / num_threads;

        // Loop to create threads, each handling a chunk of the nodes
        for (int i = 0; i < num_threads; ++i) {
            int start = i * chunk_size;
            int end = (i + 1) * chunk_size;

            threads.emplace_back([&nodes, &meshAccessors, start, end, &graphList, &graphListMutex]()
                                 {
                                     // Process each node in the chunk
                                     for (int j = start; j < end; ++j) {
                                         processNode(j, meshAccessors[j], graphList, graphListMutex);
                                     }
                                 });
        }

        // Handle any remaining nodes if nodes.size() is not divisible by num_threads
        for (size_t j = num_threads * chunk_size; j < nodes.size(); ++j) {
            processNode(j, meshAccessors[j], graphList, graphListMutex);
        }


        // Wait for all threads to complete
        for (auto& thread : threads) {
            thread.join();
        }

        // Log the time elapsed
        spdlog::get("time_log")->info("buildGraphList: {} microseconds", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));

        return graphList;
    }


   std::vector<Graph*> buildGraphListParallel(
    const std::vector<MinSG::GeometryNode*>& nodes,
    const std::vector<MeshAccessor*>& meshAccessors
) {
    // Start the timer for this function
    spdlog::stopwatch sw;

    std::cout << yellow << "Construct individual graphs: " << reset;

    std::vector<Graph*> graphList(nodes.size(), nullptr);

   // #pragma omp parallel for
    for(size_t i = 0; i < nodes.size(); ++i) {
        std::cout << i << " ";

        Graph* currentGraph = new Graph();
        MeshAccessor* currentMeshAccessor = meshAccessors[i];
        std::vector<Vertex> vertices;

        // Add all vertices to currentGraph and store them in vertices
        for(uint32_t vertexId = 0; vertexId < currentMeshAccessor->getVertexCount(); ++vertexId) {
            Vertex v = boost::add_vertex(
                GraphVertex{
                    currentMeshAccessor->getVertex(vertexId),
                    currentMeshAccessor
                },
                *currentGraph
            );
            vertices.push_back(v);
        }

        // Add all edges
        for(uint32_t vertexId = 0; vertexId < vertices.size(); ++vertexId) {
            for(uint32_t neighbourId : currentMeshAccessor->getAdjacentVertexIds(vertexId)) {
                boost::add_edge(
                    vertices[vertexId],
                    vertices[neighbourId],
                    GraphEdge{
                        distance3(
                            (*currentGraph)[vertices[vertexId]].pos,
                            (*currentGraph)[vertices[neighbourId]].pos
                        ),
                        determineDrivability(
                            (*currentGraph)[vertexId].pos,
                            (*currentGraph)[neighbourId].pos
                        ),
                        determineEdgeWeight(
                            (*currentGraph)[vertexId].pos,
                            (*currentGraph)[neighbourId].pos
                        ),
                        determineEdgeType(
                            (*currentGraph)[vertexId].pos,
                            (*currentGraph)[neighbourId].pos
                        )
                    },
                    *currentGraph
                );
            }
        }

        graphList[i] = currentGraph;
    }

    std::cout << '\n';

    spdlog::get("time_log")->info("buildGraphList: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
    return graphList;
}

}
