/** 
 * @file
 *
 * @brief This file declares all functionality related to building graphs out
 *        of meshes.
 *
 * @author Jan-Luca Hansel
 */

#ifndef BUILDGRAPH_H
#define BUILDGRAPH_H

#include <MinSG/Core/Nodes/ListNode.h>

#include <boost/graph/adjacency_list.hpp>

#include "Using.h"
#include "MeshAccessor.h"

namespace Pathfinder {

	/**
	 * @brief Build a list of all GeometryNodes in the scene graph.
	 * @details Given the root node of a scene graph, traverse the scene
	 *          graph and store all GeometryNodes that have been found 
	 *          in a vector.
	 *
	 * @param root The root node of the scene graph.
	 *
	 * @return Return a vector containing pointers to all GeometryNodes of 
	 *         the scene graph.
	 */
	std::vector<MinSG::GeometryNode*> 
	extractGeometryNodes(
		const MinSG::ListNode& root
	);


	/**
	 * @brief Create MeshAccessor objects for the given GeometryNodes.
	 * @details Go through the list of GeometryNodes and create one MeshAccessor
	 *          object per node.
	 *
	 * @param nodes A vector of pointers to GeometryNodes.
	 *
	 * @return Return a vector of pointers to MeshAccessor objects.
	 */
	std::vector<MeshAccessor*> 
	createMeshAccessors(
		const std::vector<MinSG::GeometryNode*>& nodes
	);


	/**
	 * @brief Determine which agent is able to traverse the direct connection
	 *        between the two points.
	 * @details Determine the slope of the direct connecton between the two 
	 *          points. Which agent corresponds to which slope is 
	 *          mentioned in the thesis.
	 *
	 * @param v The first point.
	 * @param w The second point.
	 *
	 * @return Returns a number that corresponds to the agents that can 
	 *         traverse an edge.
	 */
	short
	determineDrivability(
		const Geometry::Vec3& v,
		const Geometry::Vec3& w
	);

    /**
     * @author  Vineet
     * @brief   Determine edge costs so that agent will be able to
     *          choose uphill or downhill movement along with drivability
	 * @details Determine the edge costs based on difference in Z values
	 *
	 * @param v The first point.
	 * @param w The second point.
	 *
	 * @return Returns a number that corresponds to the agents that can
	 *         traverse an edge.
	 */
    float
    determineEdgeWeight(
            const Geometry::Vec3& v,
            const Geometry::Vec3& w
    );

/**
     * @author  Vineet
     * @brief   Determine edge type
	 * @details Determine the type of edge
	 *
	 * @param v The first point.
	 * @param w The second point.
	 *
	 * @return Returns a character u for uphill and d for downhill
	 */
    char
    determineEdgeType(
            const Geometry::Vec3& v,
            const Geometry::Vec3& w
    );


	/**
	 * @brief Build a graph for every mesh of a scene graph.
	 * @details Given a list of GeometryNodes, get the meshes stored in them
	 *          and convert these meshes into graphs. This is done by taking
	 *          the vertices of the mesh and translating them into the vertices
	 *          of a graph. The edges of the graph correspond to the connections 
	 *          of the mesh's vertices that form the polygons.
	 *
	 * @param nodes         A vector of pointers to the GeometryNodes of 
	 *                      the scene graph.
	 * @param meshAccessors A vector of pointers to the MeshAccessors created 
	 *                      for the GeometryNodes of nodes.
	 *
	 * @return Return a vector of pointers to graphs.
	 */
	std::vector<Graph*> 
	buildGraphList(
		const std::vector<MinSG::GeometryNode*>& nodes, 
		const std::vector<MeshAccessor*>& meshAccessors
	);

}

#endif
