/** 
 * @file
 *
 * @brief This file implements all functionality declared in BuildGraph.h.
 *
 * @author Jan-Luca Hansel
 */

#include <iostream>
#include <limits>
#include <cmath>
#include <stack>
#include <map>
#include <memory>
#include <chrono>
#include <thread>
#include <vector>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include "Vec.h"
#include "BuildGraph.h"

namespace Pathfinder {

	std::vector<MinSG::GeometryNode*> 
	extractGeometryNodes(const MinSG::ListNode& root) {
		std::cout << yellow << "Extract GeometryNodes" << reset << std::endl;

		std::vector<MinSG::GeometryNode*> geometryNodes; 
		std::stack<std::tuple<const MinSG::ListNode&, size_t>> s;
		s.push(std::tuple<const MinSG::ListNode&, size_t>(root, 0));
		size_t currentChildNumber = 0;
		while(!s.empty()) {
			std::tuple<const MinSG::ListNode&, size_t> t = s.top();
			currentChildNumber = std::get<1>(t);
			s.pop();
			if(MinSG::GeometryNode* geoNodePtr = 
			   dynamic_cast<MinSG::GeometryNode*>(
					std::get<0>(t).getChild(currentChildNumber))
			) {
				// Cast was successful -> GeometryNode was found
				geometryNodes.push_back(geoNodePtr);
			} else {
				// Cast failed -> Another ListNode
				MinSG::ListNode& next = dynamic_cast<MinSG::ListNode&>(
					*std::get<0>(t).getChild(currentChildNumber)
				);
				s.push({next, 0});
			}
			if(currentChildNumber != std::get<0>(t).countChildren()-1) {
				// There are children left -> Put on stack w/ reference to next 
				// child
				s.push({std::get<0>(t), ++currentChildNumber});
			}
		}

		return geometryNodes;
	}


	std::vector<MeshAccessor*> createMeshAccessors(
		const std::vector<MinSG::GeometryNode*>& nodes
	) {
		std::cout << yellow << "Create MeshAccessors" << reset << std::endl;
		std::vector<MeshAccessor*> meshAccessors;
		for(size_t i = 0; i < nodes.size(); ++i) {
			meshAccessors.push_back(new MeshAccessor(nodes[i]));
		}
		return meshAccessors;
	}


	short determineDrivability(
		const Geometry::Vec3& v, const Geometry::Vec3& w
	) {
		float hypotenuse = distance3(v, w);
		//float deltaY = std::abs(v[y] - w[y]);
        float deltaY = std::abs(v[y] - w[y]);
		float alphaDeg = asin(deltaY / hypotenuse) * (180 / M_PI);


        float direction = v[y] - w[y];
		//std::cout <<v<<'\n';
		//std::cout << yellow << "Determine drivability:\n" << reset;
		//std::cout << magenta << "\t(" << v << ")\n";
		//std::cout << "\t(" << w << ")\n" << reset;
		//std::cout << "\tHypotenuse: " << hypotenuse << '\n';
		//std::cout << "\tDelta y: " << deltaY << '\n';
		//std::cout << "\tAngle: " << alphaDeg << '\n';
        if (direction<=0) {
            if (alphaDeg <= 20) {
                //std::cout << "\t-> Person\n";
                return person;
            } else if (alphaDeg <= 22.5) {
                //std::cout << "\t-> Bicycle\n";
                return bicycle;
            } else if (alphaDeg <= 25) {
                //std::cout << "\t-> Car\n";
                return car;
            } else if (alphaDeg <= 27.5) {
                //std::cout << "\t-> Truck\n";
                return truck;
            } else {
                //std::cout << "\t-> Noone\n";
                return none;
            }
        }
        else { //to-do
            if (alphaDeg <= 20) {
                //std::cout << "\t-> Person\n";
                return person;
            } else if (alphaDeg <= 22.5) {
                //std::cout << "\t-> Bicycle\n";
                return bicycle;
            } else if (alphaDeg <= 25) {
                //std::cout << "\t-> Car\n";
                return car;
            } else if (alphaDeg <= 27.5) {
                //std::cout << "\t-> Truck\n";
                return truck;
            } else {
                //std::cout << "\t-> Noone\n";
                return none;
            }
        }

	}

    const double uphill_factor=1;       //different costs for uphill and downhill movements
    const double downhill_factor=1;
    float determineEdgeWeight(const Geometry::Vec3& v1, const Geometry::Vec3& v2) {

       // float deltaY = v1[y] - v2[y];       // Y axis is the ascent or descent as per PADrend
        // std::cout << "\tdeltaY: " << deltaY ;//<< '\n';
        // std::cout << "\tv2: " <<v2[z]<<"\tv1: " <<v1[z]<<" ";
        std::cout <<"\tedge weight"<< v2[y] - v1[y];
        return (v2[y] - v1[y]);
//        if (v2[y] > v1[y]) {
//
//           // std::cout << "uphill: "<<(v2[y] - v1[y]) * uphill_factor;
//            return (v2[y] - v1[y]) * uphill_factor;
//        } else {
//
//          //  std::cout << "downhill: "<<(v1[y] - v2[y]) * downhill_factor;
//            return (v1[y] - v2[y]) * downhill_factor;
//
//        }

//        if(deltaY<0)
//            return 1;
//        else
//            return 2;

    }

    std::vector<Graph*> buildGraphList(
		const std::vector<MinSG::GeometryNode*>& nodes, 
		const std::vector<MeshAccessor*>& meshAccessors
	) {
		// Start the timer for this function
		spdlog::stopwatch sw;

		std::cout << yellow << "Construct individual graphs: " << reset;


        std::vector<Graph*> graphList;
		// Create one graph per GeometryNode
		for(size_t i = 0; i < nodes.size(); ++i) {
			std::cout << i << " ";

			Graph* currentGraph = new Graph();
			MeshAccessor* currentMeshAccessor = meshAccessors[i];
			std::vector<Vertex> vertices;


			// Add all vertices to currentGraph and store them in vertices
			for(uint32_t vertexId = 0; 
				vertexId < currentMeshAccessor->getVertexCount(); ++vertexId
			) {
				Vertex v = 
					boost::add_vertex(
						GraphVertex{
							currentMeshAccessor->getVertex(vertexId), 
							currentMeshAccessor
						}, 
						*currentGraph
					);
				vertices.push_back(v);
			}
            std::cout << '\n'<<"Total number of vertices: "<<vertices.size();

			// Add all edges
			for(uint32_t vertexId = 0; vertexId < vertices.size(); ++vertexId) {
				for(uint32_t neighbourId : 
					currentMeshAccessor->getAdjacentVertexIds(vertexId)
				) {
					boost::add_edge(
						vertices[vertexId], 
						vertices[neighbourId], 
						GraphEdge{
							distance3(
								(*currentGraph)[vertices[vertexId]].pos, 
								(*currentGraph)[vertices[neighbourId]].pos
							), 
							determineDrivability(
								(*currentGraph)[vertexId].pos, 
								(*currentGraph)[neighbourId].pos
							),
                            determineEdgeWeight(  //@author: Vineet
                                    (*currentGraph)[vertexId].pos,
                                    (*currentGraph)[neighbourId].pos
                            )
						},
						*currentGraph
					);
				}
			}

			graphList.push_back(currentGraph);
		}

		std::cout << '\n';

		spdlog::get("time_log")->info("buildGraphList: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
		return graphList;
	}

}
