/**
 * @file
 *
 * @brief This file implements all functionality declared in BuildGroups.h.
 * 
 * @author Jan-Luca Hansel
 */

#include <memory>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>
#include <chrono>

#include "BuildGroups.h"

namespace Pathfinder {

	bool allVisited(const std::vector<bool>& boolVec) {
		for(auto b : boolVec) {
			if(b == false) {
				return false;
			}
		}
		return true;
	}

	
	size_t getNextUnvisited(const std::vector<bool>& boolVec) {
		for(size_t i = 0; i < boolVec.size(); ++i) {
			if(!boolVec[i]) {
				return i;
			}
		}
		return boolVec.size();
	}


	std::vector<std::vector<Graph*>> 
	buildGroups(const std::vector<MinSG::GeometryNode*>& nodes, 
				const std::vector<Graph*>& graphList) {
		std::vector<std::vector<Graph*>> groups;
		std::vector<bool> visited(graphList.size(), false);
		std::stack<size_t> s;
		while(!allVisited(visited)) {
			// Begin a new group
			std::vector<Graph*> currentGroup;
			size_t currentId = getNextUnvisited(visited);
			visited[currentId] = true;
			s.push(currentId);
			currentGroup.push_back(graphList[currentId]);
			while(!s.empty()) {
				// While the stack is not empty
				currentId = s.top();
				s.pop();
				Graph* currentGraph = graphList[currentId];
				for(size_t i = 0; i < graphList.size(); ++i) {
					// Look at each other mesh
					Graph* g = graphList[i];
					if(g == currentGraph) {
						continue;
					}
					MeshAccessor* currentGraphMeshAccessor = 
						(*currentGraph)[*boost::vertices(*currentGraph).first].ma;
					MeshAccessor* gMeshAccessor = 
						(*g)[*boost::vertices(*g).first].ma;
					if(!visited[i] && 
					   AABBintersection(currentGraphMeshAccessor->getAABB(), 
										gMeshAccessor->getAABB())
					) {
						// If another mesh intersects with the current mesh, 
						// it is added to the group and pushed on the stack
						s.push(i);
						visited[i] = true;
						currentGroup.push_back(g);
					}
				}
			}
			groups.push_back(currentGroup);
		}
		return groups;
	}

}
