/**
 * @file
 * 
 * @brief This file declares all functionality needed to create groups of
 *        possibly intersecting meshes.
 *
 * @author Jan-Luca Hansel
 */

#ifndef BUILDGROUPS_H
#define BUILDGROUPS_H

#include <stack>

#include "MeshAccessor.h"
#include "Using.h"

namespace Pathfinder {

	/**
	 * @brief Check whether all entries in boolVec are true.
	 *
	 * @param boolVec A vector of boolean values.
	 *
	 * @return Return true if all entries in boolVec are true, false otherwise.
	 */
	bool 
	allVisited(
		const std::vector<bool>& boolVec
	); 


	/**
	 * @brief Search in boolVec for the first value equal to false and 
	 *        return its index.
	 *
	 * @param boolVec A vector of boolean values.
	 *
	 * @return Return the first index of an entry in boolVec which is set 
	 *         to false. Return boolVec.size() if all entries were true.
	 */
	size_t 
	getNextUnvisited(
		const std::vector<bool>& boolVec
	); 


	/**
	 * @brief Compute the groups of possibly intersecting meshes.
	 * @details Two meshes have the possibility of intersection of their
	 *          AABBs intersect. If the AABBs of A and B and the 
	 *          AABBs of B and C intersect, A, B, and C all end up
	 *          in the same group.
	 *
	 * @param nodes     A vector of pointers to GeometryNodes.
	 * @param graphList A vector of pointers to graphs.
	 *
	 * @return Return a vector of groups of possibly intersecting meshes. 
	 *         Each group is a vector of pointers to graphs.
	 */
	std::vector<std::vector<Graph*>> 
	buildGroups(
		const std::vector<MinSG::GeometryNode*>& nodes, 
		const std::vector<Graph*>& graphList
	); 

}

#endif
