/** 
 * @file
 * 
 * @brief This file declares all functionality that is used to compare
 *        various data types.
 *
 * @author Jan-Luca Hansel
 */

#ifndef COMPARE_H
#define COMPARE_H

#include <tuple>

#include <Geometry/Vec3.h>

namespace Pathfinder {

	/**
	 * @brief Compare two Vec3 instances for a lexicographic < relation.
	 *
	 * @param l The first Vec3.
	 * @param r The second Vec3.
	 *
	 * @return Return true if l < r, false otherwise.
	 */
	bool 
	compareSmallerVec3(
		const Geometry::Vec3& l, 
		const Geometry::Vec3& r
	); 


	/**
	 * @brief Compare two Vec3 instances for equality.
	 * @details Two Vec3s are equal if their contents are the same.
	 *
	 * @param l The first Vec3.
	 * @param r The second Vec3.
	 *
	 * @return Return true if l == r, false otherwise.
	 */
	bool 
	compareEqualVec3(
		const Geometry::Vec3& l, 
		const Geometry::Vec3& r
	); 


	/**
	 * @brief Compare two Vec3 instances for a lexicographic <= relation.
	 *
	 * @param l The first Vec3.
	 * @param r The second Vec3.
	 *
	 * @return Return true if l <= r, false otherwise.
	 */
	bool 
	compareSmallerEqualVec3(
		const Geometry::Vec3& l, 
		const Geometry::Vec3& r
	); 


	/**
	 * @brief Compare two tuples that contain Vec3 instances, for a 
	 *        lexicographic < relation.
	 *
	 * @param l The first tuple.
	 * @param r The second tuple.
	 *
	 * @return Return true if l < r, false otherwise.
	 */
	bool 
	compareSmallerVec3Tuple(
		const std::tuple<Geometry::Vec3, Geometry::Vec3>& l, 
		const std::tuple<Geometry::Vec3, Geometry::Vec3>& r
	); 

	
	/**
	 * @brief Compare two tuples that contain Vec3 instances, for equality.
	 * @details Two such tuples are equal if their contents are the same.
	 *
	 * @param l The first tuple.
	 * @param r The second tuple.
	 *
	 * @return Return true if l == r, false otherwise.
	 */
	bool 
	compareEqualVec3Tuple(
		const std::tuple<Geometry::Vec3, Geometry::Vec3>& l, 
		const std::tuple<Geometry::Vec3, Geometry::Vec3>& r
	); 


	/**
	 * @brief Compare two tuples which contain tuples, for a lexicographic 
	 *        < relation.
	 *
	 * @param l The first tuple.
	 * @param r The second tuple.
	 *
	 * @return Return true if l < r, false otherwise.
	 */
	bool 
	compareSmallerVec3TupleTuple(
		const std::tuple<std::tuple<Geometry::Vec3, Geometry::Vec3>, 
						 std::tuple<Geometry::Vec3, Geometry::Vec3>>& l, 
		const std::tuple<std::tuple<Geometry::Vec3, Geometry::Vec3>, 
						 std::tuple<Geometry::Vec3, Geometry::Vec3>>& r
	); 


	/**
	 * @brief Compare two tuples which contain tulpes, for equality.
	 * @details Two such tuples are equal if their contents are the same.
	 *
	 * @param l The first tuple.
	 * @param r The second tuple.
	 *
	 * @return Return true if l == r, false otherwise.
	 */
	bool 
	compareEqualVec3TupleTuple(
		const std::tuple<std::tuple<Geometry::Vec3, Geometry::Vec3>, 
						 std::tuple<Geometry::Vec3, Geometry::Vec3>>& l, 
		const std::tuple<std::tuple<Geometry::Vec3, Geometry::Vec3>, 
						 std::tuple<Geometry::Vec3, Geometry::Vec3>>& r
	); 


	/**
	 * @brief Compare two float values for equality.
	 * @details The given values are not compared by the use of ==. This
	 *          function computes the distance of l and r and checks if this
	 *          lies within some epsilon. If so, the two values are considered
	 *          equal.
	 *
	 * @param l The first float value.
	 * @param r The second float value.
	 *
	 * @return Return true if l == r, false otherwise.
	 */
	bool 
	compareEqualFloat(
		const float& l, 
		const float& r
	); 


	/**
	 * @brief Compare two float values for a < relation.
	 * @details This function first checks if l and r are the same by the use 
	 *          of compareEqualFloat. If so, this function returns false. 
	 *          Otherwise it returns what is returned by l < r.
	 * @warning If this returns false, l <= r does not hold!
	 *
	 * @param l The first float value.
	 * @param r The second float value.
	 *
	 * @return Return true if l < r, false otherwise.
	 */
	bool 
	compareSmallerFloat(
		const float& l, 
		const float& r
	); 
}

#endif
