/**
 * @file
 *
 * @brief This file implements all functionality declared in Flat.h.
 *
 * @author Jan-Luca Hansel
 */

#include <boost/optional/optional_io.hpp>
#include <memory>
#include <chrono>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include "Flat.h"
#include "TupleAccess.h"
#include "Compare.h"

namespace Pathfinder {

	/** 
	 * Easier description of a point being above, below or on a line.
	 */
	enum relativePosition : short {below, online, above};


	short coordToDrop(const Polygon* tri) {
		// Determine the normal and search the index of the maximum value
		Geometry::Vec3 n = tri->normal;
		short maxId = x;
		float maxVal = std::abs(n[x]);
		for(short i = y; i <= z; ++i) {
			if(std::abs(n[i]) > maxVal) {
				maxVal = std::abs(n[i]);
				maxId = i;
			}
		}
		// Return the coordinate where the maximum was found
		if(maxId == x) {
			return x;
		} else if(maxId == y) {
			return y;
		} else {
			return z;
		}
	}


	std::tuple<Geometry::Vec2, Geometry::Vec2, Geometry::Vec2> 
	flattenTri(const Polygon* tri, const short& coord) {
		// Based on the maximum value, build 2D vectors which omit the component 
		// coord
		switch(coord) {
			case x: return {{tri->p1[y], tri->p1[z]}, 
							{tri->p2[y], tri->p2[z]}, 
							{tri->p3[y], tri->p3[z]}};
			case y: return {{tri->p1[x], tri->p1[z]}, 
							{tri->p2[x], tri->p2[z]}, 
							{tri->p3[x], tri->p3[z]}};
			case z: return {{tri->p1[x], tri->p1[y]}, 
							{tri->p2[x], tri->p2[y]}, 
							{tri->p3[x], tri->p3[y]}};
			default: spdlog::get("warn_log")->info("flattenTri() returns \{}.");
					 return {};
		}
	}


	Geometry::Vec2 flattenVec3(const Geometry::Vec3& a, const short& coord) {
		switch(coord) {
			case x: return {a[y], a[z]};
			case y: return {a[x], a[z]};
			case z: return {a[x], a[y]};
			default: spdlog::get("warn_log")->info("flattenVec3() returns \{}.");
					 return {}; 
		}
	}


	std::tuple<short, short, short> determineRelativePos(
		Geometry::Vec2 point, const std::tuple<Line2, Line2, Line2>& tri
	) {
		std::tuple<short, short, short> relPos = {-1, -1, -1};
		// Look at each bounding line
		for(short i = 0; i < 3; ++i) {
			// Compute the point on the current bounding line for the x of the 
			// point
			boost::optional<float> computedY = getLine2(tri, i).eval(point[x]);

			// Check if the current bounding line is vertical
			if(computedY != boost::none) {
				// Not vertical -> Compare y positions
				if(*computedY < point[y]) {
					// Point is above
					*getShortPtr(relPos, i) = above;
				} else if(compareEqualFloat(*computedY, point[y])) {
					// Point lies on the line
					*getShortPtr(relPos, i) = online;
				} else {
					// Point is below 
					*getShortPtr(relPos, i) = below;
				}
			} else {
				// Vertical -> Compare x positions 
				if(*getLine2(tri, i).xpos < point[x]) {
					// Point is above
					*getShortPtr(relPos, i) = above;
				} else if(compareEqualFloat(*getLine2(tri, i).xpos, point[x])) {
					// Point lies on the line
					*getShortPtr(relPos, i) = online;
				} else {
					// Point is below 
					*getShortPtr(relPos, i) = below;
				}
			}
		}
		return relPos;
	}


	std::vector<bool> pointVecInsideTri(
		const std::vector<Geometry::Vec3>& pointVec, 
		const std::tuple<Line2, Line2, Line2>& tri, 
		const short& droppedCoord, 
		const std::tuple<short, short, short>& relPosCenter
	) {
		std::vector<bool> inside(pointVec.size(), false);
		// Look at each point
		for(unsigned short i = 0; i < pointVec.size(); ++i) {
			// Flatten the point
			Geometry::Vec2 pointFlat = flattenVec3(pointVec[i], droppedCoord);
			// Determine the relative positions
			std::tuple<short, short, short> relPosPoint = 
				determineRelativePos(pointFlat, tri);
			// Above/Below must always match between the current point and the center point
			// online can be treated as both
			if((std::get<0>(relPosPoint) == online || 
				std::get<0>(relPosPoint) == std::get<0>(relPosCenter)) && 
			   (std::get<1>(relPosPoint) == online || 
				std::get<1>(relPosPoint) == std::get<1>(relPosCenter)) && 
			   (std::get<2>(relPosPoint) == online || 
				std::get<2>(relPosPoint) == std::get<2>(relPosCenter))) {
				inside[i] = true;
			}
		}
		return inside;
	}


	std::vector<bool> pointVecInsideTri(
		const std::vector<Geometry::Vec2>& pointVec, 
		const std::tuple<Line2, Line2, Line2>& tri, 
		const std::tuple<short, short, short>& relPosCenter
	) {
		// Start timer for this function
		spdlog::stopwatch sw;

		std::vector<bool> inside(pointVec.size(), false);
		// Look at each point
		for(unsigned short i = 0; i < pointVec.size(); ++i) {
			Geometry::Vec2 currentPoint = pointVec[i];
			// Determine the relative positions
			std::tuple<short, short, short> relPosPoint = 
				determineRelativePos(currentPoint, tri);
			// Above/Below must always match between the current point and the 
			// center point online can be treated as both
			if((std::get<0>(relPosPoint) == online || 
				std::get<0>(relPosPoint) == std::get<0>(relPosCenter)) && 
			   (std::get<1>(relPosPoint) == online || 
				std::get<1>(relPosPoint) == std::get<1>(relPosCenter)) && 
			   (std::get<2>(relPosPoint) == online || 
				std::get<2>(relPosPoint) == std::get<2>(relPosCenter))) {
				inside[i] = true;
			}
		}
		spdlog::get("time_log")->info("pointVecInsideTri(2D): {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
		return inside;
	}


	bool pointInsideTri(
		const Geometry::Vec3& point, 
		const std::tuple<Line2, Line2, Line2>& tri, 
		const short& droppedCoord, 
		const std::tuple<short, short, short>& relPosCenter
	) {
		// Flatten the point
		Geometry::Vec2 pointFlat = flattenVec3(point, droppedCoord);
		// Determine the relative positions
		std::tuple<short, short, short> relPosPoint = 
			determineRelativePos(pointFlat, tri);
		// Above/Below must always match between the current point and the 
		// center point online can be treated as both
		if((std::get<0>(relPosPoint) == online || 
			std::get<0>(relPosPoint) == std::get<0>(relPosCenter)) && 
		   (std::get<1>(relPosPoint) == online || 
			std::get<1>(relPosPoint) == std::get<1>(relPosCenter)) && 
		   (std::get<2>(relPosPoint) == online || 
			std::get<2>(relPosPoint) == std::get<2>(relPosCenter))) {
			return true;
		}
		return false;
	}


	bool pointInsideTri(
		const Geometry::Vec2& point, 
		const std::tuple<Line2, Line2, Line2>& tri, 
		const std::tuple<short, short, short>& relPosCenter
	) {
		// Determine the relative positions
		std::tuple<short, short, short> relPosPoint = 
			determineRelativePos(point, tri);
		// Above/Below must always match between the current point and the 
		// center point online can be treated as both
		if((std::get<0>(relPosPoint) == online || 
			std::get<0>(relPosPoint) == std::get<0>(relPosCenter)) && 
		   (std::get<1>(relPosPoint) == online || 
			std::get<1>(relPosPoint) == std::get<1>(relPosCenter)) && 
		   (std::get<2>(relPosPoint) == online || 
			std::get<2>(relPosPoint) == std::get<2>(relPosCenter))) {
			return true;
		}
		return false;
	}


	bool pointInRectangle(
		Geometry::Vec2& p, 
		Geometry::Vec2& lowerLeft, 
		Geometry::Vec2& upperRight
	) {
		if(lowerLeft[x] <= p[x] && 
		   p[x] <= upperRight[x] &&
		   lowerLeft[y] <= p[y] && 
		   p[y] <= upperRight[y]) {
			return true;
		} else {
			return false;
		}
	}

	
	std::tuple<Geometry::Vec2, Geometry::Vec2> buildRectangle(
		Geometry::Vec2& pi, Geometry::Vec2& pj
	) {
		// Check which operators to use by checking if pi or pj is smaller
		if(pi[x] <= pj[x] && pi[y] <= pj[y]) {
			// + -- pj
			// |     |
			// pi -- +
			return {pi, pj};
		} else if(pi[x] <= pj[x] && pi[y] > pj[y]) {
			// pi -- + 
			// |     |
			// + -- pj
			Geometry::Vec2 lowerLeft(pi[x], pj[y]);
			Geometry::Vec2 upperRight(pj[x], pi[y]);
			return {lowerLeft, upperRight};
		} else if(pi[x] > pj[x] && pi[y] <= pj[y]) {
			// pj -- + 
			// |     |
			// + -- pi
			Geometry::Vec2 lowerLeft(pj[x], pi[y]);
			Geometry::Vec2 upperRight(pi[x], pj[y]);
			return {lowerLeft, upperRight};
		} else {
			// + -- pi
			// |     |
			// pj -- +
			return {pj, pi};
		}
	}

}
