/**
 * @file
 *
 * @brief This file declares all functionality related to basic operations for 
 *        flattened Polygons.
 *
 * @author Jan-Luca Hansel
 */

#ifndef FLAT_H
#define FLAT_H

#include "Line2.h"

namespace Pathfinder {

	/**
	 * @brief Decide which coordinate shall be dropped for flattening this 
	 *        polygon.
	 * @details The component that shall be dropped is the component that has 
	 *          the greatest absolut value in the normal vector of the polygon.
	 * 
	 * @param tri The polygon.
	 *
	 * @return The component (x, y or z) which shall be dropped.
	 */
	short 
	coordToDrop(
		const Polygon* tri
	);


	/**
	 * @brief Flatten the given polygon onto the xy/xz/yz plane.
	 * @details This is done by deleting all coord elements
	 *          from the polygon's vertices.
	 *
	 * @param tri   The polygon.
	 * @param coord The coordinate which shall be dropped.
	 *
	 * @return A tuple of flattened coordinates in 2D space.
	 */
	std::tuple<Geometry::Vec2, Geometry::Vec2, Geometry::Vec2> 
	flattenTri(
		const Polygon* tri, 
		const short& coord
	); 


	/**
	 * @brief Flatten the given vector onto the xy/xz/yz plane.
	 * @details This is done by deleting the coord element from a.
	 *
	 * @param a     The vector to flatten.
	 * @param coord The coordinate which shall be dropped.
	 *
	 * @return The vector flattened to a coordinate in 2D space.
	 */
	Geometry::Vec2 
	flattenVec3(
		const Geometry::Vec3& a, 
		const short& coord
	); 


	/**
	 * @brief Determine the relative position of a point to three lines.
	 * @details For a given point and a triangle described by three lines in 
	 *          two-dimensional space, compute a tuple which contains the 
	 *          information on whether the given point lies above, on or below 
	 *          the individual bounding lines of the triangle. If a bounding 
	 *          line is vertical, above means that point is right of the line. 
	 *          If the value is set to below, this means that point is left of 
	 *          the line.
	 *
	 * @param point The point for which it shall be evaluated if it lies above, 
	 *              on or below the individual lines.
	 * @param tri   A triangle described by a tuple of three Line2 objects.
	 *
	 * @return Return a tulpe of three short values which tell if the point lies
	 *         above, on or below the corresponding line in tri.
	 */
	std::tuple<short, short, short> 
	determineRelativePos(
		Geometry::Vec2 point, 
		const std::tuple<Line2, Line2, Line2>& tri
	); 


	/**
	 * @brief For a vector of points in three-dimensional space, determine if 
	 *        the inidividual points lie inside or outside a polygon.
	 * @details A point lies inside a poylgon, if its relative positions to the 
	 *          polygon's bounding lines are the same as the relative positions
	 *          of the polygon's center point to the polygon's bounding lines.
	 *          If a point lies on a bounding line, it can be interpreted 
	 *          either way: as being above or below.
	 *
	 * @param pointVec     A vector of points where each point shall be cheked 
	 *                     for its position.
	 * @param tri          A polygon described ba a tuple of three 2D lines.
	 * @param droppedCoord The component which was dropped in the flattening of 
	 *                     tri.
	 * @param relPosCenter A tuple which describes the relative position of 
	 *                     the center point of tri to the three bounding lines 
	 *                     of tri.
	 *
	 * @return Return a vector of boolean values which indicate if the 
	 *         corresponding point in pointVec lies inside tri (true) or 
	 *         not (false).
	 */
	std::vector<bool> 
	pointVecInsideTri(
		const std::vector<Geometry::Vec3>& pointVec, 
		const std::tuple<Line2, Line2, Line2>& tri, 
		const short& droppedCoord, 
		const std::tuple<short, short, short>& relPosCenter
	); 


	/**
	 * @brief For a vector of points in two-dimensional space, determine if
	 *        the individual points lie inside or outside a polygon.
	 * @details A point lies inside a poylgon, if its relative positions to the 
	 *          polygon's bounding lines are the same as the relative positions
	 *          of the polygon's center point to the polygon's bounding lines.
	 *          If a point lies on a bounding line, it can be interpreted 
	 *          either way: as being above or below.
	 *
	 * @param pointVec     A vector of points where each point shall be cheked 
	 *                     for its position.
	 * @param tri          A polygon described ba a tuple of three 2D lines.
	 * @param relPosCenter A tuple which describes the relative position of 
	 *                     the center point of tri to the three bounding lines 
	 *                     of tri.
	 *
	 * @return Return a vector of boolean values which indicate if the 
	 *         corresponding point in pointVec lies inside tri (true) or 
	 *         not (false).
	 */
	std::vector<bool> 
	pointVecInsideTri(
		const std::vector<Geometry::Vec2>& pointVec, 
		const std::tuple<Line2, Line2, Line2>& tri, 
		const std::tuple<short, short, short>& relPosCenter
	); 


	/**
	 * @brief For a point in three-dimensional space, determine if it lies 
	 *        inside or outside a given polygon.
	 * @details A point lies inside a poylgon, if its relative positions to the 
	 *          polygon's bounding lines are the same as the relative positions
	 *          of the polygon's center point to the polygon's bounding lines.
	 *          If a point lies on a bounding line, it can be interpreted 
	 *          either way: as being above or below.
	 *
	 * @param point        The point which is to be checked.
	 * @param tri          A polygon described by a tuple of three 2D lines.
	 * @param droppedCoord The component which was dropped in the flattening 
	 *                     of tri.
	 * @param relPosCenter A tuple which describes the relative position of 
	 *                     the center point of tri to the three bounding lines 
	 *                     of tri.
	 *
	 * @return Return true if the point lies inside tri, false otherwise.
	 */
	bool 
	pointInsideTri(
		const Geometry::Vec3& point, 
		const std::tuple<Line2, Line2, Line2>& tri, 
		const short& droppedCoord, 
		const std::tuple<short, short, short>& relPosCenter
	); 


	/**
	 * @brief For a point in two-dimensional space, determine if it lies inside 
	 *        or outside a given polygon.
	 * @details A point lies inside a poylgon, if its relative positions to the 
	 *          polygon's bounding lines are the same as the relative positions
	 *          of the polygon's center point to the polygon's bounding lines.
	 *          If a point lies on a bounding line, it can be interpreted 
	 *          either way: as being above or below.
	 *
	 * @param point        The point which is to be checked.
	 * @param tri          A polygon described by a tuple of three 2D lines.
	 * @param relPosCenter A tuple which describes the relative position of the 
	 *                     center point of tri to the three bounding lines of 
	 *                     tri.
	 *
	 * @return Return true if the point lies inside tri, false otherwise.
	 */
	bool 
	pointInsideTri(
		const Geometry::Vec2& point, 
		const std::tuple<Line2, Line2, Line2>& tri, 
		const std::tuple<short, short, short>& relPosCenter
	); 


	/**
	 * @brief Given a rectangle defined by two points, check if p lies inside 
	 *        it.
	 * @details The point p lies inside the rectangle if its components are 
	 *          greater or equal to the lower left corner of the rectangle and 
	 *          smaller or equal to the upper right corner of the rectangle.
	 *
	 * @param p          The point which shall be checked.
	 * @param lowerLeft  The lower left corner of the rectangle.
	 * @param upperRight The lower left corner of the rectangle.
	 *
	 * @return Return true if p lies inside the rectangle, false otherwise.
	 */
	bool 
	pointInRectangle(
		Geometry::Vec2& p, 
		Geometry::Vec2& lowerLeft, 
		Geometry::Vec2& upperRight
	); 

	
	/**
	 * @brief Given two points, construct a tuple of points that describe 
	 *        the lower left and upper right corner of the rectangle defined 
	 *        by the two points.
	 * @details A rectangle can be defined by any two points. This function 
	 *          computes two points that define the same rectangle but are
	 *          the lower left and upper right corner of the rectangle.
	 *
	 * @param pi The first point.
	 * @param pj The second point.
	 *
	 * @return Return a tuple containing the lower left and upper right corner 
	 *         of the rectangle defined by pi and pj.
	 */
	std::tuple<Geometry::Vec2, Geometry::Vec2> 
	buildRectangle(
		Geometry::Vec2& pi, 
		Geometry::Vec2& pj
	); 

}

#endif
