/** 
 * @file
 *
 * @brief This file implements all functionality declared in Intersection.h.
 *
 * @author Jan-Luca Hansel
 */

#include <iostream>
#include <cmath>
#include <memory>
#include <chrono>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include "Intersection.h"
#include <fstream>
#include <Octree.h>

#include "TupleAccess.h"
#include "Vec.h"
#include "Flat.h"
#include "OBJ.h"
#include "Compare.h"
#include "convhull_3d/convhull_3d.h"

namespace Pathfinder {

	enum relativePosition : short {below, online, above};


	float t(const Plane& p, const Geometry::Vec3& p1, const Geometry::Vec3& p2) {
		float numerator = p.a*p2[x] + p.b*p2[y] + p.c*p2[z] + p.d;
		if(!numerator) {
			return 0;
		}
		float denominator = 
			p.a*(p1[x]-p2[x]) + p.b*(p1[y]-p2[y]) + p.c*(p1[z]-p2[z]);
		return -(numerator/denominator);
	}


	std::tuple<std::vector<short>, 
			   std::vector<short>, 
			   std::vector<short>, 
			   std::vector<short>, 
			   std::vector<short>> categorizetValueIndices(
		const std::tuple<float, float, float, float, float, float>& tValues
	) {
		std::tuple<std::vector<short>, 
				   std::vector<short>, 
				   std::vector<short>, 
				   std::vector<short>, 
				   std::vector<short>> count = {{}, {}, {}, {}, {}};
		for(short i = 0; i < 6; ++i) {
			float currenttValue = gettValue(tValues, i);
			if(std::isinf(currenttValue)) {
				// Result was inf or -inf
				if(std::signbit(currenttValue)) {
					// -inf
					std::get<0>(count).push_back(i);
				} else {
					// inf
					std::get<4>(count).push_back(i);
				}
			} else {
				// Result was a normal number
				if(currenttValue < 0) {
					std::get<0>(count).push_back(i);
				} else if(compareEqualFloat(currenttValue, 0)) { // == 0
					std::get<1>(count).push_back(i);
				} else if(0 < currenttValue && currenttValue < 1) {
					std::get<2>(count).push_back(i);
				} else if(compareEqualFloat(currenttValue, 1)) { // == 1
					std::get<3>(count).push_back(i);
				} else if(1 < currenttValue) {
					std::get<4>(count).push_back(i);
				}
			}
		}
		return count;
	}


	short determineIntersectionCase(
		const std::tuple<std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>>& categorizedtValueIndices
	) {
		if(std::get<1>(categorizedtValueIndices).size() == 6) {
			return 4;
		} else if(std::get<1>(categorizedtValueIndices).size() == 4 && 
				  std::get<3>(categorizedtValueIndices).size() == 2
		) {
			return 3;
		} else if(std::get<1>(categorizedtValueIndices).size() == 2 && 
				  std::get<3>(categorizedtValueIndices).size() == 2
		) {
			return 2;
		} else if(std::get<2>(categorizedtValueIndices).size() == 4) {
			return 1;
		} else {
			return 0;
		}
	}


	std::tuple<short, short> tIndex2endpoints(const short& id) {
		switch(id) {
			case 0: return {0, 1};
			case 1: return {0, 2};
			case 2: return {1, 0};
			case 3: return {1, 2};
			case 4: return {2, 0};
			case 5: return {2, 1};
			default: spdlog::get("warn_log")->info("tIndex2endpoints returns \{-1, -1}.");
					 return {-1, -1};
		}
	}


	short endpoints2tIndex(const std::tuple<short, short>& e) {
		if(std::get<0>(e) == 0) {
			if(std::get<1>(e) == 1) {
				return 0;
			} else { // std::get<1>(e) == 2
				return 1;
			}
		} else if(std::get<0>(e) == 1) {
			if(std::get<1>(e) == 0) {
				return 2;
			} else { // std::get<1>(e) == 2
				return 3;
			}
		} else { //std::get<0>(e) == 2
			if(std::get<1>(e) == 0) {
				return 4;
			} else { // std::get<1>(e) == 1
				return 5;
			}
		}
	}


	short getEquivalenttValueIndex(const short& id) {
		switch(id) {
			case 0: return 2;
			case 1: return 4;
			case 2: return 0;
			case 3: return 5;
			case 4: return 1;
			case 5: return 3;
			default: spdlog::get("time_log")->info("getEquivalenttValueIndex returns 6.");
					 return 6;
		}
	}


	std::tuple<bool, bool, bool> intersectionTupleValid(
		Geometry::Vec2& pi, 
		Geometry::Vec2& pj, 
		const std::tuple<boost::optional<float>, 
						 boost::optional<float>, 
						 boost::optional<float>>& intersections, 
		const std::tuple<Line2, Line2, Line2>& tri, 
		const std::tuple<short, short, short>& relPosCenter
	) {
		// Look at each computed intersection and check if the x value of 
		// the intersection lies between the x values of the endpoints of 
		// the non-boundary line. If it does, mark this as valid/true
		std::tuple<bool, bool, bool> valid = {false, false, false};

		// Create the line between pi and pj
		Line2 lij(pi, pj);

		// Construct a rectangle between pi and pj 
		std::tuple<Geometry::Vec2, Geometry::Vec2> rect = buildRectangle(pi, pj);

		// Look at each intersection
		for(short i = 0; i < 3; ++i) {
			// Get the x coordinate of the current intersection
			boost::optional<float> currentIntersectionX = 
				getOptionalFloat(intersections, i);

			// If currentIntersectionX is boost::none, this intersection is 
			// invalid
			if(currentIntersectionX == boost::none) {
				*getBoolPtr(valid, i) = false;
			} else {
				// currentIntersectionX is not boost::none 
				//   -> Compute y value to go with it
				boost::optional<float> currentIntersectionY = 
					lij.eval(*currentIntersectionX);

				// If currentIntersectionY is boost::none, this line is vertical
				if(currentIntersectionY == boost::none) {
					// Evaluate the corresponding line of tri 
					currentIntersectionY = 
						getLine2(tri, i).eval(*currentIntersectionX);
				} 

				// Construct an actual point out of the values
				Geometry::Vec2 currentIntersection(
					*currentIntersectionX, *currentIntersectionY
				);

				// Check if the intersection point lies in the valid area
				if(pointInRectangle(
						currentIntersection, 
						std::get<0>(rect), 
						std::get<1>(rect)) && 
				   pointInsideTri({currentIntersection}, tri, relPosCenter)) {
					// between pi and pj -> valid
					*getBoolPtr(valid, i) = true;
				} else {
					// not between pi and pj -> invalid
					*getBoolPtr(valid, i) = false;
				}
			}
		}
		return valid;
	}


	boost::optional<short> chooseIntersection(
		const std::tuple<Line2, Line2, Line2>& tri, 
		const std::tuple<boost::optional<float>, 
						 boost::optional<float>, 
						 boost::optional<float>>& intersections, 
		const std::tuple<bool, bool, bool>& valid, 
		Geometry::Vec2& pjFlat
	) {
		boost::optional<short> chosenLine;
		float intersectionDistance = std::numeric_limits<float>::max();
		// Look at each intersection
		for(short i = 0; i < 3; ++i) {
			if(getBool(valid, i)) {
				// Only consider valid intersections
				Line2 currentBoundingLine = getLine2(tri, i);
				float x = *getOptionalFloat(intersections, i);
				float y = *currentBoundingLine.eval(x);
				Geometry::Vec2 pNew(x, y);
				float currentDistance = distance2(pjFlat, pNew);

				if(chosenLine == boost::none) {
					// First valid intersection
					chosenLine = i;
					intersectionDistance = currentDistance;
				} else if(currentDistance < intersectionDistance) {
					// The current intersection has a smaller distance than the 
					// previously found one -> better
					chosenLine = i;
					intersectionDistance = currentDistance;
				}
			}
		}
		return chosenLine;
	}


	float distanceToNewPoint(
		const Line3Norm& lij, 
		const short& droppedCoord, 
		const float& pointX, 
		const float& pointY
	) {
		// Compute the distance to the new point on the line
		for(short i = x; i <= z; ++i) {
			if(i == droppedCoord) {
				// This cannot be used for computation becuase of too few known 
				// values
				continue;
			} else if(compareEqualFloat(lij.w[i], 0)) {
				// Cannot divide by 0
				continue;
			} else {
				// Look which to use: pointX or pointY
				float relevantComponent = 0;
				switch(droppedCoord) {
					case x: {
								switch(i) {
									case y: relevantComponent = pointX;
											break;
									case z: relevantComponent = pointY;
											break;
								}
							}
							break;
					case y: {
								switch(i) {
									case x: relevantComponent = pointX;
											break;
									case z: relevantComponent = pointY;
											break;
								}
							}
							break;
					case z: {
								switch(i) {
									case x: relevantComponent = pointX;
											break;
									case y: relevantComponent = pointY;
											break;
								}
							}
							break;
				}
				return (relevantComponent - lij.v[i])/lij.w[i];
			}
		}
		spdlog::get("warn_log")->info("distanceToNewPoint returns 0.");
		return 0;
	}


	boost::optional<Geometry::Vec3> cutPointToPolygon(
		const Geometry::Vec3& pi, 
		const Geometry::Vec3& pj, 
		const short& droppedCoord, 
		const std::vector<bool>& inside, 
		const Polygon* tri, 
		const std::tuple<Line2, Line2, Line2>& triLines, 
		const std::tuple<short, short, short>& relPosCenter
	) {
		// Create a 2D line which goes through pi and pj
		Geometry::Vec2 piFlat = flattenVec3(pi, droppedCoord);
		Geometry::Vec2 pjFlat = flattenVec3(pj, droppedCoord);
		Line2 lijFlat(piFlat, pjFlat);

		// Determine the x positions where lijFlat intersects with the bounding lines of tri
		std::tuple<boost::optional<float>, 
				   boost::optional<float>, 
				   boost::optional<float>> lijFlatBoundIntersections = 
				{line2Intersection(lijFlat, getLine2(triLines, 0)), 
				 line2Intersection(lijFlat, getLine2(triLines, 1)), 
				 line2Intersection(lijFlat, getLine2(triLines, 2))};

		// Determine which of these intersections is valid
		std::tuple<bool, bool, bool> valid = 
			intersectionTupleValid(piFlat, pjFlat, lijFlatBoundIntersections, 
								   triLines, relPosCenter
			); 

		// Get the intersection point which is closest to j and also valid
		boost::optional<short> chosenLine = 
			chooseIntersection(triLines, lijFlatBoundIntersections, valid, 
							   pjFlat
			); 

		// Determine the new point of intersection
		if(chosenLine != boost::none) {
			// A new valid point was found

			float cutPointX = 
				*getOptionalFloat(lijFlatBoundIntersections, *chosenLine);
			float cutPointY = 
				*getLine2(triLines, *chosenLine).eval(cutPointX);
	
			// Create the actual line between i and j
			Line3Norm lij(pi, pj);

			// Calculate the distance on the line
			float distance = 
				distanceToNewPoint(lij, droppedCoord, cutPointX, cutPointY);

			// Return the point that is distance away from the support vector
			return lij.eval(distance);
		}
		// No valid point could be found
		return boost::none;
	}


	std::vector<std::vector<Geometry::Vec3>> cutPlaneIntersectionsToPolygon(
		const Polygon* tri, 
		const std::vector<std::vector<Geometry::Vec3>>& intersections
	) {
		// Start timer for this function
		spdlog::stopwatch sw;

		// Write all resulting intersections into cutIntersections
		std::vector<std::vector<Geometry::Vec3>> cutIntersections;

		// Determine which coordinate to drop
		short droppedCoord = coordToDrop(tri);

		//std::cout << "\t\tDroppedCoord: " << droppedCoord << '\n';

		// Flatten the triangle
		std::tuple<Geometry::Vec2, Geometry::Vec2, Geometry::Vec2> 
			triFlat = flattenTri(tri, droppedCoord);

		// Create the three line equations for the triangle's boundaries
		std::tuple<Line2, Line2, Line2> 
			triLines = {Line2(std::get<0>(triFlat), std::get<1>(triFlat)),
						Line2(std::get<0>(triFlat), std::get<2>(triFlat)),
						Line2(std::get<1>(triFlat), std::get<2>(triFlat))};

		// Determine the center point of the triangle and its position relative 
		// to the three bounding lines
		Geometry::Vec2 center(
			(std::get<0>(triFlat)[x] + 
			 std::get<1>(triFlat)[x] + 
			 std::get<2>(triFlat)[x])/3, 
			(std::get<0>(triFlat)[y] + 
			 std::get<1>(triFlat)[y] + 
			 std::get<2>(triFlat)[y])/3
		);

		std::tuple<short, short, short> relPosCenter = 
			determineRelativePos(center, triLines);

		// Go through every intersection
		for(auto intersection : intersections) {
			// Write the results into this vector
			std::vector<Geometry::Vec3> cutIntersection;

			// For each endpoint of intersection, determine if it lies inside 
			// or outside the polygon
			std::vector<bool> inside = 
				pointVecInsideTri(
					intersection, 
					triLines, 
					droppedCoord, 
					relPosCenter
				); 

			if(intersection.size() == 1 && inside[0]) {
				// If only one point was found and its inside, add it to the 
				// result
				cutIntersection.push_back(intersection[0]);
			} else {
				// Check each pair of endpoints otherwise
				// If the second point of a pair is not inside the polygon, 
				// update the point in cutIntersection and in inside
				for(unsigned short i = 0; i < intersection.size(); ++i) {
					for(unsigned short j = 0; j < intersection.size(); ++j) {
						if(i == j) {
							// i and j need to be different -> nothing to do
						} else if(inside[j]) {
							// j is inside the polygon 
							//   -> Just add j into cutIntersection
							cutIntersection.push_back(intersection[j]);
						} else {
							boost::optional<Geometry::Vec3> cutPoint = 
								cutPointToPolygon(
									intersection[i], 
									intersection[j], 
									droppedCoord, 
									inside, 
									tri, 
									triLines, 
									relPosCenter
								); 

							// Try to move the point to lie inside the tri
							if(cutPoint != boost::none) {
								cutIntersection.push_back(*cutPoint);
							}
						}
					}
				}
			}

			// Eliminate all duplicates from cutIntersection
			std::sort(
				cutIntersection.begin(), 
				cutIntersection.end(), 
				compareSmallerVec3
			);
			auto last = 
				std::unique(
					cutIntersection.begin(), 
					cutIntersection.end(), 
					compareEqualVec3
				);
			cutIntersection.erase(last, cutIntersection.end());
			cutIntersections.push_back(cutIntersection);
		}
		spdlog::get("time_log")->info("cutPlaneIntersectionToPolygon: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
		return cutIntersections;
	}


	std::tuple<float, float, float, float, float, float> computetValues(
		const Plane& p, 
		const Polygon* tri
	) {
		return {t(p, tri->p1, tri->p2),
				t(p, tri->p1, tri->p3),
				t(p, tri->p2, tri->p1),
				t(p, tri->p2, tri->p3),
				t(p, tri->p3, tri->p1),
				t(p, tri->p3, tri->p2)};
	}


	std::vector<uint32_t> polygonInAABBIds(MeshAccessor* ma, AABB aabb) {
		// Start timer for this function
		spdlog::stopwatch sw;
		
		//std::cout << "Polygons of bigger in smaller's AABB:\n";

		std::vector<uint32_t> polygons;
		for(uint32_t id = 0; 3 * id < ma->getIndexCount(); ++id) {
			Polygon* p = ma->getPolygonPtr(id);
			if(aabb.triInAABB(p)) {
				// Intersection was detected -> Add into polygons
				polygons.push_back(id);
			}
		}
		spdlog::get("time_log")->info("polygonInAABBIds: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
		return polygons;
	}


	std::vector<Geometry::Vec3> handleIntersectionCase1(
		Polygon* tri, 
		const std::tuple<float, float, float, float, float, float>& tValues, 
		const std::tuple<std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>>& categorizedtValueIndices
	) {
		// Save the edges that hold the intersection points 
		// in relevant edges
		std::tuple<std::tuple<short, short>, 
				   std::tuple<short, short>> relevantEdges = {{}, {}};
		bool setFirst = false;

		// Look at each of the four edges in category 2
		for(short i = 0; i < 4; ++i) {
			short id = std::get<2>(categorizedtValueIndices)[i];

			// Only add the edge which has the smaller index in 
			// tValues to avoid duplicates
			if(id < getEquivalenttValueIndex(id)) {
				if(!setFirst) {
					std::get<0>(relevantEdges) = tIndex2endpoints(id);
					setFirst = true;
				} else {
					std::get<1>(relevantEdges) = tIndex2endpoints(id);
				}
			}
		}


		// Construct the two lines defined by the edges that intersect the plane
		Line3Unnorm e1(tri->getVertex(std::get<0>(std::get<0>(relevantEdges))), 
					   tri->getVertex(std::get<1>(std::get<0>(relevantEdges))));
		Line3Unnorm e2(tri->getVertex(std::get<0>(std::get<1>(relevantEdges))), 
					   tri->getVertex(std::get<1>(std::get<1>(relevantEdges))));

		// Compute the intersection point on each line and return the line
		// defined by these endpoints 
		short tIdx1 = endpoints2tIndex(std::get<0>(relevantEdges));
		short tIdx2 = endpoints2tIndex(std::get<1>(relevantEdges));

		return {e1.eval(gettValue(tValues, tIdx1)), 
				e2.eval(gettValue(tValues, tIdx2))};
	}


	std::vector<Geometry::Vec3> handleIntersectionCase2(
		const Polygon* tri, 
		const std::tuple<float, float, float, float, float, float>& tValues, 
		const std::tuple<std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>>& categorizedtValueIndices
	) {
		// Store the endpoints in the following list
		std::vector<Geometry::Vec3> intersectionEndpoints;

		// Store the index of the vertex which lies on the plane, in 
		// vertexOnPlane. This point is necessary in all cases
		short vertexOnPlane = 0;

		// Add the endpoint which is involved in the two edges 
		// resulting in t = 0
		// t() returns 0 if the second point intersects with the plane
		// -> Check the computed t values for those pairs with the same
		// second endpoint. If both are equal to 0, this point must be 
		// the intersection
		if((std::get<1>(categorizedtValueIndices)[0] == 2 && 
			std::get<1>(categorizedtValueIndices)[1] == 4) ||
		   (std::get<1>(categorizedtValueIndices)[0] == 4 && 
			std::get<1>(categorizedtValueIndices)[1] == 2)) {
			// First vertex results in t = 0
			intersectionEndpoints.push_back({tri->p1});
		} else if((std::get<1>(categorizedtValueIndices)[0] == 0 && 
				   std::get<1>(categorizedtValueIndices)[1] == 5) ||
				  (std::get<1>(categorizedtValueIndices)[0] == 5 && 
				   std::get<1>(categorizedtValueIndices)[1] == 0)) {
			// Second vertex results in t = 0
			intersectionEndpoints.push_back({tri->p2});
			vertexOnPlane = 1;
		} else if((std::get<1>(categorizedtValueIndices)[0] == 1 &&
				   std::get<1>(categorizedtValueIndices)[1] == 3) ||
				  (std::get<1>(categorizedtValueIndices)[0] == 3 && 
				   std::get<1>(categorizedtValueIndices)[1] == 1)) {
			// Third vertex results in t = 0
			intersectionEndpoints.push_back({tri->p3});
			vertexOnPlane = 2;
		}

		// Check if one edge goes through the plane to determine case i. or ii.
		if(std::get<2>(categorizedtValueIndices).size() == 0) {
			// i. Noting more needs to be done
			//std::cout << "\t\t\tCase i.\n";
		} else {
			// ii. Calculate the additional intersection point
			// Check which vertex intersects with the plane 
			//   -> The other two edges not involved with this point
			// need to be considered for the additional intersection computation
			//std::cout << "\t\t\tCase ii. ";
			intersectionEndpoints.push_back(
				handleIntersectionCase2ii(tri, vertexOnPlane, tValues)
			);
		}

		// intersectionEndpoints contains 1 (i.) or 2 (ii.) points and can be 
		// returned
		return intersectionEndpoints;
	}


	Geometry::Vec3 handleIntersectionCase2ii(
		const Polygon* tri, 
		const short& vertexOnPlane, 
		const std::tuple<float, float, float, float, float, float>& tValues
	) {
		switch(vertexOnPlane) {
			case 0: {
						// Construct a line with the two points that do not have 
						// index 0
						Line3Norm e(tri->p2, tri->p3);

						// Return the intersection point of the line with the 
						// plane
						return e.eval(gettValue(tValues, endpoints2tIndex({1, 2})));
					}
			case 1: {
						// Construct a line with the two points that do not have 
						// index 1
						Line3Norm e(tri->p1, tri->p3);

						// Return the intersection point of the line with the 
						// plane
						return e.eval(gettValue(tValues, endpoints2tIndex({0, 2})));
					}
			case 2: {
						// Construct a line with the two points that do not have 
						// index 2
						Line3Norm e(tri->p1, tri->p2);

						// Return the intersection point of the line with the 
						// plane
						return e.eval(gettValue(tValues, endpoints2tIndex({0, 1})));
					}
			default: spdlog::get("warn_log")->info("handleIntersectionCase2ii returns \{}.");
					 return {};
		}
	}


	std::vector<Geometry::Vec3> handleIntersectionCase3(
		Polygon* tri, 
		const std::tuple<std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>, 
						 std::vector<short>>& categorizedtValueIndices
	) {
		// e1 and e2 yield 0. However, reversed, they yield 1 -> unwanted!
		short e1 = 
			getEquivalenttValueIndex(std::get<3>(categorizedtValueIndices)[0]);
		short e2 = 
			getEquivalenttValueIndex(std::get<3>(categorizedtValueIndices)[1]);

		for(short i = 0; i < 4; ++i) {
			// Look at each edge that yields 0
			short current = std::get<1>(categorizedtValueIndices)[i];

			if(current == e1 || current == e2) {
				// They are not the desired edges since they result in 1 when 
				// reversed
				continue;
			} else {
				// Desired edge found
				// To avoid duplicates, only return the edge 
				// with the smaller index in tValues
				if(current < getEquivalenttValueIndex(current)) {
					return {tri->getVertex(std::get<0>(tIndex2endpoints(current))), 
							tri->getVertex(std::get<1>(tIndex2endpoints(current)))};
				}
			}
		}
		spdlog::get("warn_log")->info("handleIntersectionCase3 retruns \{}.");
		return {};
	}

	//@author Vineet Check if object is parallel to axis
	bool isParallelToAxis(const Geometry::Vec3& normal) {

		//std::cout << "Checking if object is parallel to axis" << std::endl;

		// Retrieve and print vector components for debugging
		float x = normal.x();
		float y = normal.y();
		float z = normal.z();

		//std::cout << "Vector components: x = " << x << ", y = " << y << ", z = " << z << std::endl;

		float epsilon = 1e-6;  //epsilon is used to check if the components of a vector are close enough to zero

		bool isParallel = (std::fabs(x) < epsilon && std::fabs(y) < epsilon) ||   //fabs() function returns the absolute (positive) value of a number
						  (std::fabs(x) < epsilon && std::fabs(z) < epsilon) ||
						  (std::fabs(y) < epsilon && std::fabs(z) < epsilon);

		//std::cout << "Is parallel to axis: " << (isParallel ? "Yes" : "No") << std::endl;

		return isParallel;
	}

	// Function to create a convex hull using convhull_3d
	std::vector<Geometry::Vec3> createConvexHull(const std::vector<Geometry::Vec3>& points) {
		// Convert points to the format required by convhull_3d
		std::vector<ch_vertex> vertices(points.size());
		for (size_t i = 0; i < points.size(); ++i) {
			vertices[i].x = points[i].x();
			vertices[i].y = points[i].y();
			vertices[i].z = points[i].z();
		}

		// Prepare the output containers for faces
		int* face_array = nullptr;
		int num_faces = 0;

		// Call convhull_3d library function to create the convex hull
		convhull_3d_build(vertices.data(), static_cast<int>(vertices.size()), &face_array, &num_faces);

		// Copy face data
		std::vector<int> faces(face_array, face_array + num_faces * 3);

		// Clean up allocated arrays by convhull_3d_build
		free(face_array);

		// Extract the vertices of the convex hull
		std::vector<Geometry::Vec3> hullVertices;
		for (size_t i = 0; i < faces.size(); i += 3) {
			hullVertices.push_back(Geometry::Vec3(vertices[faces[i]].x, vertices[faces[i]].y, vertices[faces[i]].z));
			hullVertices.push_back(Geometry::Vec3(vertices[faces[i + 1]].x, vertices[faces[i + 1]].y, vertices[faces[i + 1]].z));
			hullVertices.push_back(Geometry::Vec3(vertices[faces[i + 2]].x, vertices[faces[i + 2]].y, vertices[faces[i + 2]].z));
		}

		return hullVertices;
	}


	//Under test for object not parallel to axis
	std::vector<std::tuple<Graph*, Graph*, std::vector<Geometry::Vec3>>> 
	intersectionEXP(Graph* smaller, Graph* bigger) {    ///Experimental
		// Start timer for this function
		spdlog::stopwatch sw;

		std::cout << yellow << "Start intersection" << reset << std::endl;

		//TODO Octree here

		std::vector<std::vector<Geometry::Vec3>> intersections;

		// Easier access to the MeshAccessor objects
		MeshAccessor* smallerMeshAccessor = 
			(*smaller)[*boost::vertices(*smaller).first].ma;
		MeshAccessor* biggerMeshAccessor = 
			(*bigger)[*boost::vertices(*bigger).first].ma;

		// Put the ids of all polygons into smallerPolygons since all of these 
		// will be checked
		std::vector<uint32_t> smallerPolygons;
		for(uint32_t id = 0; 3*id < smallerMeshAccessor->getIndexCount(); ++id) {
			smallerPolygons.push_back(id);
		}

		// Put the ids of all polygons which intersect with the AABB of smaller 
		// into biggerPolygons since all of these will be checked
		std::vector<uint32_t> biggerPolygons =
			polygonInAABBIds(biggerMeshAccessor, smallerMeshAccessor->getAABB());



		//----------------------------------------------------------
		// TODO Adjust AABB of bigger!
		// Now we have two vectors of possibly intersecting polygons
		// TODO? I skip over the preceding discard of polygons and 
		// apply the formula to all other polygons
		//----------------------------------------------------------	

		// For every poylgon in smallerPolygons, check for intersection with 
		// every polygon of biggerPolygons
		for(uint32_t smallId = 0; smallId < smallerPolygons.size(); ++smallId) {
			Polygon* smallerTri = smallerMeshAccessor->getPolygonPtr(smallId);
			if (isParallelToAxis(smallerTri->normal)) {   //@author Vineet
				// Perform AABB intersection checks or other axis-specific operations

			// Check if the current polygon of smaller intersects with the AABB 
			// of bigger
			if(!biggerMeshAccessor->getAABB().triInAABB(smallerTri)) {
				// This polygon cannot intersect -> skip
				continue;
			}

			// Intersection possible
			// Construct the plane defined by smallerTri
			Plane p(smallerTri->p1, smallerTri->p2, smallerTri->p3);

			// this list will contain all intersections between smallerTri and 
			// all polygons in biggerPolygons
			std::vector<std::vector<Geometry::Vec3>> localIntersections;

			// Look at each polygon in biggerPolygons
			for(auto biggerId : biggerPolygons) {
				Polygon* biggerTri = 
					biggerMeshAccessor->getPolygonPtr(biggerId);

				// Compute all 6 t values and save the results in tValues
				std::tuple<float, float, float, float, float, float> tValues = 
					computetValues(p, biggerTri);

				// Categorize the indices of the computed values into categories 
				// according to membership in specific intervals
				std::tuple<std::vector<short>, 
						   std::vector<short>, 
						   std::vector<short>, 
						   std::vector<short>, 
						   std::vector<short>> categorizedtValueIndices = 
					categorizetValueIndices(tValues);

				// Look at the number of t values in each category and 
				// decide if an intersection occurs
				// If an intersection occurs, decide which case it is
				std::vector<Geometry::Vec3> localIntersection;
				// std::cout << "determineIntersectionCase" << std::endl;

				switch(determineIntersectionCase(categorizedtValueIndices)) {
					case 0: // No intersection -> Nothing needs to be added
							break;
					case 1: // Case 1: Calculate endpoints with the help of t
							localIntersection = 
								handleIntersectionCase1(
									biggerTri, 
									tValues, 
									categorizedtValueIndices
								);
							break;
					case 2: // Case 2: 
							localIntersection = 
								handleIntersectionCase2(
									biggerTri, 
									tValues, 
									categorizedtValueIndices
								);
							break;
					case 3: // Case 3: Add the edge that never yields 1
							localIntersection = 
								handleIntersectionCase3(
									biggerTri, 
									categorizedtValueIndices
								);
							break;
					case 4: // Case 4: Add all vertices
							localIntersection = 
								{biggerTri->p1, biggerTri->p2, biggerTri->p3};
							break;
				}

				// Check if a point of smaller lies inside bigger
				// Flatten all three ways and check if any vertex lies in all 
				// resulting triangles
				// Drop x
				std::tuple<Geometry::Vec2, Geometry::Vec2, Geometry::Vec2> 
					biggerFlatX = flattenTri(biggerTri, x);
				std::tuple<Line2, Line2, Line2> 
					biggerLinesX = 
						{Line2(std::get<0>(biggerFlatX), 
							   std::get<1>(biggerFlatX)),
						 Line2(std::get<0>(biggerFlatX), 
							   std::get<2>(biggerFlatX)),
						 Line2(std::get<1>(biggerFlatX), 
							   std::get<2>(biggerFlatX))};
				Geometry::Vec2 centerX(
					(std::get<0>(biggerFlatX)[x] + 
					 std::get<1>(biggerFlatX)[x] + 
					 std::get<2>(biggerFlatX)[x])/3, 
					(std::get<0>(biggerFlatX)[y] + 
					 std::get<1>(biggerFlatX)[y] + 
					 std::get<2>(biggerFlatX)[y])/3
				);
				std::tuple<short, short, short> relPosCenterX = 
					determineRelativePos(centerX, biggerLinesX);

				// Drop y
				std::tuple<Geometry::Vec2, Geometry::Vec2, Geometry::Vec2> 
					biggerFlatY = flattenTri(biggerTri, y);
				std::tuple<Line2, Line2, Line2> 
					biggerLinesY = 
						{Line2(std::get<0>(biggerFlatY), 
							   std::get<1>(biggerFlatY)),
						 Line2(std::get<0>(biggerFlatY), 
							   std::get<2>(biggerFlatY)),
						 Line2(std::get<1>(biggerFlatY), 
							   std::get<2>(biggerFlatY))};
				Geometry::Vec2 centerY(
					(std::get<0>(biggerFlatY)[x] + 
					 std::get<1>(biggerFlatY)[x] + 
					 std::get<2>(biggerFlatY)[x])/3, 
					(std::get<0>(biggerFlatY)[y] + 
					 std::get<1>(biggerFlatY)[y] + 
					 std::get<2>(biggerFlatY)[y])/3);
				std::tuple<short, short, short> relPosCenterY = 
					determineRelativePos(centerY, biggerLinesY);

				// Drop z
				std::tuple<Geometry::Vec2, Geometry::Vec2, Geometry::Vec2> 
					biggerFlatZ = flattenTri(biggerTri, z);
				std::tuple<Line2, Line2, Line2> 
					biggerLinesZ = 
						{Line2(std::get<0>(biggerFlatZ), 
							   std::get<1>(biggerFlatZ)),
						 Line2(std::get<0>(biggerFlatZ), 
							   std::get<2>(biggerFlatZ)),
						 Line2(std::get<1>(biggerFlatZ), 
							   std::get<2>(biggerFlatZ))};
				Geometry::Vec2 centerZ(
					(std::get<0>(biggerFlatZ)[x] + 
					 std::get<1>(biggerFlatZ)[x] + 
					 std::get<2>(biggerFlatZ)[x])/3, 
					(std::get<0>(biggerFlatZ)[y] + 
					 std::get<1>(biggerFlatZ)[y] + 
					 std::get<2>(biggerFlatZ)[y])/3);
				std::tuple<short, short, short> relPosCenterZ = 
					determineRelativePos(centerZ, biggerLinesZ);

				for(short i = 0; i < 3; ++i) {
					Geometry::Vec3 currentVertex = smallerTri->getVertex(i);
					if(pointInsideTri(
							currentVertex, biggerLinesX, x, relPosCenterX) &&   
					   pointInsideTri(
							currentVertex, biggerLinesY, y, relPosCenterY) &&   
					   pointInsideTri(
							currentVertex, biggerLinesZ, z, relPosCenterZ)) {
						// Add that point to the intersection
						localIntersection.push_back(currentVertex);
					}
				}
				localIntersections.push_back(localIntersection);
			}

			// Cut all found intersections to the smallerTri
			std::vector<std::vector<Geometry::Vec3>> cutLocalIntersections = 
				cutPlaneIntersectionsToPolygon(smallerTri, localIntersections);

			// Copy all cutLocalIntersections into intersections
			for(unsigned short i = 0; i < cutLocalIntersections.size(); ++i) {
				// Check if the current intersection is valid
				std::vector<Geometry::Vec3> currentCutLocalIntersection = 
					cutLocalIntersections[i];
				if(currentCutLocalIntersection.size() != 0) {
					intersections.push_back(currentCutLocalIntersection);
				}
			}

			} else
			{
				std::cout << "object is not parallel to axis" << std::endl;

				// Use convhull_3d to create convex hull for non-parallel object
				std::vector<Geometry::Vec3> points;

				// Assume points are filled with vertices from smallerTri and polygons from biggerPolygons

				for(auto& id : biggerPolygons) {
					Polygon* biggerTri = biggerMeshAccessor->getPolygonPtr(id);
					points.push_back(biggerTri->p1);
					points.push_back(biggerTri->p2);
					points.push_back(biggerTri->p3);
				}
				points.push_back(smallerTri->p1);
				points.push_back(smallerTri->p2);
				points.push_back(smallerTri->p3);


				if (points.size() < 4) {
					std::cerr << "Error: At least 4 points are required to compute a 3D convex hull." << std::endl;
					continue; // Skip this iteration
				}
				// Convert points to ch_vertex format required by convhull_3d
				std::vector<ch_vertex> vertices(points.size());
				for (size_t i = 0; i < points.size(); ++i) {
					vertices[i].x = static_cast<CH_FLOAT>(points[i].x());
					vertices[i].y = static_cast<CH_FLOAT>(points[i].y());
					vertices[i].z = static_cast<CH_FLOAT>(points[i].z());
				}

                // Print vertices before passing to convhull_3d_build
                std::cout << "Vertices to be used for convex hull computation:" << std::endl;
				// for (const auto& vertex : vertices) {
				// 	std::cout << "Vertex: (" << vertex.x << ", " << vertex.y << ", " << vertex.z << ")" << std::endl;
				// }

                // Prepare data structures for convex hull computation
                int* faceIndices = nullptr;
                int nFaces = 0;

                std::cout << "before convhull_3d" << std::endl;
                // Compute the convex hull
                convhull_3d_build(vertices.data(), vertices.size(), &faceIndices, &nFaces);
                std::cout << "after convhull_3d" << std::endl;
                // Check if faceIndices and nFaces are valid
                if (!faceIndices || nFaces <= 0) {
                    std::cerr << "Error: Invalid convex hull faces." << std::endl;
                    return {}; // Return an empty vector
                }

				// Extract convex hull vertices and faces
				std::vector<Geometry::Vec3> convexHullVertices(vertices.size());
				for (size_t i = 0; i < vertices.size(); ++i) {
					convexHullVertices[i] = Geometry::Vec3(
						static_cast<float>(vertices[i].x),
						static_cast<float>(vertices[i].y),
						static_cast<float>(vertices[i].z));
				}

                //  print the convex hull vertices and faces
                std::cout << "Convex hull vertices:" << std::endl;
                // for (const auto& vertex : convexHullVertices) {
                //     std::cout << vertex.x() << ", " << vertex.y() << ", " << vertex.z() << std::endl;
                // }

                std::cout << "Convex hull faces:" << std::endl;
                // for (int i = 0; i < nFaces; ++i) {
                //     std::cout << faceIndices[3 * i] << ", " << faceIndices[3 * i + 1] << ", " << faceIndices[3 * i + 2] << std::endl;
                // }

			// Save convex hull vertices and faces to OBJ file
				std::string filename = "convex_hull_" + std::to_string(smallId) + ".obj";
				std::ofstream output(filename, std::ios::out | std::ios::app);

				// Write vertices to the OBJ file
				for (const auto& vertex : convexHullVertices) {
					output << "v " << vertex.getX() << " " << vertex.getY() << " " << vertex.getZ() << std::endl;
				}

				// Write faces to the OBJ file
				for (int i = 0; i < nFaces; ++i) {
					output << "f " << (faceIndices[3 * i] + 1) << " " << (faceIndices[3 * i + 1] + 1) << " " << (faceIndices[3 * i + 2] + 1) << std::endl;
				}

				output.close();

                // Free allocated memory
                free(faceIndices);
			}
			}

		// Add the mesh information to each intersection
		std::vector<std::tuple<Graph*, Graph*, std::vector<Geometry::Vec3>>> res;
		for(size_t i = 0; i < intersections.size(); ++i) {
			res.push_back({smaller, bigger, intersections[i]});
		}

		spdlog::get("time_log")->info("Intersection: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));

		return res;
	}

	std::vector<std::tuple<Graph*, Graph*, std::vector<Geometry::Vec3>>>
	intersection(Graph* smaller, Graph* bigger) { //Original
		// Start timer for this function
		spdlog::stopwatch sw;

		std::cout << yellow << "Start intersection" << reset << std::endl;

		//TODO Octree here

		std::vector<std::vector<Geometry::Vec3>> intersections;

		// Easier access to the MeshAccessor objects
		MeshAccessor* smallerMeshAccessor =
			(*smaller)[*boost::vertices(*smaller).first].ma;
		MeshAccessor* biggerMeshAccessor =
			(*bigger)[*boost::vertices(*bigger).first].ma;

		// Put the ids of all polygons into smallerPolygons since all of these
		// will be checked
		std::vector<uint32_t> smallerPolygons;
		for(uint32_t id = 0; 3*id < smallerMeshAccessor->getIndexCount(); ++id) {
			smallerPolygons.push_back(id);
		}

		// Put the ids of all polygons which intersect with the AABB of smaller
		// into biggerPolygons since all of these will be checked
		std::vector<uint32_t> biggerPolygons =
			polygonInAABBIds(biggerMeshAccessor, smallerMeshAccessor->getAABB());

		//----------------------------------------------------------
		// TODO Adjust AABB of bigger!
		// Now we have two vectors of possibly intersecting polygons
		// TODO? I skip over the preceding discard of polygons and
		// apply the formula to all other polygons
		//----------------------------------------------------------

		// For every poylgon in smallerPolygons, check for intersection with
		// every polygon of biggerPolygons
		for(uint32_t smallId = 0; smallId < smallerPolygons.size(); ++smallId) {
			Polygon* smallerTri = smallerMeshAccessor->getPolygonPtr(smallId);

			// Check if the current polygon of smaller intersects with the AABB
			// of bigger
			if(!biggerMeshAccessor->getAABB().triInAABB(smallerTri)) {
				// This polygon cannot intersect -> skip
				continue;
			}

			// Intersection possible
			// Construct the plane defined by smallerTri
			Plane p(smallerTri->p1, smallerTri->p2, smallerTri->p3);

			// this list will contain all intersections between smallerTri and
			// all polygons in biggerPolygons
			std::vector<std::vector<Geometry::Vec3>> localIntersections;

			// Look at each polygon in biggerPolygons
			for(auto biggerId : biggerPolygons) {
				Polygon* biggerTri =
					biggerMeshAccessor->getPolygonPtr(biggerId);

				// Compute all 6 t values and save the results in tValues
				std::tuple<float, float, float, float, float, float> tValues =
					computetValues(p, biggerTri);

				// Categorize the indices of the computed values into categories
				// according to membership in specific intervals
				std::tuple<std::vector<short>,
						   std::vector<short>,
						   std::vector<short>,
						   std::vector<short>,
						   std::vector<short>> categorizedtValueIndices =
					categorizetValueIndices(tValues);

				// Look at the number of t values in each category and
				// decide if an intersection occurs
				// If an intersection occurs, decide which case it is
				std::vector<Geometry::Vec3> localIntersection;
				switch(determineIntersectionCase(categorizedtValueIndices)) {
					case 0: // No intersection -> Nothing needs to be added
							break;
					case 1: // Case 1: Calculate endpoints with the help of t
							localIntersection =
								handleIntersectionCase1(
									biggerTri,
									tValues,
									categorizedtValueIndices
								);
							break;
					case 2: // Case 2:
							localIntersection =
								handleIntersectionCase2(
									biggerTri,
									tValues,
									categorizedtValueIndices
								);
							break;
					case 3: // Case 3: Add the edge that never yields 1
							localIntersection =
								handleIntersectionCase3(
									biggerTri,
									categorizedtValueIndices
								);
							break;
					case 4: // Case 4: Add all vertices
							localIntersection =
								{biggerTri->p1, biggerTri->p2, biggerTri->p3};
							break;
				}

				// Check if a point of smaller lies inside bigger
				// Flatten all three ways and check if any vertex lies in all
				// resulting triangles
				// Drop x
				std::tuple<Geometry::Vec2, Geometry::Vec2, Geometry::Vec2>
					biggerFlatX = flattenTri(biggerTri, x);
				std::tuple<Line2, Line2, Line2>
					biggerLinesX =
						{Line2(std::get<0>(biggerFlatX),
							   std::get<1>(biggerFlatX)),
						 Line2(std::get<0>(biggerFlatX),
							   std::get<2>(biggerFlatX)),
						 Line2(std::get<1>(biggerFlatX),
							   std::get<2>(biggerFlatX))};
				Geometry::Vec2 centerX(
					(std::get<0>(biggerFlatX)[x] +
					 std::get<1>(biggerFlatX)[x] +
					 std::get<2>(biggerFlatX)[x])/3,
					(std::get<0>(biggerFlatX)[y] +
					 std::get<1>(biggerFlatX)[y] +
					 std::get<2>(biggerFlatX)[y])/3
				);
				std::tuple<short, short, short> relPosCenterX =
					determineRelativePos(centerX, biggerLinesX);

				// Drop y
				std::tuple<Geometry::Vec2, Geometry::Vec2, Geometry::Vec2>
					biggerFlatY = flattenTri(biggerTri, y);
				std::tuple<Line2, Line2, Line2>
					biggerLinesY =
						{Line2(std::get<0>(biggerFlatY),
							   std::get<1>(biggerFlatY)),
						 Line2(std::get<0>(biggerFlatY),
							   std::get<2>(biggerFlatY)),
						 Line2(std::get<1>(biggerFlatY),
							   std::get<2>(biggerFlatY))};
				Geometry::Vec2 centerY(
					(std::get<0>(biggerFlatY)[x] +
					 std::get<1>(biggerFlatY)[x] +
					 std::get<2>(biggerFlatY)[x])/3,
					(std::get<0>(biggerFlatY)[y] +
					 std::get<1>(biggerFlatY)[y] +
					 std::get<2>(biggerFlatY)[y])/3);
				std::tuple<short, short, short> relPosCenterY =
					determineRelativePos(centerY, biggerLinesY);

				// Drop z
				std::tuple<Geometry::Vec2, Geometry::Vec2, Geometry::Vec2>
					biggerFlatZ = flattenTri(biggerTri, z);
				std::tuple<Line2, Line2, Line2>
					biggerLinesZ =
						{Line2(std::get<0>(biggerFlatZ),
							   std::get<1>(biggerFlatZ)),
						 Line2(std::get<0>(biggerFlatZ),
							   std::get<2>(biggerFlatZ)),
						 Line2(std::get<1>(biggerFlatZ),
							   std::get<2>(biggerFlatZ))};
				Geometry::Vec2 centerZ(
					(std::get<0>(biggerFlatZ)[x] +
					 std::get<1>(biggerFlatZ)[x] +
					 std::get<2>(biggerFlatZ)[x])/3,
					(std::get<0>(biggerFlatZ)[y] +
					 std::get<1>(biggerFlatZ)[y] +
					 std::get<2>(biggerFlatZ)[y])/3);
				std::tuple<short, short, short> relPosCenterZ =
					determineRelativePos(centerZ, biggerLinesZ);

				for(short i = 0; i < 3; ++i) {
					Geometry::Vec3 currentVertex = smallerTri->getVertex(i);
					if(pointInsideTri(
							currentVertex, biggerLinesX, x, relPosCenterX) &&
					   pointInsideTri(
							currentVertex, biggerLinesY, y, relPosCenterY) &&
					   pointInsideTri(
							currentVertex, biggerLinesZ, z, relPosCenterZ)) {
						// Add that point to the intersection
						localIntersection.push_back(currentVertex);
					}
				}
				localIntersections.push_back(localIntersection);
			}

			// Cut all found intersections to the smallerTri
			std::vector<std::vector<Geometry::Vec3>> cutLocalIntersections =
				cutPlaneIntersectionsToPolygon(smallerTri, localIntersections);

			// Copy all cutLocalIntersections into intersections
			for(unsigned short i = 0; i < cutLocalIntersections.size(); ++i) {
				// Check if the current intersection is valid
				std::vector<Geometry::Vec3> currentCutLocalIntersection =
					cutLocalIntersections[i];
				if(currentCutLocalIntersection.size() != 0) {
					intersections.push_back(currentCutLocalIntersection);
				}
			}
		}

		// Add the mesh information to each intersection
		std::vector<std::tuple<Graph*, Graph*, std::vector<Geometry::Vec3>>> res;
		for(size_t i = 0; i < intersections.size(); ++i) {
			res.push_back({smaller, bigger, intersections[i]});
		}

		spdlog::get("time_log")->info("Intersection: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));

		return res;
	}


	
}
