/** 
 * @file
 *
 * @brief This file declares all functionality needed for computing
 *        intersections between graphs.
 *
 * @author Jan-Luca Hansel
 */

#ifndef INTERSECTION_H
#define INTERSECTION_H

#include <boost/optional/optional_io.hpp>

#include "Plane.h"
#include "Line2.h"
#include "Line3Norm.h"
#include "Line3Unnorm.h"

namespace Pathfinder {

	/**
	 * @brief Compute the t value for a plane and two points that form a line.
	 * @details The computation of the t value is decribed in the thesis. It
	 *          describes the distance of the intersection point of the plane
	 *          with the line formed by p1 and p2, from one of the endpoints.
	 *
	 * @param p  The plane.
	 * @param p1 The first endpoint of the line.
	 * @param p2 The second endpoint of the line.
	 *
	 * @return Return t according to the formula in the thesis.
	 */
	float 
	t(
		const Plane& p, 
		const Geometry::Vec3& p1, 
		const Geometry::Vec3& p2
	); 


	/**
	 * @brief Sort each computed t into a tuple of vectors where each vector 
	 *        corresponds to one specific interval.
	 * @details Each vector stores the indices of those values in tValues which 
	 *          fall into the interval represented by that vector. The intervals 
	 *          are:
	 *             0 = (-infinity,0), 
	 *             1 = [0,0], 
	 *             2 = (0,1), 
	 *             3 = [1,1], 
	 *             4 = (1,infinity).
	 *
	 * @param tValues A vector of 6 elements which contains the computed t 
	 *                values for one poylgon.
	 *
	 * @return Return a tuple containing 5 vectors that in turn contain short
	 *         values. 
	 */
	std::tuple<std::vector<short>, 
			   std::vector<short>, 
			   std::vector<short>, 
			   std::vector<short>, 
			   std::vector<short>> 
	categorizetValueIndices(
		const std::tuple<float, float, float, float, float, float>& tValues
	); 


	/**
	 * @brief Determine which intersection case should be considered for the 
	 *        computed t values.
	 * @details Given the computed t values for an intersection, a case 
	 *          distinction can be made. The details are provided in the thesis.
	 *
	 * @param categorizedtValueIndices A tuple of 5 vectors which shows how many 
	 *                                 t values fall into which interval.
	 *
	 * @return Return a number representing the found intersection case. 
	 * 0 = no intersection, 1 = (1), ..., 4 = (4).
	 */
	short 
	determineIntersectionCase(
		const std::tuple<std::vector<short>, 
			        	 std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>>& categorizedtValueIndices
	); 


	/**
	 * @brief Translate the index of the tValues vector to the endpoints which 
	 *        were involved in the t value computation.
	 * @details Every possible combination of two different vertices of a 
	 *          polygon has a certain index in the t values tuple. This 
	 *          function provides a convenient way to obtain the two endpoints
	 *          involved in the computation of a certain t value.
	 *
	 * @param id The index of the tValues vector.
	 *
	 * @return Return a tuple of two numbers. These numbers represent the 
	 *         endpoints used for computing the t value with the index id in the
	 *         tvalues tuple.
	 */
	std::tuple<short, short> 
	tIndex2endpoints(
		const short& id
	); 


	/**
	 * @brief Translate a pair of endpoint indices into the corresponding index
	 *        of the tValue vector.
	 * @details Every possible combination of two different vertices of a 
	 *          polygon has a certain index in the t values tuple. This 
	 *          function provides a convenient way to obtain the t value tuple 
	 *          index computed from the two given endpoints.
	 *
	 * @param e An edge represented by a vector of two endpoint indices.
	 *
	 * @return Return the index of the computed t value in the tuple tValues.
	 */
	short 
	endpoints2tIndex(
		const std::tuple<short, short>& e
	); 


	/**
	 * @brief For a given index of the tValues tuple, return the index of the 
	 *        tValues tuple which contains the t value that was comuputed by the
	 *        same vertices but in reverse order.
	 * @details Example: t value 0 and 2 are computed by the same two vertices
	 *          of a given poylgon. However, the order of vertices is switched
	 *          in the computations of the two t values.
	 *
	 * @param id The index of the tValues tuple.
	 *
	 * @return Return the index of the element in the tValues tuple which 
	 *         contains a value computed with the same endpoints in reverse 
	 *         order.
	 */
	short 
	getEquivalenttValueIndex(
		const short& id
	);


	/**
	 * @brief For a given tuple of found intersections, determine if these
	 *        are valid. 
	 * @details An intersection is valid if its x position lies
	 *          between the x positions of the two points pi and pj.
	 *
	 * @param pi            The first endpoint of the common intersection line.
	 * @param pj            The second endpoint of the common intersection line.
	 * @param intersections A tuple of three optional values representing
	 *                      the x positions of the intersections of the line 
	 *                      through pi and pj and the three bounding lines of 
	 *                      the polygon. If an intersection was found, a float 
	 *                      value will reside at the corresponding position. If
	 *                      two lines were found to be parallel, a boost::none 
	 *                      value will be placed at the corresponding index.
	 * @param tri           The triangle with which the intersections were 
	 *                      computed.
	 * @param centerAbove   A tulpe describing the relative position of the 
	 *                      center point of tri to its bounding lines.
	 *
	 * @return Return a tuple of bool values. If a value is set to true, this
	 *         means that the corresponding intersection in intersections is 
	 *         valid. If the value reads false, the intersection is invalid.
	 */
	std::tuple<bool, bool, bool> 
	intersectionTupleValid(
		Geometry::Vec2& pi, 
		Geometry::Vec2& pj, 
		const std::tuple<boost::optional<float>, 
						 boost::optional<float>, 
						 boost::optional<float>>& intersections, 
		const std::tuple<Line2, Line2, Line2>& tri, 
		const std::tuple<bool, bool, bool>& centerAbove
	); 


	/**
	 * @brief When moving a point on an intersection line, determine which 
	 *        intersection to use as a new endpoint. 
	 * @details This point should be the intersection which is valid and the 
	 *          closest to the point which shall be moved.
	 *
	 * @param tri           The current polygon described by a tuple of 
	 *                      bounding lines.
	 * @param intersections A tuple that contains the computed intersections of 
	 *                      an intersection line with tri.
	 * @param valid         A tuple that indicates which of the intersections 
	 *                      in intersections are valid.
	 * @param pjFlat        The flattened version of the point that is to be 
	 *                      moved.
	 *
	 * @return Return the index of the intersection which is closest to the 
	 *         point that is to be moved. If no such intersection exists, 
	 *         boost::none is returned.
	 */
	boost::optional<short> 
	chooseIntersection(
		const std::tuple<Line2, Line2, Line2>& tri, 
		const std::tuple<boost::optional<float>, 
				         boost::optional<float>, 
				         boost::optional<float>>& intersections, 
		const std::tuple<bool, bool, bool>& valid, 
		Geometry::Vec2& pjFlat
	); 


	/**
	 * @brief When calculating the point on a line that corresponds to a 
	 *        2D point, determine the distance this new point has from the 
	 *        support vector of the line.
	 *
	 * @param lij          The 3D line for which a distance from the support
	 *                     vector to a point shall be calculated.
	 * @param droppedCoord The component that was dropped while flattening in 
	 *                     the current step.
	 * @param pointX       The x position of the 2D point.
	 * @param pointY       The y position of the 2D point.
	 *
	 * @return Return the distance from the new point on the line to the support 
	 *         vector of the line.
	 */
	float 
	distanceToNewPoint(
		const Line3& lij, 
		const short& droppedCoord, 
		const float& pointX, 
		const float& pointY
	); 


	/**
	 * @brief For a given polygon and a pair of endpoints, determine a new 
	 *        endpoint for this intersection line which lies inside tri.
	 *
	 * @param pi           The first endpoint of the intersection line.
	 * @param pj           The second endpoint of the intersection line.
	 * @param droppedCoord The coordinate that was dropped while flattening
	 *                     the tri.
	 * @param inside       The vector which holds the information on which 
	 *                     points lie inside tri.
	 * @param tri          The current flattened polygon described by three
	 *                     bounding lines.
	 *
	 * @return Return the new endpoint for the intersection. If no such point 
	 *         could be found, boost::none is returned.
	 */
	boost::optional<Geometry::Vec3> 
	cutPointToPolygon(
		const Geometry::Vec3& pi, 
		const Geometry::Vec3& pj, 
		const short& droppedCoord, 
		const std::vector<bool>& inside, 
		const std::tuple<Line2, Line2, Line2>& tri
	); 


	/**
	 * @brief For a given polygon and a list of intersections with the plane 
	 *        defined by that polygon, edit these intersections to fit the size
	 *        of the actual polygon.
	 * @details This might involve discarding intersections or cutting them to 
	 *          the appropriate size.
	 *
	 * @param tri           A polygon defined by three vertices.
	 * @param intersections A vector of vectors which contain the endpoints of 
	 *                      the intersections with plane defined by tri.
	 *
	 * @return Return a vector of vectors containing the endpoints of the edited
	 *         intersections. If some intersection points needed to be 
	 *         discarded, they are returned as boost::none. The indices 
	 *         are unchanged from intersections.
	 */
	std::vector<std::vector<boost::optional<Geometry::Vec3>>>
	cutPlaneIntersectionsToPolygon(
		const std::vector<Geometry::Vec3>& tri, 
		const std::vector<std::vector<Geometry::Vec3>>& intersections
	); 


	/**
	 * @brief Given a plane and a polygon, compute the six possible t values.
	 *
	 * @param p   The plane.
	 * @param tri The polygon.
	 *
	 * @return Return a tuple containing all 6 possible t values. The mapping
	 * of tuple position and vertex pairing is the following:
	 *     0 -- 0, 1
	 *     1 -- 0, 2
	 *     2 -- 1, 0
	 *     3 -- 1, 2
	 *     4 -- 2, 0
	 *     5 -- 2, 1
	 */
	std::tuple<float, float, float, float, float, float> 
	computetValues(
		const Plane& p, 
		const Polygon* tri
	); 


	/**
	 * @brief Compute which polygons of the mesh managed by ma intersect with 
	 *        the AABB aabb.
	 *
	 * @param ma   The MeshAccessor of the mesh whose polygons shall be checked 
	 *             for intersection.
	 * @param aabb The AABB that shall be used for checking intersection.
	 *
	 * @return Return a vector containing all indices of those polygons of the 
	 *         mesh managed by ma which intersect with the AABB aabb.
	 */
	std::vector<uint32_t>
	polygonInAABBIds(
		MeshAccessor* ma, 
		AABB aabb
	); 


	/**
	 * @brief Do the necessary computations to obtain the intersection endpoints 
	 *        for intersection case 1.
	 *
	 * @param tri                      The polygon that intersects the current 
	 *                                 plane.
	 * @param tValues                  The computed t values for the 
	 *                                 intersection of tri and the current 
	 *                                 plane.
	 * @param categorizedtValueIndices The tuple containing the assignment of 
	 *                                 the indices in the tValues vector to the 
	 *                                 different value intervals.
	 *
	 * @return Return a vector containing the endpoints of the intersection.
	 */
	std::vector<Geometry::Vec3> 
	handleIntersectionCase1(
		Polygon* tri, 
		const std::tuple<float, float, float, float, float, float>& tValues, 
		const std::tuple<std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>>& categorizedtValueIndices
	); 


	/**
	 * @brief Do the necessary computations to obtain the intersection endpoints 
	 *        for intersection case 2. 
	 * @details This includes the computation of the endpoint that lies on the 
	 *          current plane as well as deciding if case 2i. or 2ii. is 
	 *          present.
	 *
	 * @param tri                      The polygon that intersects the current 
	 *                                 plane.
	 * @param tValues                  The computed t values for the 
	 *                                 intersection of tri and the current plane.
	 * @param categorizedtValueIndices The tuple containing the assignment of 
	 *                                 the indices in the tValues vector to the 
	 *                                 different value intervals.
	 *
	 * @return Returns a vector containing the endpoints of the intersection.
	 */
	std::vector<Geometry::Vec3>	
	handleIntersectionCase2(
		const Polygon* tri, 
		const std::tuple<float, float, float, float, float, float>& tValues, 
		const std::tuple<std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>>& categorizedtValueIndices
	); 


	/**
	 * @brief Do the necessary computations to obtain the additional 
	 *        intersection endpoint for intersection case 2ii.
	 *
	 * @param tri           The polygon that intersects the current plane.
	 * @param vertexOnPlane The index of the vertex which lies on the current 
	 *                      plane.
	 * @param tValues       The computed t values for the intersection of tri 
	 *                      and the current plane.
	 *
	 * @return Return the computed additional intersection endpoint.
	 */
	Geometry::Vec3 
	handleIntersectionCase2ii(
		const Polygon* tri, 
		const short& vertexOnPlane, 
		const std::tuple<float, float, float, float, float, float>& tValues
	); 


	/**
	 * @brief Do the necessary computations to obtain the intersection endpoints 
	 *        for intersection case 3.
	 *
	 * @param tri                      The polygon that intersects the current 
	 *                                 plane.
	 * @param categorizedtValueIndices The tuple containing the assignment of 
	 *                                 the indices in the tValues vector to the 
	 *                                 different value intervals.
	 *
	 * @return Return a vector containing the endpoints of the intersection.
	 */
	std::vector<Geometry::Vec3> 
	handleIntersectionCase3(
		Polygon* tri, 
		const std::tuple<std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>>& categorizedtValueIndices
	); 


	/**
	 * @brief Compute the intersection between the meshes.
	 *
	 * @param smaller The smaller mesh.
	 * @param bigger  The bigger mesh.
	 *
	 * @return Return a vector with all found intersections.
	 */
	std::vector<std::tuple<Graph*, Graph*, std::vector<Geometry::Vec3>>> 
	intersection(
		Graph* smaller,	
		Graph* bigger
	); 


}

#endif
