/** 
 * @file 
 *
 * @brief This file declares the struct Line2 that represents a two-dimensional
 *        line.
 *
 * @author Jan-Luca Hansel
 */

#ifndef LINE2_H
#define LINE2_H

#include <boost/optional.hpp>
#include <boost/optional/optional_io.hpp>

#include "Using.h"
#include "Geometry/Vec2.h"

namespace Pathfinder {

	/**
	 * @brief This struct represents a vertical 2D line or a line of the form 
	 *        y = mx + b.
	 */
	struct Line2 {

		/** 
		 * The slope of the line.
		 */
		boost::optional<float> m; 


		/** 
		 * The y-intersect of the line.
		 */
		boost::optional<float> b;


		/** 
		 * The x value of a vertical line.
		 */
		boost::optional<float> xpos;


		/** 
		 * Indicate if this line is vertical.
		 */
		bool vertical; 
	

		/**
		 * @brief A constructor that creates a non-vertical line from the two 
		 *        parameters that describe a 2D line.
		 *
		 * @param m The slope of the line.
		 * @param b The y-intercept of the line.
		 */
		Line2(
			const float& m, 
			const float& b
		);


		/**
		 * @brief A constructor that creates a line from two points.
		 * @details The necessary parameters are calculated from the two 
		 *          given points.
		 *
		 * @param u The first point.
		 * @param v The second point.
		 */
		Line2(
			Geometry::Vec2& u, 
			Geometry::Vec2& v
		); 


		/**
		 * @brief Evaluate the line function for a given x.
		 *
		 * @param x The point at where the line function shall be evaluated.
		 *
		 * @return Return the corresponding y value to the given input if the 
		 *         line is not vertical. Return boost::none if the line is 
		 *         vertical.
		 */
		boost::optional<float> 
		eval(
			const float& x
		);

	};


	/**
	 * @brief Compute the x value of the intersection of two lines.
	 *
	 * @param a The first line.
	 * @param b The second line.
	 *
	 * @return Return the x value of the intersection point of a and b if
	 *         Return boost::none is a and b are parallel.
	 */
	boost::optional<float> 
	line2Intersection(
		const Line2& a, 
		const Line2& b
	); 

}

#endif
