/**
 * @file
 *
 * @brief This file implements all functionality declared in Merge.h.
 *
 * @author Jan-Luca Hansel
 */

#include <memory>
#include <chrono>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include "BuildGraph.h"
#include "Merge.h"
#include "Intersection.h"
#include "Vec.h"
#include "Compare.h"

namespace Pathfinder {

	std::vector<std::pair<Vertex, Graph*>> 
	determineMergePoints(
		const Geometry::Vec3& p, 
		Graph* graphA, 
		Graph* graphB, 
		const Octree* octree,
		AABB& aabb
	) {
		auto vertexItsA = boost::vertices(*graphA);
		auto vertexItsB = boost::vertices(*graphB);
		std::pair<Vertex, Graph*> closestPointGraphA = {0, graphA};
		std::pair<Vertex, Graph*> closestPointGraphB = {0, graphB};
		for(auto vit = vertexItsA.first; vit != vertexItsA.second; ++vit) {
			if(!aabb.pointInAABB((*graphA)[*vit].pos) && 
			   distance3(p, (*graphA)[*vit].pos) < 
			   distance3(p, (*graphA)[closestPointGraphA.first].pos)) {
				closestPointGraphA.first = *vit;
			}
		}
		for(auto vit = vertexItsB.first; vit != vertexItsB.second; ++vit) {
			if(!aabb.pointInAABB((*graphB)[*vit].pos) &&
			   distance3(p, (*graphB)[*vit].pos) < 
			   distance3(p, (*graphB)[closestPointGraphB.first].pos)) {
				closestPointGraphB.first = *vit;
			}
		}
		return {closestPointGraphA, closestPointGraphB};
	}


	Graph* 
	merge(
		const std::vector<Graph*>& group, 
		Octree* octree
	) {
		// Start timer for this function
		spdlog::stopwatch sw;

		 if(group.size() == 1) {
		 	return group[0];
		 }

		std::cout << yellow << "Begin Merge\n" << reset;

		// Compute all pairwise intersections and save them in the 
		// intersections vector
		std::vector<
			std::vector<
				std::tuple<Graph*, Graph*, std::vector<Geometry::Vec3>>
			>
		> intersections;

		for(size_t i = 0; i < group.size() - 1; ++i) {
			for(size_t j = i + 1; j < group.size(); ++j) {
				AABB aabbi = 
					(*group[i])[(*boost::vertices(*group[i]).first)].ma->getAABB();
				AABB aabbj = 
					(*group[j])[(*boost::vertices(*group[j]).first)].ma->getAABB();
				if(AABBintersection(aabbi, aabbj)) {
					// Intersection possible
					Geometry::Vec3 diffi = 
						{aabbi.max[x] - aabbi.min[x], 
						 aabbi.max[y] - aabbi.min[y], 
						 aabbi.max[z] - aabbi.min[z]
					};
					Geometry::Vec3 diffj = 
						{aabbj.max[x] - aabbj.min[x], 
						 aabbj.max[y] - aabbj.min[y], 
						 aabbj.max[z] - aabbj.min[z]
					};
					if(compareSmallerVec3(diffi, diffj)) {
						intersections.push_back(
							intersection(group[i], group[j])
						);
					} else {
						intersections.push_back(
							intersection(group[j], group[i])
						);
					}
				}
			}
		}

		// Aggregate all graphs into one not yet connected graph
		Graph* graph = new Graph();
		// (Graph where point is taken from, Vertex descriptor in the source 
		//  graph, Vertex descriptor in the destination graph, Position)
		std::vector<
			std::tuple<Graph*, Vertex, Vertex, Geometry::Vec3>
		> groupToGraph; // Resuse for later? TODO

		for(size_t i = 0; i < group.size(); ++i) {
			auto vertexIts = boost::vertices(*group[i]);
			// Look at every vertex of the current graph
			for(auto vit = vertexIts.first; vit != vertexIts.second; ++vit) {
				// Check if the current position already exists in the graph
				// TODO combined notation for point appearing in more than one graph?
				Geometry::Vec3 pos = (*group[i])[*vit].pos;
				bool found = false;
				for(auto t : groupToGraph) {
					if(std::get<3>(t) == pos) {
						found = true;
						break;
					}
				}
				if(!found) {
					Vertex v = boost::add_vertex(*graph);
					(*graph)[v] = (*group[i])[*vit];
					groupToGraph.push_back({group[i], *vit, v, pos});
				}
			}
			
			// Look at each edge of the current graph
			auto edgeIts = boost::edges(*group[i]);
            // Ensure that the range is valid before entering the loop
            if (edgeIts.first != edgeIts.second) {
                for (auto eit = edgeIts.first; eit != edgeIts.second; ++eit) {
                    // Get endpoints of edge in group[it]
                    // Get vertex description in graph (LUT?)
                    // Add edge with vertex descriptors in graph
                    // Set GraphEdge as is in group[it]
                    Vertex v1 = boost::source(*eit, *group[i]);
                    Vertex v1graph = 0;
                    for (auto t: groupToGraph) {
                        // Ensure that eit is valid before dereferencing
                        //if (eit != nullptr) {
                            if (v1 == std::get<1>(t)) {
                                v1graph = std::get<2>(t);
                                break;
                            }
                       // }
                    }
                    Vertex v2 = boost::target(*eit, *group[i]);
                    Vertex v2graph = 0;
                    for (auto t: groupToGraph) {
                        if (v2 == std::get<1>(t)) {
                            v2graph = std::get<2>(t);
                            break;
                        }
                    }
                    std::pair<Edge, bool> e =
                            boost::add_edge(v1graph, v2graph, *graph);
                    (*graph)[e.first] = (*group[i])[*eit];
                }
            }
		}

		// Create new links for each intersection
		for(auto intersectionList : intersections) {
			// Get the AABB of the current intersection
			float minX = std::numeric_limits<float>::max();
			float minY = std::numeric_limits<float>::max();
			float minZ = std::numeric_limits<float>::max();
			float maxX = std::numeric_limits<float>::min();
			float maxY = std::numeric_limits<float>::min();
			float maxZ = std::numeric_limits<float>::min();

			// Check if the current intersection is a single point
			bool singlePoint = true;

			for(auto intersection : intersectionList) {
				if(std::get<2>(intersection).size() > 1) {
					singlePoint = false;
				}

				// Look at each intersection point
				for(auto p : std::get<2>(intersection)) {
					if(p[x] < minX) {
						minX = p[x];
					} 
					if(p[y] < minY) {
						minY = p[y];
					} 
					if(p[z] < minZ) {
						minZ = p[z];
					}
					if(p[x] > maxX) {
						maxX = p[x];
					} 
					if(p[y] > maxY) {
						maxY = p[y];
					} 
					if(p[z] > maxZ) {
						maxZ = p[z];
					}
				}
			}

			// If the AABB is flat and the intersection consists of more than 
			// a single point, make it a volume
			if(!singlePoint && compareEqualFloat(minX, maxX)) {
				minX -= 0.5;
				maxX += 0.5;
			}
			if(!singlePoint && compareEqualFloat(minY, maxY)) {
				minY -= 0.5;
				maxY += 0.5;
			}
			if(!singlePoint && compareEqualFloat(minZ, maxZ)) {
				minZ -= 0.5;
				maxZ += 0.5;
			}

			AABB aabb({minX, minY, minZ}, {maxX, maxY, maxZ});

			for(auto intersection : intersectionList) {
				// Get the two graphs for the current intersection
				Graph* graphA = std::get<0>(intersection);
				Graph* graphB = std::get<1>(intersection);

				// Look at each intersection point
				for(auto p : std::get<2>(intersection)) {

					std::vector<std::pair<Vertex, Graph*>> mergePoints = 
						determineMergePoints(p, graphA, graphB, octree, aabb);

					// Check if the intersection already lies inside graph 
					// and insert if not
					Vertex v;
					bool found = false;
					auto vertexIts = boost::vertices(*graph);
					for(auto vit = vertexIts.first; vit != vertexIts.second; ++vit) {
						if(p == (*graph)[*vit].pos) {
							found = true;
							v = *vit;
							break;
						}
					}

					if(!found) {
						v = boost::add_vertex(*graph);
					}

					// v should be a valid vertex descriptor now
					(*graph)[v].pos = p;
					(*graph)[v].ma = nullptr;
					
					// Add edges to neighbours
					// Find neighbours in graph
					std::vector<Vertex> mergePointsGraph;
					for(auto mergePoint : mergePoints) {
						for(auto vit = vertexIts.first; vit != vertexIts.second; ++vit) {
							if((*mergePoint.second)[mergePoint.first].pos == (*graph)[*vit].pos) {
								mergePointsGraph.push_back(*vit);
								break;
							}
						}
					}

					// Establish edges to all neighbours
					for(size_t i = 0; i < mergePointsGraph.size(); ++i) {
						boost::add_edge(
							v, 
							mergePointsGraph[i], 
							GraphEdge{
								distance3(
									(*graph)[v].pos, 
									(*graph)[mergePointsGraph[i]].pos
								), 
								determineDrivability(
									(*graph)[v].pos,
									(*graph)[mergePointsGraph[i]].pos
								),
                               determineEdgeWeight(
                                        (*graph)[v].pos,
                                        (*graph)[mergePointsGraph[i]].pos
                                ),
                                determineEdgeType(
                                        (*graph)[v].pos,
                                        (*graph)[mergePointsGraph[i]].pos
                                )
							}, 
							*graph
						);
					}

				}
			}


			// Delete all vertices inside the AABB
			auto vertexIts = boost::vertices(*graph);
			for(auto vit = vertexIts.first; vit != vertexIts.second; ++vit) {
				if(aabb.pointInAABB((*graph)[*vit].pos)) {
					boost::clear_vertex(*vit, *graph);
				}
			}
		}

		spdlog::get("time_log")->info("merge: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));

		return graph;
	}

}
