/**
 * @file
 *
 * @brief This file declares all functionality related to merging graphs.
 * 
 * @author Jan-Luca Hansel
 */

#ifndef MERGE_H
#define MERGE_H

#include <vector>

#include "Octree.h"

namespace Pathfinder {

	/**
	 * @brief Get the points to which a intersection point shall be connected.
	 * @details The point p shall be connected to the two nearest points 
	 *          in either graph that are outside the given AABB.
	 *
	 * @param p      The point that shall be connected.
	 * @param graphA The first graph.
	 * @param graphB The second graph.
	 * @param octree An octree that encompasses the whole scene.
	 * @param aabb   An AABB that defines a volume where no connections
	 *               shall be made.
	 *
	 * @return Returns the two points that shall be used for the new connections.
	 */
	std::vector<std::pair<Vertex, Graph*>> 
	determineMergePoints(
		const Geometry::Vec3& p,
		Graph* graphA,
		Graph* graphB,
		const Octree* octree,
		AABB& aabb
	);


	/**
	 * @brief Create one big graph out of the given group.
	 *
	 * @param group  The group whose graphs shall be merged.
	 * @param octree A pointer to the octree that stores all information of the 
	 *               scene.
	 *
	 * @return Return the merged graph.
	 */
	Graph* 
	merge(
		const std::vector<Graph*>& group, 
		Octree* octree
	); 

}

#endif
