/** 
 * @file
 *
 * @brief This file implements all functionality declared in MeshAccessor.h.
 *
 * @author Jan-Luca Hansel
 */

#include <limits>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include <Rendering/MeshUtils/MeshUtils.h>

#include "MeshAccessor.h"
#include "Using.h"

namespace Pathfinder {


	bool MeshAccessor::triIdValid(const uint32_t& id) {
		return 3 * id < meshIndexData.getIndexCount();
	}


	bool MeshAccessor::vertexIdValid(const uint32_t& id) {
		return id < this->getVertexCount();
	}


	Polygon* MeshAccessor::newPolygonPtr(const uint32_t& id) {
		if(this->triIdValid(id)) {
			std::tuple<uint32_t, uint32_t, uint32_t> vIds = 
				triAccessorRef->getIndices(id);
			return new Polygon(
				worldTransformationMatrix.transformPosition(
					connectivityAccessorRef->getVertex(std::get<0>(vIds))),
				worldTransformationMatrix.transformPosition(
					connectivityAccessorRef->getVertex(std::get<1>(vIds))),
				worldTransformationMatrix.transformPosition(
					connectivityAccessorRef->getVertex(std::get<2>(vIds))),
				geoNodePtr);
		} else {
			spdlog::get("warn_log")->info("newPolygonPtr returns nullptr.");
			return nullptr;
		}
	}


	MeshAccessor::MeshAccessor(MinSG::GeometryNode* geoNodePtr) : 
		geoNodePtr(geoNodePtr), 
		meshPtr(geoNodePtr->getMesh()), 
		meshVertexData(meshPtr->openVertexData()), 
		meshIndexData(meshPtr->openIndexData()),
		worldTransformationMatrix(geoNodePtr->getWorldTransformationMatrix()),
		connectivityAccessorRef(Rendering::MeshUtils::ConnectivityAccessor::create(meshPtr)),
		triAccessorRef(Rendering::MeshUtils::TriangleAccessor::create(meshPtr)),
		polygons(std::vector<Polygon*>()) {
		// Merge close vertices such that searching for adjacent vertices does 
		// not throw an exception
		// TODO Uncomment this line if an exception is thrown
		//Rendering::MeshUtils::mergeCloseVertices(meshPtr, 10);

		// Go through the mesh and find all max/min values
			float minX = std::numeric_limits<float>::max();
			float minY = std::numeric_limits<float>::max();
			float minZ = std::numeric_limits<float>::max();
			float maxX = std::numeric_limits<float>::min();
			float maxY = std::numeric_limits<float>::min();
			float maxZ = std::numeric_limits<float>::min();
			for(uint32_t id = 0; id < meshPtr->getVertexCount(); ++id) {
				Geometry::Vec3 pos = 
					worldTransformationMatrix.transformPosition(
					connectivityAccessorRef->getVertex(id));
				if(pos[x] < minX) {
					minX = pos[x];
				} else if(pos[x] > maxX) {
					maxX = pos[x];
				}
				if(pos[y] < minY) {
					minY = pos[y];
				} else if(pos[y] > maxY) {
					maxY = pos[y];
				}
				if(pos[z] < minZ) {
					minZ = pos[z];
				} else if(pos[z] > maxZ) {
					maxZ = pos[z];
				}
			}		
			aabb = AABB({minX, minY, minZ}, {maxX, maxY, maxZ});
		
		for(size_t i = 0; 3 * i < meshIndexData.getIndexCount(); ++i) {
			polygons.push_back(newPolygonPtr(i));
		}
	}


	void MeshAccessor::deletePolygons() {
		//std::cout << "Destructor\n";
		for(size_t i = 0; i < polygons.size(); ++i) {
			//std::cout << polygons[i] << " -> ";
			delete polygons[i];
			polygons[i] = nullptr;
			//std::cout << polygons[i] << std::endl;
		}
	}

	
	Polygon* MeshAccessor::getPolygonPtr(const uint32_t& id) {
		if(triIdValid(id)) {
			return polygons[id];
		}
		spdlog::get("warn_log")->info("getPolygonPtr returns nullptr.");
		return nullptr;
	}


	Geometry::Vec3 MeshAccessor::getVertex(const uint32_t& id) {
		if(vertexIdValid(id)) {
			return worldTransformationMatrix.transformPosition(
				connectivityAccessorRef->getVertex(id));
		}
		spdlog::get("warn_log")->info("getVertex returns \{}.");
		return {};
	}


	std::vector<uint32_t> MeshAccessor::getAdjacentVertexIds(
		const uint32_t& id
	) {
		if(vertexIdValid(id)) {
			return connectivityAccessorRef->getVertexAdjacentVertices(id);
		}
		spdlog::get("warn_log")->info("getAdjacentVertexIds returns \{}.");
		return {};
	}


	AABB MeshAccessor::getAABB() {
		return aabb;
	}


	uint32_t MeshAccessor::getVertexCount() {
		return meshPtr->getVertexCount();
	}


	uint32_t MeshAccessor::getIndexCount() {
		return meshIndexData.getIndexCount();
	}


	MinSG::GeometryNode* MeshAccessor::getGeoNodePtr() const {
		return geoNodePtr;
	}


	bool operator==(const MeshAccessor& l, const MeshAccessor& r) {
		return l.getGeoNodePtr() == r.getGeoNodePtr();
	}
}
