/**
 * @file
 *
 * @brief This file declares the struct MeshAccessor that presents a
 *        convenient way to access meshes. 
 *
 * @author Jan-Luca Hansel
 */

#ifndef MESHACCESSOR_H
#define MESHACCESSOR_H

#include <cstdint>

#include <Rendering/Mesh/Mesh.h>
#include <Rendering/Mesh/MeshIndexData.h>
#include <Rendering/Mesh/VertexAttributeAccessors.h>
#include <Rendering/MeshUtils/ConnectivityAccessor.h>
#include <Rendering/MeshUtils/TriangleAccessor.h>

#include "AABB.h"

namespace Pathfinder {

	/**
	 * @brief This class is used to simplify the accessing of meshes, their
	 *        vertices and polygons.
	 */
	class MeshAccessor {
	
		private:

			/** 
			 * A pointer to the GeometryNode which holds the mesh that is to be 
			 * managed.
			 */
			MinSG::GeometryNode* geoNodePtr; 


			/** 
			 * A pointer to the managed mesh.
			 */
			Rendering::Mesh* meshPtr; 


			/** 
			 * The VertexData which holds information about the vertices of the 
			 * mesh.
			 */
			Rendering::MeshVertexData& meshVertexData; 


			/**
			 * The IndexData which holds information about the indices of the 
			 * vertices of the mesh.
			 */
			Rendering::MeshIndexData& meshIndexData; 


			/** 
			 * The matrix to convert local to world coordinates.
			 */
			Geometry::Matrix4x4 worldTransformationMatrix; 


			/** 
			 * A reference to the ConnectivityAccessor which enables access to 
			 * the connectivity between vertices.
			 */
			Util::Reference<Rendering::MeshUtils::ConnectivityAccessor> 
				connectivityAccessorRef; 


			/** 
			 * A reference to the triangle accessor which enables polygon 
			 * access.
			 */
			Util::Reference<Rendering::MeshUtils::TriangleAccessor> 
				triAccessorRef; 


			/** 
			 * The AABB of the managed mesh.
			 */
			AABB aabb; 


			/** 
			 * A list of pointers to Polygon objects.
			 */
			std::vector<Polygon*> polygons; 


			/**
			 * @brief For a given Id, decide if this Id is a valid Id for 
			 *        accessing a polygon of this mesh.
			 * 
			 * @param id The Id which is to be checked.
			 *
			 * @return Return true if id is valid. Return false otherwise.
			 */
			bool 
			triIdValid(
				const uint32_t& id
			); 


			/**
			 * @brief For a given Id, decide if this Id is a valid Id for 
			 *        accessing a vertex of this mesh.
			 *
			 * @param id The Id which is to be checked.
			 *
			 * @return Return true if it is valid. Return false otherwise.
			 */
			bool 
			vertexIdValid(
				const uint32_t& id
			);


			/**
			 * @brief For a given Id, create a new Polygon object for the 
			 *        corresponding polygon in the managed mesh.
			 *
			 * @param id The polygon's Id in the managed mesh.
			 *
			 * @return Return a pointer to a polygon object that represents 
			 *         the queried polygon of the mesh.
			 */
			Polygon* 
			newPolygonPtr(
				const uint32_t& id
			); 


		public:

			/**
			 * @brief A constructor which initializes all member variables.
			 *
			 * @param geoNodePtr A pointer to a GeometryNode.
			 */
			MeshAccessor(
				MinSG::GeometryNode* geoNodePtr
			);


			/**
			 * @brief Delete all allocated polygon objects.
			 */
			void 
			deletePolygons(
			);

			
			/**
			 * @brief For a given Id, return the corresponding polygon in world 
			 *        coordinates.
			 *
			 * @param id The polygon's Id.
			 *
			 * @return Return a pointer to a polygon object that represents the 
			 *         queried polygon of the mesh.
			 */
			Polygon* 
			getPolygonPtr(
				const uint32_t& id
			); 


			/**
			 * @brief Return the positon of the vertex corresponding to the 
			 *        given Id. 
			 * @details The position is given in world coordinates.
			 * 
			 * @param id The index of the vertex.
			 * 
			 * @return Return the postion of the vertex with the Id id in world 
			 *         coordinates.
			 */
			Geometry::Vec3 
			getVertex(
				const uint32_t& id
			);

			
			/**
			 * @brief Return a list containing all ids of vertices adjacent to
			 *        the vertex id.
			 *
			 * @param id The id of the vertex whose neighbours shall be 
			 *           returned.
			 *
			 * @return Return a list of the ids of all adjacent vertices.
			 */
			std::vector<uint32_t> 
			getAdjacentVertexIds(
				const uint32_t& id
			); 


			/**
			 * @brief Return the aabb member variable.
			 *
			 * @return Return the contents of aabb.
			 */
			AABB 
			getAABB(
			);


			/**
			 * @brief Return the number of vertices in the managed mesh.
			 *
			 * @return return the number of vertices in the mesh.
			 */
			uint32_t 
			getVertexCount(
			);


			/**
			 * @brief Return the number of indices in the managed mesh.
			 *
			 * @return Return the number of indices in the mesh.
			 */
			uint32_t 
			getIndexCount(
			);


			/**
			 * @brief Return the pointer stored in the member variable 
			 *        geoNodePtr.
			 *
			 * @return Return the contents of geoNodePtr.
			 */
			MinSG::GeometryNode* 
			getGeoNodePtr(
			) const; 

	};


	/**
	 * @brief Overload operator==. Two MeshAccessor objects are considered the 
	 *        same if both objects describe the same GeometryNode.
	 *
	 * @param l The left operand.
	 * @param r The right operand.
	 *
	 * @return Return true if the the geoNodePtr member variables of both 
	 *         operands are the same. Return false otherwise.
	 */
	bool 
	operator==(
		const MeshAccessor& l, 
		const MeshAccessor& r
	);

}

#endif
