/**
 * @file
 *
 * @brief This file implements all functionality declared in OBB.h.
 *
 * @author vineet
 */

#include <iostream>
#include <memory>
#include <chrono>
#include <array>
#include <cmath>
#include <vector>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include "OBB.h"
#include "Using.h"
#include "Vec.h"

namespace Pathfinder {


    OBB::OBB(const Geometry::Vec3& center, const Geometry::Vec3& halfExtents, const Geometry::_Matrix3x3<float>& rotation)
           : center(center), halfExtents(halfExtents), rotation(rotation) {}

    Geometry::Vec3 OBB::getCenter() const {
        return center;
    }

    Geometry::Vec3 OBB::getHalfExtents() const {
        return halfExtents;
    }

    Geometry::_Matrix3x3<float> OBB::getRotation() const {
        return rotation;
    }


    OBB::OBB(const Geometry::Vec3& min, const Geometry::Vec3& max) {
        // Calculate center from OBB
        center = (min + max) * 0.5f;

        // Calculate extents from half dimensions of OBB
        halfSizes = (max - min) * 0.5f;

        // object data points
        std::vector<Geometry::Vec3> points = getPointsFromObject();

        // Calculate OBB axes
        float covariance[3][3];
        computeCovariance(points, covariance);

        Geometry::Vec3 eigenvalues;
        Geometry::Vec3 eigenvectors[3];
        eigenDecomposition(covariance, eigenvalues, eigenvectors);

        // Set OBB axes
        orientation[0] = eigenvectors[0];
        orientation[1] = eigenvectors[1];
        orientation[2] = eigenvectors[2];
    }

    bool OBB::triNotInOBB(const Polygon* p, const short& component) {
        // Transform triangle points to OBB space
        Geometry::Vec3 p1_transformed;
        p1_transformed[0] = dot(orientation[0], p->p1 - center);
        p1_transformed[1] = dot(orientation[1], p->p1 - center);
        p1_transformed[2] = dot(orientation[2], p->p1 - center);

        Geometry::Vec3 p2_transformed;
        p2_transformed[0] = dot(orientation[0], p->p2 - center);
        p2_transformed[1] = dot(orientation[1], p->p2 - center);
        p2_transformed[2] = dot(orientation[2], p->p2 - center);

        Geometry::Vec3 p3_transformed;
        p3_transformed[0] = dot(orientation[0], p->p3 - center);
        p3_transformed[1] = dot(orientation[1], p->p3 - center);
        p3_transformed[2] = dot(orientation[2], p->p3 - center);

        // Check if all points are outside the OBB extents along the given axis
        if ((std::abs(p1_transformed[component]) > halfSizes[component] &&
             std::abs(p2_transformed[component]) > halfSizes[component] &&
             std::abs(p3_transformed[component]) > halfSizes[component]) ||
            (std::abs(p1_transformed[component]) < -halfSizes[component] &&
             std::abs(p2_transformed[component]) < -halfSizes[component] &&
             std::abs(p3_transformed[component]) < -halfSizes[component])) {
            return true;
        }
        return false;
    }

    bool OBB::triInOBB(const Polygon* p) {
        if (triNotInOBB(p, 0) || triNotInOBB(p, 1) || triNotInOBB(p, 2)) {
            // Triangle cannot intersect
            return false;
        }
        return true;
    }

    bool OBB::pointInOBB(Geometry::Vec3 p) {
        // Transform point to OBB space
        Geometry::Vec3 p_transformed;
        p_transformed[0] = dot(orientation[0], p - center);
        p_transformed[1] = dot(orientation[1], p - center);
        p_transformed[2] = dot(orientation[2], p - center);

        // Check if point is within extents along all axes
        if (std::abs(p_transformed[0]) <= halfSizes[0] &&
            std::abs(p_transformed[1]) <= halfSizes[1] &&
            std::abs(p_transformed[2]) <= halfSizes[2]) {
            return true;
        }
        return false;
    }

    float OBB::projectOBB(const OBB& obb, const Geometry::Vec3& axis) {
        float projection = 0.0f;
        for (int i = 0; i < 3; ++i) {
            projection += obb.halfSizes[i] * std::abs(dot(obb.orientation[i], axis));
        }
        return projection;
    }

    bool OBB::isAxisSeparating(const OBB& a, const OBB& b, const Geometry::Vec3& axis) {
        float aProj = projectOBB(a, axis);
        float bProj = projectOBB(b, axis);
        float distance = std::abs(dot(axis, b.center - a.center));
        return distance <= (aProj + bProj);
    }



    bool OBB::OBBintersection(const OBB& a, const OBB& b) {
        // Axes to test
        std::array<Geometry::Vec3, 15> axes;
        for (int i = 0; i < 3; ++i) {
            axes[i] = a.orientation[i];
            axes[i + 3] = b.orientation[i];
        }
        int index = 6;
        for (int i = 0; i < 3; ++i) {
            for (int j = 0; j < 3; ++j) {
                axes[index++] = cross(a.orientation[i], b.orientation[j]);
            }
        }

        // Check each axis
        for (const auto& axis : axes) {
            if (!isAxisSeparating(a, b, axis)) {
                return false;
            }
        }
        return true;
    }



    float OBB::dot(const Geometry::Vec3& a, const Geometry::Vec3& b) {
        return a.x() * b.x() + a.y() * b.y() + a.z() * b.z();
    }

    Geometry::Vec3 OBB::cross(const Geometry::Vec3& a, const Geometry::Vec3& b) {
        return Geometry::Vec3(a.y() * b.z() - a.z() * b.y(),
                              a.z() * b.x() - a.x() * b.z(),
                              a.x() * b.y() - a.y() * b.x());
    }


    //OBBTree: A Hierarchical Structure for Rapid Interference Detection
    //S. Gottschalk
    void OBB::computeCovariance(const std::vector<Geometry::Vec3>& points, float covariance[3][3]) {
        // Initialize covariance matrix to zero
        for (int i = 0; i < 3; ++i) {
            for (int j = 0; j < 3; ++j) {
                covariance[i][j] = 0.0f;
            }
        }

        // Calculate mean of points
        Geometry::Vec3 mean(0.0f, 0.0f, 0.0f);
        for (const Geometry::Vec3& point : points) {
            mean += point;
        }
        mean /= static_cast<float>(points.size());

        // Compute covariance
        for (const Geometry::Vec3& point : points) {
            Geometry::Vec3 centered_point = point - mean;
            covariance[0][0] += centered_point.x() * centered_point.x();
            covariance[0][1] += centered_point.x() * centered_point.y();
            covariance[0][2] += centered_point.x() * centered_point.z();
            covariance[1][0] += centered_point.y() * centered_point.x();
            covariance[1][1] += centered_point.y() * centered_point.y();
            covariance[1][2] += centered_point.y() * centered_point.z();
            covariance[2][0] += centered_point.z() * centered_point.x();
            covariance[2][1] += centered_point.z() * centered_point.y();
            covariance[2][2] += centered_point.z() * centered_point.z();
        }

        // Normalize covariance
        float invSize = 1.0f / static_cast<float>(points.size() - 1);
        for (int i = 0; i < 3; ++i) {
            for (int j = 0; j < 3; ++j) {
                covariance[i][j] *= invSize;
            }
        }
    }


    //https://builtin.com/data-science/eigendecomposition
    void OBB::eigenDecomposition(const float covariance[3][3], Geometry::Vec3 eigenvalues, Geometry::Vec3 eigenvectors[3]) {
        // Choose a random initial guess for eigenvector (normalized)
        float initialGuess[3] = {1.0f, 0.0f, 0.0f};
        float norm = sqrtf(dot(initialGuess, initialGuess, 3));
        for (int i = 0; i < 3; ++i) {
            initialGuess[i] /= norm;
        }

        // Maximum iterations and convergence threshold
        const int maxIterations = 100;
        const float convergenceThreshold = 1e-6f;

        // Perform Power Method iterations
        float oldEigenvalue = 0.0f;
        float newEigenvalue;
        for (int i = 0; i < maxIterations; ++i) {
            // Matrix-vector multiplication (using covariance matrix)
            Geometry::_Vec3<float> tempVec;
            for (int j = 0; j < 3; ++j) {
                tempVec[j] = 0.0f;
                for (int k = 0; k < 3; ++k) {
                    tempVec[j] += covariance[j][k] * initialGuess[k];
                }
            }

            // Update eigenvector and eigenvalue
            newEigenvalue = dot(tempVec, initialGuess, 3);
            scaleVector(tempVec, 1.0f / newEigenvalue, 3);
            norm = sqrtf(dot(tempVec, tempVec, 3));
            for (int i = 0; i < 3; ++i) {
                initialGuess[i] = tempVec[i] / norm;
            }

            // Check for convergence
            if (std::abs(newEigenvalue - oldEigenvalue) < convergenceThreshold) {
                break;
            }
            oldEigenvalue = newEigenvalue;
        }

        // Copy results to output structures
        eigenvalues.x() = newEigenvalue;
        eigenvectors.x() = initialGuess[0];
        eigenvectors.y = initialGuess[1];
        eigenvectors.z = initialGuess[2];
    }

    std::vector<Geometry::Vec3> OBB::getPointsFromObject() {
        return std::vector<Geometry::Vec3>();
    }

}
