/** 
 * @file
 *
 * @brief This file implements all functionality declared in OBJ.h.
 *
 * @author Jan-Luca Hansel
 */

#include <fstream>
#include <memory>
#include <chrono>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#define CONVHULL_3D_ENABLE
#include "convhull_3d/convhull_3d.h"

#include "OBJ.h"
#include "Sort.h"
#include "Compare.h"

namespace Pathfinder {

	std::tuple<size_t, size_t, size_t> 
	getOBJIndices(const std::list<Geometry::Vec3>& vertices, 
				  const Geometry::Vec3& a, 
				  const Geometry::Vec3& b, 
				  const Geometry::Vec3& c) {
		// Start timer for this function
		spdlog::stopwatch sw;

		// Get position of the vertices in the list vertices
		size_t aId, bId, cId;
		size_t counter = 0;
		for(auto &v : vertices) {
			if(compareEqualVec3(a, v)) {
				aId = counter;
			} else if(compareEqualVec3(b, v)) {
				bId = counter;
			} else if(compareEqualVec3(c, v)) {
				cId = counter;
			}
			++counter;
		}
		spdlog::get("time_log")->info("getOBJIndices: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
		return std::tuple<size_t, size_t, size_t>(aId, bId, cId);
	}


	void writeOBJFromMesh(
		const MinSG::GeometryNode* geoNodePtr, 
		const std::string& path
	) {
		// Start timer for this function
		spdlog::stopwatch sw;

		std::cout << yellow << "Compute an OBJ file from a mesh\n" << reset;

		// Create the file
		std::ofstream output;
		output.open(path);

		// Process all data and write it to the file
		output << "# From mesh\n";
		// Setup vertex access
		Rendering::Mesh* geoNodeMeshPtr = geoNodePtr->getMesh();
		Util::Reference<Rendering::MeshUtils::TriangleAccessor> triAccessorPtr = 
			Rendering::MeshUtils::TriangleAccessor::create(geoNodeMeshPtr);
		Geometry::Matrix4x4 worldCoordTransMat = 
			geoNodePtr->getWorldTransformationMatrix();
		Rendering::MeshIndexData& geoNodeMeshIndexData = 
			geoNodeMeshPtr->openIndexData();
		Util::Reference<Rendering::MeshUtils::ConnectivityAccessor> meshCARef = 
			Rendering::MeshUtils::ConnectivityAccessor::create(geoNodeMeshPtr);
		// Save each vertex in a list
		std::list<Geometry::Vec3> vertices;
		for(uint32_t id = 0; id < geoNodeMeshPtr->getVertexCount(); ++id) {
			vertices.push_back(
				worldCoordTransMat.transformPosition(meshCARef->getVertex(id))
			);
		}
		// Sort the list and throw away all duplicates
		vertices.sort(compareSmallerVec3);
		vertices.unique();
		// Print all v commands
		for(auto &v : vertices) {
			output << "v " << v[x] << " " << v[y] << " " << v[z] << '\n';
		}
		// Look at each polygon and save the polygons in a list
		std::list<std::tuple<size_t, size_t, size_t>> polygons;
		for(uint32_t id = 0; 3*id < geoNodeMeshIndexData.getIndexCount(); ++id) {
			// Get polygon
			std::tuple<uint32_t, uint32_t, uint32_t> tri = 
				triAccessorPtr->getIndices(id);
			polygons.push_back(
				getOBJIndices(vertices, 
							  worldCoordTransMat.transformPosition(
									meshCARef->getVertex(std::get<0>(tri))), 
							  worldCoordTransMat.transformPosition(
									meshCARef->getVertex(std::get<1>(tri))), 
							  worldCoordTransMat.transformPosition(
									meshCARef->getVertex(std::get<2>(tri)))
				)
			);
		}
		// Sort the list 
		for(auto &p : polygons) {
			sortSizeTTuple(p);
		}
		polygons.sort();
		// Print all f commands
		for(auto &p : polygons) {
			output << "f " << (std::get<0>(p)+1) 
				   << " " << (std::get<1>(p)+1)  
				   << " " << (std::get<2>(p)+1) 
				   << '\n';
		}

		output.close();
		spdlog::get("time_log")->info("writeOBJFromMesh: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
	}


	void writeOBJFromGraph(const Graph* graph, const std::string& path) {
		// Start timer for this function
		spdlog::stopwatch sw;

		std::cout << yellow << "Compute an OBJ file from a graph\n" << reset;

		// Create the file
		std::ofstream output;
		output.open(path);

		// Process the data and write it to the file
		output << "# From graph\n";
		std::list<Geometry::Vec3> vertices;
		// Save all vertices in a list 
		auto vertexIts = boost::vertices(*graph);
		for(auto it = vertexIts.first; it != vertexIts.second; ++it) {
			vertices.push_back((*graph)[*it].pos);
		}
		// Sort the list
		vertices.sort(compareSmallerVec3);
		// Print all v commands
		for(auto &v : vertices) {
			output << "v " << v[x] << " " << v[y] << " " << v[z] << '\n'; 
		}
		// Save all edges
		std::list<std::tuple<Geometry::Vec3, Geometry::Vec3>> edges;
		auto edgeIterators = boost::edges(*graph);
		for(auto it = edgeIterators.first; it != edgeIterators.second; ++it) {
			edges.push_back(
				{(*graph)[boost::source(*it, *graph)].pos, 
				 (*graph)[boost::target(*it, *graph)].pos}
			);
		}
		// Eliminate all duplicates
		for(auto &e : edges) {
			sortVec3Tuple(e);
		}
		edges.sort(compareSmallerVec3Tuple);
		edges.unique(compareEqualVec3Tuple);
		// Find all edges which have a common vertex
		std::list<std::tuple<std::tuple<Geometry::Vec3, Geometry::Vec3>, 
							 std::tuple<Geometry::Vec3, Geometry::Vec3>>> 
			edgePairsOneVertexEqual;
		for(auto &edge1 : edges) {
			for(auto &edge2 : edges) {
				if(compareEqualVec3Tuple(edge1, edge2)) {
					continue;
				} else if(
					compareEqualVec3(std::get<0>(edge1), std::get<0>(edge2)) ||
					compareEqualVec3(std::get<0>(edge1), std::get<1>(edge2)) ||
					compareEqualVec3(std::get<1>(edge1), std::get<0>(edge2)) ||
					compareEqualVec3(std::get<1>(edge1), std::get<1>(edge2))) {
					edgePairsOneVertexEqual.push_back({edge1, edge2});
				} 
			}
		}
		// Eliminate all duplicates
		for(auto &e : edgePairsOneVertexEqual) {
			sortVec3TupleTuple(e);
		}
		edgePairsOneVertexEqual.sort(compareSmallerVec3TupleTuple);
		edgePairsOneVertexEqual.unique(compareEqualVec3TupleTuple);
		// Go through all edgePairs and all edges and check if an edge forms a 
		// triangle togther with an edgePair
		std::list<std::tuple<size_t, size_t, size_t>> polygons;
		for(auto &edgePair : edgePairsOneVertexEqual) {
			for(auto &edge : edges) {
				// Check if edge is contained in edgePair
				if(compareEqualVec3Tuple(edge, std::get<0>(edgePair)) || 
				   compareEqualVec3Tuple(edge, std::get<1>(edgePair))) {
					continue;
				}
				// Check which vertex is the edge shared by the two edges of the edgePair
				if(compareEqualVec3(std::get<0>(std::get<0>(edgePair)), 
									std::get<0>(std::get<1>(edgePair)))
				) {
					// Check if edge completes the edgePair to a triangle
					if((compareEqualVec3(std::get<0>(edge), 
										 std::get<1>(std::get<0>(edgePair))) && 
					    compareEqualVec3(std::get<1>(edge), 
										 std::get<1>(std::get<1>(edgePair)))) ||
					   (compareEqualVec3(std::get<1>(edge), 
										 std::get<1>(std::get<0>(edgePair))) && 
					    compareEqualVec3(std::get<0>(edge), 
										 std::get<1>(std::get<1>(edgePair))))) {
						polygons.push_back(
							getOBJIndices(vertices, 
										  std::get<0>(std::get<0>(edgePair)), 
										  std::get<0>(edge), 
										  std::get<1>(edge)
							)
						);
					}
				}
				if(compareEqualVec3(std::get<0>(std::get<0>(edgePair)), 
									std::get<1>(std::get<1>(edgePair)))) {
					// Check if edge completes the edgePair to a triangle
					if((compareEqualVec3(std::get<0>(edge), 
										 std::get<1>(std::get<0>(edgePair))) && 
					    compareEqualVec3(std::get<1>(edge), 
										 std::get<0>(std::get<1>(edgePair)))) ||
					   (compareEqualVec3(std::get<1>(edge), 
										 std::get<1>(std::get<0>(edgePair))) && 
					    compareEqualVec3(std::get<0>(edge), 
										 std::get<0>(std::get<1>(edgePair))))) {
						polygons.push_back(
							getOBJIndices(vertices, 
										  std::get<0>(std::get<0>(edgePair)), 
										  std::get<0>(edge), 
										  std::get<1>(edge)
							)
						);
					}
				}
				if(compareEqualVec3(std::get<1>(std::get<0>(edgePair)), 
									std::get<0>(std::get<1>(edgePair)))) {
					// Check if edge completes the edgePair to a triangle
					if((compareEqualVec3(std::get<0>(edge), 
										 std::get<0>(std::get<0>(edgePair))) && 
					    compareEqualVec3(std::get<1>(edge), 
										 std::get<1>(std::get<1>(edgePair)))) ||
					   (compareEqualVec3(std::get<1>(edge), 
										 std::get<0>(std::get<0>(edgePair))) && 
					    compareEqualVec3(std::get<0>(edge), 
										 std::get<1>(std::get<1>(edgePair))))) {			
						polygons.push_back(
							getOBJIndices(vertices, 
										  std::get<1>(std::get<0>(edgePair)), 
										  std::get<0>(edge), 
										  std::get<1>(edge)
							)
						);
					}
				}
				if(compareEqualVec3(std::get<1>(std::get<0>(edgePair)), 
									std::get<1>(std::get<1>(edgePair)))) {
					// Check if edge completes the edgePair to a triangle
					if((compareEqualVec3(std::get<0>(edge), 
										 std::get<0>(std::get<0>(edgePair))) && 
						compareEqualVec3(std::get<1>(edge), 
										 std::get<0>(std::get<1>(edgePair)))) ||
					   (compareEqualVec3(std::get<1>(edge), 
										 std::get<0>(std::get<0>(edgePair))) && 
						compareEqualVec3(std::get<0>(edge), 
										 std::get<0>(std::get<1>(edgePair))))) {
						polygons.push_back(
							getOBJIndices(vertices, 
										  std::get<1>(std::get<0>(edgePair)), 
										  std::get<0>(edge), 
										  std::get<1>(edge)
							)
						);
					}
				}
			}
		}
		// Remove all duplicates
		for(auto &p : polygons) {
			sortSizeTTuple(p);
		}
		polygons.sort();
		polygons.unique();
		// print the f commands to define the faces
		for(auto &tri : polygons) {
			output << "f " << (std::get<0>(tri)+1) 
				   << " " << (std::get<1>(tri)+1) 
				   << " " << (std::get<2>(tri)+1)
				   << '\n';
		}

		output.close();
		spdlog::get("time_log")->info("writeOBJFromGraph: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
	}


	void appendPointcloudOBJFile(
		const std::vector<Geometry::Vec3>& points, 
		const std::string& path
	) {
		// Start timer for this function
		spdlog::stopwatch sw;

		std::ofstream output;
		output.open(path, std::ofstream::out | std::ofstream::app);
		for(auto p : points) {
			output << "v " << p << '\n';
		}
		output.close();
		spdlog::get("time_log")->info("appendPointCloudOBJFile: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
	}


	void writeIntersectionOBJFiles(
		const std::vector<
				  std::tuple<Graph*, Graph*, std::vector<Geometry::Vec3>>
			  >& intersections, 
		const std::string& path
	) {
		// Start timer for this function
		spdlog::stopwatch sw;

		std::cout << yellow << "Compute the OBJ file for an intersection\n" 
				  << reset;

		// There is nothing to do when no intersections have been found
		if(intersections.size() == 0) {
			return;
		}
  
		// filename = AddressOfFirstGeoNode_AddressOfSecondGeoNode.obj
		std::stringstream filenameStream;
		filenameStream << std::get<0>(intersections[0]) << "_" 
					   << std::get<0>(intersections[0]);
		std::string filename = filenameStream.str();

		// Write pointcloud to an OBJ file
		for(size_t i = 0; i < intersections.size(); ++i) {
			// Get the current intersection
			std::tuple<Graph*, Graph*, std::vector<Geometry::Vec3>> 
				currentIntersection = intersections[i];
			std::vector<Geometry::Vec3> points = 
				std::get<2>(currentIntersection);

			appendPointcloudOBJFile(
				points, path + "pointclouds/" + filename + ".obj"
			);
		}

		// Write convexHull of pointCloud to an OBJ file
		// Determine the path of the input file
		std::string inputOBJStr = path + "pointclouds/" + filename;
		char* inputOBJ = (char*) inputOBJStr.c_str();

		// The following code is taken from 
		// https://github.com/leomccormack/convhull_3d
		ch_vertex* vertices = NULL;
		int nVertices;
		extractVerticesFromObjFile(inputOBJ, &vertices, &nVertices);

		// Create the output path
		std::string outputOBJStr = path + "convexHulls/" + filename;
		char* outputOBJ = (char*) outputOBJStr.c_str();

		int* faceIndices = NULL;
		int nFaces;
		convhull_3d_build(vertices, nVertices, &faceIndices, &nFaces);
		convhull_3d_export_obj(
			vertices, nVertices, faceIndices, nFaces, 1, outputOBJ
		);

		free(vertices);
		free(faceIndices);

		spdlog::get("time_log")->info("writeIntersectionOBJFiles: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
	}

}
