/** 
 * @file
 *
 * @brief This file declares all functionality related to creating OBJ files.
 *
 * @author Jan-Luca Hansel
 */

#ifndef OBJ_H
#define OBJ_H

#include "Using.h"
#include "MeshAccessor.h"

namespace Pathfinder {

	/**
	 * @brief Given three points, get their names for an OBJ file.
	 * @details Search through the vertex list and return the indices of the 
	 *          
	 * @param vertices A list of all vertices.
	 * @param a        The first point.
	 * @param b        The second point.
	 * @param c        The third point.
	 *
	 * @return Returns a tuple that contains the names of the three points.
	 */
	std::tuple<size_t, size_t, size_t> 
	getOBJIndices(
		const std::list<Geometry::Vec3>& vertices, 
		const Geometry::Vec3& a, 
		const Geometry::Vec3& b, 
		const Geometry::Vec3& c
	); 


	/**
	 * @brief Go through the mesh associated with geoNodePtr and print 
	 *        the appropriate OBJ commands
	 *
	 * @param geoNodePtr A pointer to the GeometryNode.
	 * @param path       The path to the file where the data shall be written to.
	 */
	void 
	writeOBJFromMesh(
		const MinSG::GeometryNode* geoNodePtr, 
		const std::string& path
	);


	/**
	 * @brief Go through graph and output appropriate OBJ commands.
	 *
	 * @param graph The graph which shall be transformed.
	 * @param path  The path to the file where the data shall be written to.
	 */
	void 
	writeOBJFromGraph(
		const Graph* graph, 
		const std::string& path
	);


	/**
	 * @brief Given a list of points, append all of these points onto the file 
	 *        at location path.
	 *
	 * @param points A vector which contains the points that shall be written to 
	 *               a file.
	 * @param path   The path to OBJ file.
	 */
	void 
	appendPointcloudOBJFile(
		const std::vector<Geometry::Vec3>& points, 
		const std::string& path
	); 


	/**
	 * @brief Given a list of intersections, write all convex hulls of the 
	 *        intersections to OBJ files at the specified path.
	 *
	 * @param intersections A vector containing all intersections for a group.
	 * @param path          The path to the directory where the OBJ files shall 
	 *                      be written to.
	 */
	void 
	writeIntersectionOBJFiles(
		const std::vector<std::tuple<Graph*, Graph*, std::vector<Geometry::Vec3>>>&
			intersections, 
		const std::string& path
	); 

}

#endif
