/**
 * @file
 *
 * @brief This file implements all functionality declared in Octree.h.
 * 
 * @author Jan-Luca Hansel
 */

#include <iostream>
#include <limits>
#include <memory>
#include <chrono>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include <Geometry/Vec3.h>

#include "Octree.h"

namespace Pathfinder {

	Octree::Octree(const Geometry::Vec3& min, const Geometry::Vec3& max) : 
		min(min),
		max(max), 
		child{nullptr, nullptr, nullptr, nullptr, 
			  nullptr, nullptr, nullptr, nullptr}, 
		polygons(std::vector<Polygon*>()), 
		graphVertices(std::vector<std::pair<Vertex, Graph*>>()) {}


	Octree::Octree(
		const float& size, 
		const std::vector<Polygon*>& polygons, 
		const std::vector<std::pair<Vertex, Graph*>>& graphVertices
	) : 
		min(Geometry::Vec3(-size, -size, -size)), 
		max(Geometry::Vec3(size, size, size)), 
		child{nullptr, nullptr, nullptr, nullptr, 
			  nullptr, nullptr, nullptr, nullptr}, 
		polygons(polygons), 
		graphVertices(graphVertices) 
	{
		this->buildOctree(polygons, graphVertices, "");
	}


	void Octree::deleteOctree(const bool& root) {
		// Start timer for this function
		spdlog::stopwatch sw;

		// Only delete polygons in the root node
		if(root) {
			// Delete all polygons 
			for(auto p : polygons) {
				delete p;
				p = nullptr;
			}
		}

		// Delete all children
		for(short i = 0; i < 8; ++i) {
			if(child[i] != nullptr) {
				child[i]->deleteOctree(false);
				delete child[i];
				child[i] = nullptr;
			}
		}
	
		spdlog::get("time_log")->info("deleteOctree: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
	}


	void Octree::buildOctree(
		const std::vector<Polygon*>& polygonList, 
		const std::vector<std::pair<Vertex, Graph*>>& graphVertices, 
		const std::string& indent
	) {
		// Start timer for this function
		spdlog::stopwatch sw;

		// Set the polygons and nodes for this octree
		this->polygons = polygonList;
		this->graphVertices = graphVertices;

		// Check if theshold has been reached
		if(polygons.size() > threshold) {
			// Create empty octrees and store them in child
			Geometry::Vec3 center = 
				Geometry::Vec3(
					(this->min[x] + this->max[x]) / 2, 
					(this->min[y] + this->max[y]) / 2, 
					(this->min[z] + this->max[z]) / 2
				);

			child[0] = 
				new Octree(this->min, center);
			child[1] = 
				new Octree(
					Geometry::Vec3(this->min[x], this->min[y], center[z]), 
					Geometry::Vec3(center[x], center[y], this->max[z])
				);
			child[2] = 
				new Octree(
					Geometry::Vec3(this->min[x], center[y], this->min[z]), 
					Geometry::Vec3(center[x], this->max[y], center[z])
				);
			child[3] = 
				new Octree(
					Geometry::Vec3(this->min[x], center[y], center[z]), 
					Geometry::Vec3(center[x], this->max[y], this->max[z])
				);
			child[4] = 
				new Octree(
					Geometry::Vec3(center[x], this->min[y], this->min[z]), 
					Geometry::Vec3(this->max[x], center[y], center[z])
				);
			child[5] = 
				new Octree(
					Geometry::Vec3(center[x], this->min[y], center[z]), 
					Geometry::Vec3(this->max[x], center[y], this->max[z])
				);
			child[6] = 
				new Octree(
					Geometry::Vec3(center[x], center[y], this->min[z]), 
					Geometry::Vec3(this->max[x], this->max[y], center[z])
				);
			child[7] = 
				new Octree(center, this->max);

			// For each child
			for(short i = 0; i < 8; ++i) {
				std::vector<Polygon*> polygonsInChild;
				std::vector<std::pair<Vertex, Graph*>> graphVerticesInChild;

				// Build the list of Polygons for that child
				for(auto p : polygons) {
					if(child[i]->polygonInOctree(p, indent)) {
						polygonsInChild.push_back(p);
					}
				}
				// Build the list of GraphNodes for that child
				for(auto p : graphVertices) {
					if(child[i]->pointInOctree((*(p.second))[p.first].pos, 
											   indent)
					) {
						graphVerticesInChild.push_back(p);
					}
				}

				// Recursively call BuildOctree for that child
				child[i]->buildOctree(
					polygonsInChild, graphVerticesInChild, indent + "\t"
				);
			}
		}
		spdlog::get("time_log")->info("buildOctree: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
	}


	bool Octree::polygonInOctree(
		const Polygon* polygon, 
		const std::string& indent
	) {
		if(min[x] <= polygon->max[x] &&		// i.minX <= j.maxX
		   max[x] >= polygon->min[x] &&		// i.maxX >= j.minX
		   min[y] <= polygon->max[y] &&		// i.minY <= j.maxY
		   max[y] >= polygon->min[y] &&		// i.maxY >= j.minY
		   min[z] <= polygon->max[z] &&		// i.minZ <= j.maxZ
		   max[z] >= polygon->min[z]) {		// i.maxZ >= j.minZ
			return true;
		} else {
			return false;
		}
	}


	bool Octree::pointInOctree(
		const Geometry::Vec3& p, 
		const std::string& indent
	) {
		if(min[x] <= p[x] &&	// min <= node
		   min[y] <= p[y] &&
		   min[z] <= p[z] &&
		   p[x] <= max[x] &&	// node <= max
		   p[y] <= max[y] &&
		   p[z] <= max[z]) {
			return true;
		} else {
			return false;
		}
	}


	std::vector<std::pair<Vertex, Graph*>> Octree::getNeighbours(
		const Geometry::Vec3& p, const std::string& indent
	) {
		if(child[0] != nullptr && child[0]->pointInOctree(p, indent)) {
			//std::cout << indent << "\tCall getNeighbour on child 0\n";
			return child[0]->getNeighbours(p, indent + "\t");
		} else if(child[1] != nullptr && child[1]->pointInOctree(p, indent)) {
			//std::cout << indent << "\tCall getNeighbour on child 1\n";
			return child[1]->getNeighbours(p, indent + "\t");
		} else if(child[2] != nullptr && child[2]->pointInOctree(p, indent)) {
			//std::cout << indent << "\tCall getNeighbour on child 2\n";
			return child[2]->getNeighbours(p, indent + "\t");
		} else if(child[3] != nullptr && child[3]->pointInOctree(p, indent)) {
			//std::cout << indent << "\tCall getNeighbour on child 3\n";
			return child[3]->getNeighbours(p, indent + "\t");
		} else if(child[4] != nullptr && child[4]->pointInOctree(p, indent)) {
			//std::cout << indent << "\tCall getNeighbour on child 4\n";
			return child[4]->getNeighbours(p, indent + "\t");
		} else if(child[5] != nullptr && child[5]->pointInOctree(p, indent)) {
			//std::cout << indent << "\tCall getNeighbour on child 5\n";
			return child[5]->getNeighbours(p, indent + "\t");
		} else if(child[6] != nullptr && child[6]->pointInOctree(p, indent)) {
			//std::cout << indent << "\tCall getNeighbour on child 6\n";
			return child[6]->getNeighbours(p, indent + "\t");
		} else if(child[7] != nullptr && child[7]->pointInOctree(p, indent)) {
			//std::cout << indent << "\tCall getNeighbour on child 7\n";
			return child[7]->getNeighbours(p, indent + "\t");
		} else {
			//std::cout << indent << "\tReturn graphVertices.\n";
			return graphVertices;
		}
	}


	Octree* createOctree(const std::vector<Graph*>& graphList) {
		// Start timer for this function
		spdlog::stopwatch sw;

		std::vector<Polygon*> polygons; 
		std::vector<std::pair<Vertex, Graph*>> graphVertices;
		float size = std::numeric_limits<float>::min();

		// Go through each graph
		for(Graph* graph : graphList) {
			// Get the first vertex of graph to obtain a MeshAccessor
			Vertex v = *boost::vertices(*graph).first;
			MeshAccessor* ma = (*graph)[v].ma;

			// Create polygons and add them to polygons
			for(size_t i = 0; 3 * i < ma->getIndexCount(); ++i) {
				// Create a new Polygon and add it to polygons.
				polygons.push_back(ma->getPolygonPtr(i));
			}

			// Add graph vertices to graphVertices
			auto iterators = boost::vertices(*graph);
			for(auto it = iterators.first; it != iterators.second; ++it) {
				graphVertices.push_back({*it, graph});
				// Adjust size if this vertex has a bigger absolute value in 
				// any of it's position's components than size
				Geometry::Vec3 currentPos = (*graph)[*it].pos;
				for(short i = x; i <= z; ++i) {
					if(std::abs(currentPos[i]) > size) {
						size = std::abs(currentPos[i]);
					}
				}
			}
		}
		Octree* res = new Octree(size, polygons, graphVertices);
		spdlog::get("time_log")->info("createOctree: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
		return res;
	}


	void printOctree(const Octree* octree, const std::string& indent) {
		// Start timer for this function
		spdlog::stopwatch sw;

		std::cout << indent << "\\--+\n";
		std::cout << indent << "   |- min: " << octree->min << '\n';
		std::cout << indent << "   |- max: " << octree->max << '\n';
		std::cout << indent << "   |- Polygons: "
							<< octree->polygons.size() << '\n';
		for(auto p : octree->polygons) {
			std::cout << indent << "   |    |- " 
								<< p->p1 << "; " << p->p2 << "; " << p->p3 
								<< '\n';
		}
		std::cout << indent << "   |- Vertices: " 
							<< octree->graphVertices.size() << '\n';
		for(auto p : octree->graphVertices) {
			std::cout << indent << "   |    |- " 
								<< (*p.second)[p.first].pos << " from " 
								<< p.second << '\n';
		}
		for(short i = 0; i < 8; ++i) {
			if(octree->child[i] == nullptr) {
				std::cout << indent << "   |\\-- Null" << '\n';
			} else {
				printOctree(octree->child[i], indent + "   |");
			}
		}
		spdlog::get("time_log")->info("printOctree: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
	}

}
