/**
 * @file
 *
 * @brief This file declares the struct Octree that implements the octree
 *        data structure.
 * 
 * @author Jan-Luca Hansel
 */

#ifndef OCTREE_H
#define OCTREE_H

#include "Using.h"

namespace Pathfinder {

	/** 
	 * If this threshold is reached, stop octree recursion.
	 */
	const short threshold = 10; 


	/**
	 * @brief This struct implements the octree data structure.
	 * @details Each octree stores a list of polygons and a list of vertices and
	 *          is subdivided as follows:
	 *
	 *             6 7   y x
	 *         2 3 4 5   |/
	 *         0 1	     o--z
	 */
	struct Octree {

		/** 
		 * The AABBmin of the current cube.
		 */
		Geometry::Vec3 min; 


		/** 
		 * The AABBmax of the current cube.
		 */
		Geometry::Vec3 max; 


		/** 
		 * Store pointers to the eight children of the octree.
		 */
		std::vector<Octree*> child; 


		/** 
		 * Store pointers to Polygons that lie in this octree.
		 */
		std::vector<Polygon*> polygons; 


		/** 
		 * Store graph vertices that lie in this octree together with a 
		 * reference to their graph.
		 */
		std::vector<std::pair<Vertex, Graph*>> graphVertices; 


		/**
		 * @brief A basic constructor that initializes all data members with 
		 *        dummy data.
		 *
		 * @param AABBmin The smallest point of the octree.
		 * @param AABBmax The biggest point of the octree.
		 */
		Octree(
			const Geometry::Vec3& AABBmin, 
			const Geometry::Vec3& AABBmax
		);


		/**
		 * @brief A constructor that creates an octree and fills it with data.
		 * @details Creates the root node and calls buildOctree.
		 *
		 * @param size          The octree will span (-size, -size, -size) to 
		 *                      (size, size, size).
		 * @param polygons      A list of references to polygons that shall be 
		 *                      inserted into the octree.
		 * @param graphVertices A list of references to vertices that shall be 
		 *                      inserted into the octree.
		 */
		Octree(
			const float& size, 
			const std::vector<Polygon*>& polygons, 
			const std::vector<std::pair<Vertex, Graph*>>& graphVertices
		);


		/**
		 * @brief Free all memory allocated to this octree and polygons.
		 *
		 * @param root A flag that tells the function if it is invoked on the 
		 *             root of an octree. This is used to avoid a double free() 
		 *             on the polygons.
		 */
		void 
		deleteOctree(
			const bool& root
		);


		/**
		 * @brief For two lists of Polygons and vertices, sort them into an 
		 *        empty octree.
		 * @details Create new child octrees as long as the number of poylgons
		 *          for the current octree is greater than the defined
		 *          threshold. 
		 * 
		 * @param polygons      A list of pointers to Polygons that shall be 
		 *                      inserted into the octree.
		 * @param graphVertices A list of vertex-graph pairs that shall be 
		 *                      inserted into the octree.
		 * @param indent        The indentation used for output.
		 */
		void 
		buildOctree(
			const std::vector<Polygon*>& polygons, 
			const std::vector<std::pair<Vertex, Graph*>>& graphVertices, 
			const std::string& indent
		);
		
		
		/**
		 * @brief Check if a poylgon lies inside an octree.
		 * @details Uses a variant of the AABB intersection detection.
		 *
		 * @param polygon The poylgon that is to be checked.
		 * @param indent  The indentation used for output.
		 *
		 * @return Return true if polygon lies inside octree, false otherwise.
		 */
		bool 
		polygonInOctree(
			const Polygon* polygon, 
			const std::string& indent
		);


		/**
		 * @brief Check if a point lies inside an octree.
		 *
		 * @param p      The point that is to be checked.
		 * @param indent The indentation used for output.
		 *
		 * @return Return true if node lies inside octree, false otherwise.
		 */
		bool 
		pointInOctree(
			const Geometry::Vec3& p, 
			const std::string& indent
		);


		/**
		 * @brief Given a point p in 3D space, return all vertices associated to 
		 *        the smallest existing octree that has p in it.
		 *
		 * @param p      The point for which all vertices in its proximity shall 
		 *               be returned.
		 * @param indent The indentation used for output.
		 *
		 * @return Return a list of vertices that are near p.
		 */
		std::vector<std::pair<Vertex, Graph*>> 
		getNeighbours(
			const Geometry::Vec3& p, 
			const std::string& indent
		);

	}; 


	/**
	 * @brief Given the output of buildGraphList, create an octree containing 
	 *        all polygons and vertices.
	 *
	 * @param graphList The output of the builGraphList function.
	 *
	 * @return Return a pointer to the octree that was constructed.
	 */
	Octree* 
	createOctree(
		const std::vector<Graph*>& graphList
	);


	/**
	 * @brief Output a graphical representation of an octree to the console.
	 *
	 * @param octree The octree that shall be printed.
	 * @param indent The indentation that is prepended to every output to signal 
	 *               depth.
	 */
	void 
	printOctree(
		const Octree* octree, 
		const std::string& indent
	);
 
}

#endif
