/** 
 * @file
 *
 * @brief This file implements all functionality declared in Polygon.h.
 *
 * @author Jan-Luca Hansel
 */

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include "Vec.h"

namespace Pathfinder {

	Polygon::Polygon(
		const Geometry::Vec3& p1, 
		const Geometry::Vec3& p2, 
		const Geometry::Vec3& p3, 
		MinSG::GeometryNode* geoNodePtr
	) : 
		p1(p1), 
		p2(p2), 
		p3(p3), 
		center(Polygon::polygonCenter(p1, p2, p3)), 
		normal(Polygon::polygonNormal(p1, p2, p3)), 
		geoNodePtr(geoNodePtr) 
	{
		min = p1;
		max = p1;
		if(p2 < min) {
			min = p2;
		} else if(max < p2) {
			max = p2;
		}
		if(p3 < min) {
			min = p3;
		} else if(max < p3) {
			max = p3;
		}
	}
		
		
	Polygon::Polygon(
		const std::tuple<Geometry::Vec3, Geometry::Vec3, Geometry::Vec3>& points, 
		MinSG::GeometryNode* geoNodePtr
	) : 
		p1(std::get<0>(points)), 
		p2(std::get<1>(points)), 
		p3(std::get<2>(points)), 
		center(Polygon::polygonCenter(
			std::get<0>(points), std::get<1>(points), std::get<2>(points))
		), 
		normal(Polygon::polygonNormal(
			std::get<0>(points), std::get<1>(points), std::get<2>(points))
		), 
		geoNodePtr(geoNodePtr) 
	{
		min = p1;
		max = p1;
		if(p2 < min) {
			min = p2;
		} else if(max < p2) {
			max = p2;
		}
		if(p3 < min) {
			min = p3;
		} else if(max < p3) {
			max = p3;
		}
	}


	Geometry::Vec3 Polygon::getVertex(const short& idx) {
		switch(idx) {
			case 0 : return p1;
			case 1 : return p2;
			case 2 : return p3;
			default: spdlog::get("warn_log")->info("getVertex returns \{}.");
					 return {};
		}
	}


	Geometry::Vec3 Polygon::polygonCenter(
		const Geometry::Vec3& p1, 
		const Geometry::Vec3& p2, 
		const Geometry::Vec3& p3
	) {
		return Geometry::Vec3((p1[x] + p2[x] + p3[x])/3, (p1[y] + p2[y] + p3[y])/3, (p1[z] + p2[z] + p3[z])/3);
	}
	

	Geometry::Vec3 Polygon::polygonNormal(
		const Geometry::Vec3& p1, 
		const Geometry::Vec3& p2, 
		const Geometry::Vec3& p3
	) {
		Geometry::Vec3 i = {p2[x] - p1[x], p2[y] - p1[y], p2[z] - p1[z]};
		Geometry::Vec3 j = {p3[x] - p1[x], p3[y] - p1[y], p3[z] - p1[z]};
		return crossProduct(i, j);
	}
}
