/** 
 * @file
 *
 * @brief This file declares the struct Polygon that presents a convenient 
 *        way to access polygons.
 *
 * @author Jan-Luca Hansel
 */

#ifndef POLYGON_H
#define POLYGON_H

#include <Geometry/Vec3.h>
#include <MinSG/Core/Nodes/GeometryNode.h>

namespace Pathfinder {

	/**
	 * @brief A struct that represents a single polygon.
	 */
	struct Polygon {

		/** 
		 * The first vertex of the polygon.
		 */
		Geometry::Vec3 p1; 


		/** 
		 * The second vertex of the polygon.
		 */
		Geometry::Vec3 p2; 


		/** 
		 * The third vertex of the polygon.
		 */
		Geometry::Vec3 p3; 


		/** 
		 * The center point of the polygon.
		 */
		Geometry::Vec3 center; 


		/** 
		 * The normal vector of the polygon.
		 */
		Geometry::Vec3 normal; 


		/** 
		 * The lower corner of this polygon's AABB.
		 */
		Geometry::Vec3 min; 


		/** 
		 * The upper corner of this polygon's AABB.
		 */
		Geometry::Vec3 max; 


		/** 
		 * A pointer to the GeometryNode from which this polygon originates.
		 */
		MinSG::GeometryNode* geoNodePtr; 


		/**
		 * @brief A constructor that creates a poylgon out of three points.
		 *
		 * @param p1         The first vertex.
		 * @param p2         The second vertex.
		 * @param p3         The third vertex.
		 * @param geoNodePtr A pointer to the GeometryNode containing the 
		 *                   source mesh.
		 */
		Polygon(
			const Geometry::Vec3& p1, 
			const Geometry::Vec3& p2, 
			const Geometry::Vec3& p3, 
			MinSG::GeometryNode* geoNodePtr
		);
			
			
		/**
		 * @brief A constructor that creates a polyon out of a tuple of points.
		 *
		 * @param points     A tuple containing the three vertices.
		 * @param geoNodePtr A pointer to the GeometryNode containing the 
		 *                   source mesh.
		 */
		Polygon(
			const std::tuple<Geometry::Vec3, Geometry::Vec3, Geometry::Vec3>& points, 
			MinSG::GeometryNode* geoNodePtr
		);


		/**
		 * @brief Return the vertex associated with the given index.
		 * @details 0 = p1, 1 = p2, 2 = p3.
		 *
		 * @param idx The index of the vertex.
		 * 
		 * @return The vertex associated to idx.
		 */
		Geometry::Vec3 
		getVertex(
			const short& idx
		);

		
		/**
		 * @brief Compute the center point of a polygon.
		 *
		 * @param p1 The first vertex of the polygon.
		 * @param p2 The second vertex of the polygon.
		 * @param p3 The third vertex of the polygon.
		 *
		 * @return The center point of the polygon defined by p1, p2 and p3.
		 */
		static
		Geometry::Vec3 
		polygonCenter(
			const Geometry::Vec3& p1, 
			const Geometry::Vec3& p2, 
			const Geometry::Vec3& p3
		);


		/**
		 * @brief Compute the normal vector of a polygon.
		 * 
		 * @param p1 The first vertex of the polygon.
		 * @param p2 The second vertex of the polygon.
		 * @param p3 The third vertex of the polygon.
		 *
		 * @return The normal vector of the polygon defined by p1, p2 and p3.
		 */
		static
		Geometry::Vec3 
		polygonNormal(
			const Geometry::Vec3& p1, 
			const Geometry::Vec3& p2, 
			const Geometry::Vec3& p3
		);
		
	};
}

#endif
