/**
 * @file
 *
 * @brief This file implements all functionality declared in Search.h.
 *
 * @author Jan-Luca Hansel
 */

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include <boost/graph/graph_utility.hpp>

#include <E_Geometry/E_Vec3.h>

#include "Analysis.h"
#include "Search.h"
#include "Astar.h"
#include "Vec.h"
#include "AstarAlternativeHeuristic.h"

namespace Pathfinder {

	const float allowedStartEndDistance = 5;


	boost::optional<std::pair<Vertex, Vertex>> findStartGoal(
		const Geometry::Vec3& start, 
		const Geometry::Vec3& goal, 
		const Graph* graph
	) {
		// Start the timer for this function
		spdlog::stopwatch sw;

		// Take vertex 0 as th initial values
		auto vertexIterators = boost::vertices(*graph);
		Vertex s = 0;
		Vertex g = 0;
		float minDistanceS = distance3(start, (*graph)[0].pos);
		float minDistanceG = distance3(goal, (*graph)[0].pos);

		// Look at every other vertex and find the vertices that are closest
		// to the start and goal
		for(auto vit = vertexIterators.first; vit != vertexIterators.second; ++vit) {
			float currentDistanceS = distance3(start, (*graph)[*vit].pos);
			float currentDistanceG = distance3(goal, (*graph)[*vit].pos);
			if(currentDistanceS < minDistanceS) {
				minDistanceS = currentDistanceS;
				s = *vit;
			}
			if(currentDistanceG < minDistanceG) {
				minDistanceG = currentDistanceG;
				g = *vit;
			}
		}
	
		if(minDistanceS <= allowedStartEndDistance && minDistanceG <= allowedStartEndDistance) {
			std::pair<Vertex, Vertex> sg = {s, g};
			spdlog::get("time_log")->info("findStartGoal: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
			return sg;
		} else {
			spdlog::get("warn_log")->info("findStartGoal: Start/Goal too far away from any vertex.");
			spdlog::get("time_log")->info("findStartGoal: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
			return boost::none;
		}
	}

	boost::optional<Vertex> findNode(
		const Geometry::Vec3& node,

		const Graph* graph
	) {
		// Start the timer for this function
		spdlog::stopwatch sw;

		// Take vertex 0 as th initial values
		auto vertexIterators = boost::vertices(*graph);
		Vertex s = 0;
		float minDistanceS = distance3(node, (*graph)[0].pos);

		// Look at every other vertex and find the vertices that are closest
		// to the start and goal
		for(auto vit = vertexIterators.first; vit != vertexIterators.second; ++vit) {
			float currentDistanceS = distance3(node, (*graph)[*vit].pos);
		//	std::cout << " current and min distance "<<currentDistanceS <<" "<<minDistanceS<<'\n';
			if(currentDistanceS < minDistanceS) {
				minDistanceS = currentDistanceS;
				// print all vertices for debugging
				//std::cout<<"findNode: "<<(*graph)[*vit].pos<<std::endl;
				s = *vit;
			}
		}

		if(minDistanceS <= allowedStartEndDistance ) {
			spdlog::get("time_log")->info("findNode: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
			// print start or goal node for debugging
			//std::cout<<"found a Node: "<<(*graph)[s].pos<<std::endl;
			return s;
		} else {
			spdlog::get("warn_log")->info("findNode: Start/Goal too far away from any vertex.");
			spdlog::get("time_log")->info("findNode: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
			return boost::none;
		}
	}


	EScript::Array* findPath(
		const MinSG::ListNode& root, 
		const Geometry::Vec3& start, 
		const Geometry::Vec3& goal, 
		const short& agent
	) {
		// Start the timer for this function
		spdlog::stopwatch sw;

		// Do a static analysis of the scene
		std::vector<Graph*> graphs = sceneAnalysis(root);

		// Store the found path in this vector
		std::vector<Geometry::Vec3> vec3path;

		// TODO Find in which graph a path can be found
		// TODO At the moment: Assume there is only one graph
		Graph* graphStart=nullptr;
		Graph* graphGoal=nullptr;

		float startDistance = std::numeric_limits<float>::max();
		float goalDistance = std::numeric_limits<float>::max();
		Vertex startNode;
		Vertex goalNode;

		for(auto graph : graphs) {
			// Check if the graph is connected before attempting to find a path

			// Find the vertices that are the closest to Node (start or goal)

			boost::optional<Vertex> startNodeTemp= findNode (start, graph);

			// Find the vertices that are the closest to Node (start or goal)

			boost::optional<Vertex> goalNodeTemp= findNode (goal, graph);

			// // Find the vertices that are the closest to start and goal
			// boost::optional<std::pair<Vertex, Vertex>> startGoalPair =
			// 	findStartGoal(start, goal, graph);

			// If the start and end are too far away from any vertex in graph,
			// continue with the next graph
			if(startNodeTemp != boost::none) {
				float startTemp=distance3((*graph)[*startNodeTemp].pos,start);
				if(startTemp<startDistance) {
					startDistance=distance3((*graph)[*startNodeTemp].pos,start);
					graphStart= graph;
						startNode=*startNodeTemp;
				}

			}

			if(goalNodeTemp != boost::none) {
				float goalTemp=distance3((*graph)[*goalNodeTemp].pos,goal);
				if(goalTemp<goalDistance) {
					goalDistance=distance3((*graph)[*goalNodeTemp].pos,goal);
					graphGoal= graph;
						goalNode=*goalNodeTemp;
				}
			}
		}

		// if (!isGraphConnected(*graph)) {
		// 	std::cerr << "Graph " << graph << " is not connected!" << std::endl;
		// 	continue;
		// }

		if(graphGoal==graphStart) {
			// Try to find a path in the current graph
			std::vector<Vertex> path =
//				astar(graph, (*startGoalPair).first, (*startGoalPair).second, agent);

			// Try to find a path in the current graph
					AstarAlternativeHeuristic(graphStart, startNode, goalNode, agent);


			// Translate vertex descriptions into Vec3s and add them to vec3path
			vec3path = {start};
			for(auto v : path) {
				vec3path.push_back((*graphStart)[v].pos);
			}
			vec3path.push_back(goal);
		}
		spdlog::get("time_log")->info("findPath: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
		spdlog::drop_all();

		EScript::Array* arr = EScript::Array::create(vec3path);
		return arr;
	}

}
