/**
 * @file
 *
 * @brief This file declares the findPath function. This function can be 
 *        called in an EScript script to start the algorithm.
 *
 * @author Jan-Luca Hansel
 */

#ifndef SEARCH_H
#define SEARCH_H

#include <EScript/EScript/Objects/Collections/Array.h>

namespace Pathfinder {

	/**
	 * @brief Given 2 points in 3D space, approximate these points by nodes
	 *        in the graph while adhering to a maximum distance threshold.
	 *
	 * @param start The start point.
	 * @param goal  The goal point.
	 * @param graph The graph in which adequate start and goal vertices shall
	 *              be searched.
	 *
	 * @return Return a pair of vertices that approximate start and goal. 
	 *         If no such vertices exist, return boost::none.
	 */
	boost::optional<std::pair<Vertex, Vertex>>
	findStartGoal(
		const Geometry::Vec3& start, 
		const Geometry::Vec3& goal, 
		const Graph* graph
	);


	/**
	 * @brief Do the scene analysis and pathfinding.
	 *
	 * @param root  The root node of the scene graph.
	 * @param start The start point of the pathfinding.
	 * @param goal  The goal point of the pathfinding.
	 * @param agent The agent that shall be used for pathfinding.
	 *
	 * @return Return an EScript array which contains all nodes of the found 
	 *         path.
	 */
	EScript::Array*
	findPath(
		const MinSG::ListNode& root, 
		const Geometry::Vec3& start, 
		const Geometry::Vec3& goal,
		const short& agent
	);


}

#endif
