/** 
 * @file
 *
 * @brief This file declares all functionality related to providing random 
 *        access to tuples containing various data types.
 *
 * @author Jan-Luca Hansel
 */

#ifndef TUPLEACCESS_H
#define TUPLEACCESS_H

#include <vector>

#include "Line2.h"

namespace Pathfinder {

	/**
	 * @brief Access the element at index idx
	 *
	 * @param tValues The tValues tuple containing 6 float values.
	 * @param idx     The index on which access shall be performed.
	 *
	 * @return The t value stored at position idx.
	 */
	float 
	gettValue(
		const std::tuple<float, float, float, float, float, float>& tValues, 
		const short& idx
	); 


	/**
	 * @brief Access the element at index idx
	 *
	 * @param categorizedtValueIndices The tuple containing 5 vectors which 
	 *                                 represent the intervals used for case 
	 *                                 distinction.
	 * @param idx                      The index on which access shall be 
	 *                                 performed.
	 *
	 * @return The vector stored at position idx.
	 */
	std::vector<short> 
	gettValueIndexCategoryVector(
		const std::tuple<std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>, 
				         std::vector<short>>& categorizedtValueIndices, 
		const short& idx
	); 


	/**
	 * @brief Access the element at index idx
	 *
	 * @param lines The tuple containing three Line2 objects.
	 * @param idx   The index on which access shall be performed.
	 *
	 * @return The Line2 object stored at position idx.
	 */
	Line2 
	getLine2(
		const std::tuple<Line2, Line2, Line2>& lines, 
		const short& idx
	); 


	/**
	 * @brief Access the element at index idx and return a pointer to it.
	 *
	 * @param boolTuple The tuple containing three bool values.
	 * @param idx       The index on which access shall be performed.
	 *
	 * @return A pointer to the desired position in boolTuple.
	 */
	bool* 
	getBoolPtr(
		std::tuple<bool, bool, bool>& boolTuple,
		const short& idx
	); 


	/**
	 * @brief Access the element at index idx and return a pointer to it.
	 *
	 * @param shortTuple The tuple containing three short values.
	 * @param idx        The index on which access shall be performed.
	 *
	 * @return A pointer to the desired position in shortTuple.
	 */
	short* 
	getShortPtr(
		std::tuple<short, short, short>& shortTuple, 
		const short& idx
	); 


	/**
	 * @brief Access the element at index idx.
	 *
	 * @param optionalFloatTuple The tuple containing boost::optional<float> 
	 *                           values.
	 * @param idx                The index on which access shall be performed.
	 *
	 * @return The value of the tuple at the desired index.
	 */
	boost::optional<float> 
	getOptionalFloat(
		const std::tuple<boost::optional<float>, 
				         boost::optional<float>, 
				         boost::optional<float>>& optionalFloatTuple, 
		const short& idx
	); 


	/**
	 * @brief Access the element at index idx.
	 *
	 * @param boolTuple The tuple containing tree bool values.
	 * @param idx       The index on which access shall be performed.
	 *
	 * @return The value of the tuple at the desired index.
	 */
	bool 
	getBool(
		const std::tuple<bool, bool, bool>& boolTuple, 
		const short& idx
	); 


	/**
	 * @brief Access the vertex at index idx
	 *
	 * @param tri A tuple describing a polygon.
	 * @param idx The index on which access shall be performed.
	 *
	 * @return The vertex stored at position idx.
	 */
	Geometry::Vec3 
	getTriVertex(
		const std::tuple<Geometry::Vec3, Geometry::Vec3, Geometry::Vec3>& tri, 
		const short& idx
	); 


	/**
	 * @brief Access the vertex at index idx
	 *
	 * @param tri A tuple describing a flattened polygon.
	 * @param idx The index on which access shall be performed.
	 *
	 * @return The flattened vertex stored at position idx.
	 */
	Geometry::Vec2 
	getFlatTriVertex(
		const std::tuple<Geometry::Vec2, Geometry::Vec2, Geometry::Vec2>& tri, 
		const short& idx
	);
 
}

#endif
