/** 
 * @file
 *
 * @brief This file implements all functionality declared in Vec.h.
 *
 * @author Jan-Luca Hansel
 */

#include <cmath>


#include "Vec.h"
#include "Compare.h"
#include <algorithm>

namespace Pathfinder {

	Geometry::Vec3 crossProduct(
		const Geometry::Vec3& a, 
		const Geometry::Vec3& b
	) {
		return {a[y]*b[z]-a[z]*b[y], a[z]*b[x]-a[x]*b[z], a[x]*b[y]-a[y]*b[x]};
	}


	float dotProduct(const Geometry::Vec3 &a, const Geometry::Vec3 &b, int i) {
		return a[x]*b[x] + a[y]*b[y] + a[z]*b[z];
	}


	Geometry::Vec3 scalarMultiplication(
		const float& a, 
		const Geometry::Vec3& b
	) {
		return Geometry::Vec3(a*b[x], a*b[y], a*b[z]);
	}


	bool operator<(const Geometry::Vec3& l, const Geometry::Vec3& r) {
		if(l[x] < r[x]) {
			return true;
		} else if(compareEqualFloat(l[x], r[x])) {
			if(l[y] < r[y]) {
				return true;
			} else if(compareEqualFloat(l[y], r[y])) {
				if(l[z] < r[z]) {
					return true;
				}
			}
		}
		return false;
	}


	float distance2(Geometry::Vec2& p1, Geometry::Vec2& p2) {
		return sqrt(pow(p1[x] - p2[x], 2) + pow(p1[y] - p2[y], 2));
	}


	float distance3(const Geometry::Vec3 &v1, const Geometry::Vec3 &v2) {
        return
        sqrt(pow(v1[x]-v2[x], 2) + pow(v1[y]-v2[y], 2) +pow(v1[z]-v2[z], 2));
	}

    float distanceManhatten(const Geometry::Vec3 &v1, const Geometry::Vec3 &v2) {
        // Find Manhattan distance
        // using the formula
        // |x1 - x2| + |y1 - y2|
        return (abs(v1[x]-v2[x])
              + abs(v1[y]-v2[y])
              +abs(v1[z]-v2[z]));
    }

    float distanceChebyshev(const Geometry::Vec3 &v1, const Geometry::Vec3 &v2) {
        // Find Chebyshev distance
        // using the formula
        // max(|x1 - x2| , |y1 - y2|)
        return std::max(
                std::max(abs(v1[x]-v2[x]), abs(v1[y]-v2[y]))
                        ,abs(v1[z]-v2[z]));
    }

    float distanceDiagonal(const Geometry::Vec3 &v1, const Geometry::Vec3 &v2) {
        // Find Diagonal distance.
        // When D = 1 and D2 = sqrt(2), this is called the octile distance
        // using the formula
        // D * max(dx, dy) + (D2-D) * min(dx, dy)
        int dx = abs(v1[x] - v2[x]);
        int dy = abs(v1[y] - v2[y]);
        int dz = abs(v1[z] - v2[z]);
        int dmin = std::min(std::min(dx,dy),dz);
        int dmax = std::max(std::max(dx,dy),dz);
        int dmid = dx+dy+dz-dmin-dmax;
        int D1 = 1 ;
        int D2 = 1.3;//1.414;     //square root of 2
        int D3 = 1.4;//1.442;     //cube root of 3

        return ((D3-D2)* dmin + (D2-D1) * dmid + D1 * dmax);
    }

}
