/** 
 * @file
 *
 * @brief This file declares all functionality related to vectors.
 *
 * @author Jan-Luca Hansel
 */

#ifndef VEC_H
#define VEC_H

#include <Geometry/Vec2.h>

#include "Using.h"

namespace Pathfinder {

	/**
	 * @brief Compute the cross product of two vectors.
	 *
	 * @param a The first vector.
	 * @param b The second vector.
	 *
	 * @return The cross product of a and b.
	 */
	Geometry::Vec3 
	crossProduct(
		const Geometry::Vec3& a, 
		const Geometry::Vec3& b
	); 


	/**
	 * @brief Compute the dot product of two vectors.
	 *
	 * @param a The first vector.
	 * @param b The second vector.
	 *
	 * @return The dot product of a and b.
	 */
    float
	dotProduct(const Geometry::Vec3 &a, const Geometry::Vec3 &b, int i);


	/**
	 * @brief Scale a vector by a scalar.
	 *
	 * @param a The scaling factor.
	 * @param b The vector.
	 *
	 * @return Vector b scaled by a.
	 */
	Geometry::Vec3 
	scalarMultiplication(
		const float& a, 
		const Geometry::Vec3& b
	); 


	/**
	 * @brief Overload operator<.
	 *
	 * @param l The left operand.
	 * @param r The right operand.
	 *
	 * @return True if l is lexicographically smaller than r.
	 */
	bool 
	operator<(
		const Geometry::Vec3& l, 
		const Geometry::Vec3& r
	); 


	/**
	 * @brief Compute the euclidean distance between two points in 2D.
	 *
	 * @param p1 The first point.
	 * @param p2 The second point.
	 *
	 * @return The euclidean distance between p1 and p2.
	 */
	float 
	distance2(
		Geometry::Vec2& p1, 
		Geometry::Vec2& p2
	); 


	/**
	 * @brief Compute the euclidean distance between two nodes.
	 * @details Given three points in 3D, compute the length of the straight 
	 *          line between the two points.
	 *
	 * @param v1 The position of the first node.
	 * @param v2 The position of the second node.
	 *
	 * @return The euclidean distance between the two nodes.
	 */
	float
	distance3(
		const Geometry::Vec3 &v1, 
		const Geometry::Vec3 &v2
	);

    /**
	 * @brief Compute the Manhatten distance between two nodes.
	 * @details Given three points in 3D, compute the length of the straight
	 *          line between the two points.
	 *
	 * @param v1 The position of the first node.
	 * @param v2 The position of the second node.
	 *
	 * @return The Manhatten distance between the two nodes.
	 */
    float
    distanceManhatten(
            const Geometry::Vec3 &v1,
            const Geometry::Vec3 &v2
    );

    /**
	 * @brief Compute the Chebyshev distance between two nodes.
	 * @details Given three points in 3D, compute the length of the straight
	 *          line between the two points.
	 *
	 * @param v1 The position of the first node.
	 * @param v2 The position of the second node.
	 *
	 * @return The Chebyshev distance between the two nodes.
	 */
    float
    distanceChebyshev(
            const Geometry::Vec3 &v1,
            const Geometry::Vec3 &v2
    );

    /**
    * @brief Compute the Diagonal distance between two nodes.
    * @details Given three points in 3D, compute the length of the straight
    *          line between the two points.
    *
    * @param v1 The position of the first node.
    * @param v2 The position of the second node.
    *
    * @return The Diagonal distance between the two nodes.
    */
    float
    distanceDiagonal(
            const Geometry::Vec3 &v1,
            const Geometry::Vec3 &v2
    );

}

#endif
