//
// Created by vinee on 8/7/24.
//

/**
 * @file
 *
 * @brief This file implements all functionality declared in AstarAlternativeHeuristic.h.
 *
 * @author Vineet
 */

#include <algorithm>
#include <stack>

#include <spdlog/spdlog.h>
#include <spdlog/sinks/basic_file_sink.h>
#include <spdlog/stopwatch.h>
#include <spdlog/fmt/chrono.h>

#include <boost/optional/optional_io.hpp>

#include "AstarAlternativeHeuristic.h"

#include <queue>

#include "Vec.h"

namespace Pathfinder {

    typedef float (*distanceFunction)(const Geometry::_Vec3<float> &v1, const Geometry::_Vec3<float> &v2);
    distanceFunction distance=  //distanceDiagonal;       // Diagonal distance
                                    distance3;            //Euclidean distance
                                //distanceChebyshev;      //Chebyshev distance
                                //distanceManhatten;      //Manhattan distance


	boost::optional<Vertex> nextSuccessor(
		const Graph* graph, const Vertex& v, const Vertex& goal,
		const short& agent, const std::vector<Vertex>& considered
	) {
		// Get all candidate successors
		std::vector<Vertex> candidates;
		auto incidentEIts = boost::out_edges(v, *graph);
		for(auto eit = incidentEIts.first; eit != incidentEIts.second; ++eit) {
			// Edge must be traversable for the agent
			if(agent >= (*graph)[*eit].drivability) {
               // std::cout << "\tdrivability: "<<(*graph)[*eit].drivability <<"\tagent" <<agent;
				candidates.push_back(boost::target(*eit, *graph));

            }

		}

		// Do not return a vertex that is already on the stack
		for(auto v : considered) {
			candidates.erase(
				std::remove(candidates.begin(), candidates.end(), v),
				candidates.end()
			);
		}

		if(candidates.size() != 0) {
			// Successors exist. Get the one with the smallest Euclidean
			// distance to the goal
			float minDistance =
                    //distance3((*graph)[candidates[0]].pos, (*graph)[goal].pos);
                    distance((*graph)[candidates[0]].pos, (*graph)[goal].pos);
			Vertex chosen = candidates[0];
			for(size_t i = 1; i < candidates.size(); ++i) {
				float current =

                        //distance3((*graph)[candidates[i]].pos, (*graph)[goal].pos);
                        distance((*graph)[candidates[i]].pos, (*graph)[goal].pos);
				if(current < minDistance) {
					minDistance = current;
					chosen = candidates[i];
				}
			}
	//		std::cout << " v val "<<v<<'\n';
	//		std::cout << " considered val "<<&considered<<'\n';


	//		for (Vertex i: considered)
	//			std::cout << i << ' ';

	//		std::cout <<'\n';
	//		std::cout << " chosen "<<chosen<<'\n';
			return chosen;
		}

		return boost::none;
	}

    float est =0;
	float cost(const Graph* graph, const Vertex& ni, const Vertex& nj) {
		// Get the edge between ni and nj
		auto e = boost::edge(ni, nj, *graph);

        // @author: Vineet
        //when determining next vertex, A star uses edge weight as one of the decision factor
        // adding edge weight along with distance between two edges

		// Check if the edge exists
		if(e.second) {
            if((*graph)[e.first].edge_weight ==0) {
                return (*graph)[e.first].distance;
            }else {
                est = (*graph)[e.first].distance - 0.125* (*graph)[e.first].distance+ 0.125* (*graph)[e.first].distance*(*graph)[e.first].edge_weight;
//                std::cout << "\nedge distance: " << (*graph)[e.first].distance;
//                std::cout << "\t edge weight: " << (*graph)[e.first].edge_weight;
//                std::cout << "\t total estimate: " << est;
//                std::cout << "\t type: " << (*graph)[e.first].edge_type;
                return est;
            }
		}
		spdlog::get("warn_log")->info("cost returns std::numeric_limits<float>::max().");
		return std::numeric_limits<float>::max();
	}

	std::vector<Vertex> AstarAlternativeHeuristicOri(
		const Graph* graph, const Vertex& start,
		const Vertex& goal, const short& agent
	) {
		// Start timer for this function
		spdlog::stopwatch sw;

		std::cout << yellow << "Start A*" << reset << '\n';

		std::stack<Vertex> s;
		std::vector<Vertex> considered;
		bool found = false;

		// g and h values for every vertex
		std::vector<float>
			g(boost::num_vertices(*graph), std::numeric_limits<float>::max());
		std::vector<float>
			h(boost::num_vertices(*graph), std::numeric_limits<float>::max());

		// Set g and h for the start node
		g[start] = 0;
		h[start] = distance3((*graph)[start].pos, (*graph)[goal].pos);

		// Initial threshold
		float B = h[start];
		int PathCounter = 0;
		int NodesCounter= 0;

		// Try to find a path
		while(!found && B < std::numeric_limits<float>::max()) {
			s.push(start);
			considered = {start};
			float Bnew = std::numeric_limits<float>::max();
			while(!found && s.size() != 0) {
				// Check if all nodes lie on the stack
				if(s.size() == boost::num_vertices(*graph)) {
					return {};
				}
				Vertex ni = s.top();
				if(ni != goal) {
					boost::optional<Vertex> nj =
						nextSuccessor(graph, ni, goal, agent, considered);
					if(nj != boost::none) {
						considered.push_back(*nj);
						g[*nj] =
							std::min(g[*nj], g[ni] + cost(graph, ni, *nj));
						h[*nj] =
							distance3((*graph)[*nj].pos, (*graph)[goal].pos);
						if((g[*nj] + h[*nj]) > B) {
							Bnew = std::min(Bnew, (g[*nj] + h[*nj]));
						} else {
							s.push(*nj);
						}
					} else {
						s.pop();
					}
				} else {
					std::cout << "Found a path!\n";
					found = true;
					std::vector<Vertex> p;
					while(s.size() != 0) {
						p.push_back(s.top());
						s.pop();
					}
					std::reverse(p.begin(), p.end());
					spdlog::get("time_log")->info("astar: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
					spdlog::get("time_log")->info("PathCounter: {}", PathCounter);
					spdlog::get("time_log")->info("NodesCounter: {}", NodesCounter);
					std::cout << " PathCounter: " << PathCounter << " NodesCounter: " << NodesCounter;

					return p;
				}
				++PathCounter;
			}
			B = Bnew;
			++NodesCounter;
			spdlog::get("node_log")->info("[{}] {}", B, considered.size());
		}
		spdlog::get("time_log")->info("astar: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
		return {};
	}

void printPath(const PathTaken* goal) {
        const PathTaken* current = goal;
        while (current) {
            current->print();
            current = current->predecessor;
        }
    }

	//checks if there is an obstacle at the current vertex
    bool isObstacle(const Pathfinder::Graph& graph, const Pathfinder::Vertex& vertex) {
		std::cout <<"is_obstacle " << graph[vertex].is_obstacle<< std::endl;
		//to-do
	    return graph[vertex].is_obstacle;
	}

    bool hasLineOfSight(const Pathfinder::Graph* graph, const Pathfinder::Vertex& start, const Pathfinder::Vertex& end) {
		// Bresenham's line algorithm is used to determine if there's a direct line of sight
		auto pos1 = (*graph)[start].pos;
	    auto pos2 = (*graph)[end].pos;

	    int x0 = pos1.x(); // x() is a method of Geometry::Vec3 pos1
	    int y0 = pos1.y(); //  y() is a method of Geometry::Vec3 pos1
	    int x1 = pos2.x(); //  x() is a method of Geometry::Vec3 pos2
	    int y1 = pos2.y(); //  y() is a method of Geometry::Vec3 pos2

	    int dx = abs(x1 - x0);   //absolute difference between the x-coordinates of the end and start points
	    int dy = abs(y1 - y0); //absolute difference between the y-coordinates of the end and start points
	    int sx = x0 < x1 ? 1 : -1; //step direction in x and y based on whether the end point is greater than or less than the start point
	    int sy = y0 < y1 ? 1 : -1;
	    int err = dx - dy; //manage the approximation of the line

	    while (true) {
	        Pathfinder::Vertex current = start; // Calculate current vertex based on (x0, y0) if needed
	        if (isObstacle(*graph, current)) {
	        	std::cout <<"returns false " << std::endl;
	            return false;  //return false if an obstacle is found
	        }
	        if (x0 == x1 && y0 == y1) {
	            break;
	        }
	        int e2 = err * 2;
	        //x0 and y0 are adjusted by comparing with e2 to determine whether to step in the x or y direction
	        if (e2 > -dy) {
	            err -= dy;
	            x0 += sx;
	        //	std::cout <<" e2 > -dy " << err <<" "<< x0<<std::endl;

	        }
	        if (e2 < dx) {
	            err += dx;
	            y0 += sy;
	       // 	std::cout <<" e2 < dx " << err <<" "<< y0<<std::endl;

	        }
	    	//std::cout <<" e2 value " << e2<<std::endl;

	    }
	    return true;
	}




	std::vector<Pathfinder::Vertex> smoothenPath(const Pathfinder::Graph* graph, const std::vector<Pathfinder::Vertex>& path) {
		if (path.empty()) return {};

		// Convert the vector path to a list to use iterators
		std::list<Pathfinder::Vertex> pathList(path.begin(), path.end());

		// Print the contents of the pathList
		std::cout << "Original Path (List): ";
		for (const Pathfinder::Vertex& vertex : pathList) {
			std::cout << vertex << " "; // Print each vertex using its output stream operator (<<)
		}
		std::cout << std::endl;

		// Create two iterators pointing at the front of the path list
		auto e1 = pathList.begin();
		auto e2 = pathList.begin();
		++e2; // Increment e2 so it points to the next element

		// Step through the edges, checking for line of sight
		while (e2 != pathList.end()) {
			if (hasLineOfSight(graph, *e1, *e2)) {
				// If there is line of sight, remove the intermediate point
				e2 = pathList.erase(std::next(e1));
			} else {
				// Otherwise, advance both iterators
				++e1;
				++e2;
			}
		}

		// Convert the list back to a vector
		std::vector<Pathfinder::Vertex> smoothedPath(pathList.begin(), pathList.end());
		return smoothedPath;
	}

	std::vector<Vertex> AstarAlternativeHeuristicTest(const Graph* graph, const Vertex& start, const Vertex& goal, const short& agent) {
    spdlog::stopwatch sw;

    std::cout << "Start A*" << '\n';

    std::stack<Vertex> s;
    std::vector<Vertex> considered;
    bool found = false;

    std::vector<float> g(boost::num_vertices(*graph), std::numeric_limits<float>::max());
    std::vector<float> h(boost::num_vertices(*graph), std::numeric_limits<float>::max());

    g[start] = 0;
    h[start] = distance3((*graph)[start].pos, (*graph)[goal].pos);

    float B = h[start];
    int PathCounter = 0;
    int NodesCounter = 0;

    while (!found && B < std::numeric_limits<float>::max()) {
        s.push(start);
        considered = {start};
        float Bnew = std::numeric_limits<float>::max();
        while (!found && s.size() != 0) {
            if (s.size() == boost::num_vertices(*graph)) {
                return {};
            }
            Vertex ni = s.top();
            if (ni != goal) {
                boost::optional<Vertex> nj = nextSuccessor(graph, ni, goal, agent, considered);
                if (nj != boost::none) {
                    considered.push_back(*nj);
                    g[*nj] = std::min(g[*nj], g[ni] + cost(graph, ni, *nj));
                    h[*nj] = distance3((*graph)[*nj].pos, (*graph)[goal].pos);
                    if ((g[*nj] + h[*nj]) > B) {
                        Bnew = std::min(Bnew, (g[*nj] + h[*nj]));
                    } else {
                        s.push(*nj);
                    }
                } else {
                    s.pop();
                }
            } else {
                std::cout << "Found a path!\n";
                found = true;
                std::vector<Vertex> p;
                while (s.size() != 0) {
                    p.push_back(s.top());
                    s.pop();
                }
                std::reverse(p.begin(), p.end());

                // Smooth the path
                std::vector<Vertex> smoothedPath = smoothenPath(graph, p);

                spdlog::get("time_log")->info("astar: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
                spdlog::get("time_log")->info("PathCounter: {}", PathCounter);
                spdlog::get("time_log")->info("NodesCounter: {}", NodesCounter);
                std::cout << " PathCounter: " << PathCounter << " NodesCounter: " << NodesCounter;

                return  p;//smoothedPath;
            }
            ++PathCounter;
        }
        B = Bnew;
        ++NodesCounter;
        spdlog::get("node_log")->info("[{}] {}", B, considered.size());
    }
    spdlog::get("time_log")->info("astar: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
    return {};
}

std::vector<Pathfinder::Vertex> AstarAlternativeHeuristicCLsA(
    const Pathfinder::Graph* graph, const Pathfinder::Vertex& start,
    const Pathfinder::Vertex& goal, const short& agent
) {
    spdlog::stopwatch sw;

    std::cout << "Start A*" << std::endl;

    std::unordered_map<Pathfinder::Vertex, Pathfinder::Vertex> parent;
    std::vector<Pathfinder::Vertex> considered;
    bool found = false;

    std::vector<float> g(boost::num_vertices(*graph), std::numeric_limits<float>::max());
    std::vector<float> h(boost::num_vertices(*graph), std::numeric_limits<float>::max());

    g[start] = 0;
    h[start] = distance((*graph)[start].pos, (*graph)[goal].pos);

    float B = h[start];
    int PathCounter = 0;
    int NodesCounter = 0;

    using PQElement = std::pair<float, Pathfinder::Vertex>;
    std::priority_queue<PQElement, std::vector<PQElement>, std::greater<PQElement>> openSet;
    openSet.emplace(g[start] + h[start], start);
    considered.push_back(start);

    while (!found && B < std::numeric_limits<float>::max()) {
        float Bnew = std::numeric_limits<float>::max();
        while (!found && !openSet.empty()) {
            Pathfinder::Vertex ni = openSet.top().second;
            openSet.pop();

            if (ni != goal) {
                boost::optional<Pathfinder::Vertex> nj = nextSuccessor(graph, ni, goal, agent, considered);
                if (nj != boost::none) {
                    considered.push_back(*nj);
                    float tentative_g = g[ni] + cost(graph, ni, *nj);
                    if (tentative_g < g[*nj]) {
                        parent[*nj] = ni;
                        g[*nj] = tentative_g;
                        h[*nj] = distance((*graph)[*nj].pos, (*graph)[goal].pos);
                        float f = g[*nj] + h[*nj];
                        if (f > B) {
                            Bnew = std::min(Bnew, f);
                        } else {
                            openSet.emplace(f, *nj);
                        }
                    }
                }
            } else {
            	std::cout << "\nFound a path!\n";
            	found = true;
            	std::vector<Pathfinder::Vertex> p;
            	for (Pathfinder::Vertex v = goal; v != start; v = parent[v]) {
            		p.push_back(v);
            	}
            	p.push_back(start);
            	std::reverse(p.begin(), p.end());

            	std::cout << "Original path:\n";
            	for (auto vertex : p) {
            		std::cout << vertex << " ";
            	}
            	std::cout << std::endl;

            	auto smoothPath = smoothenPath(graph, p);

            	std::cout << "Smoothed path:\n";
            	for (auto vertex : smoothPath) {
            		std::cout << vertex << " ";
            	}
            	std::cout << std::endl;

            	spdlog::get("time_log")->info("AstarAlternativeHeuristic: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
            	spdlog::get("time_log")->info("PathCounter: {}", PathCounter);
            	spdlog::get("time_log")->info("NodesCounter: {}", NodesCounter);
            	std::cout << " PathCounter: " << PathCounter << " NodesCounter: " << NodesCounter;

            	return smoothPath;
            }
            ++PathCounter;
        }
        B = Bnew;
        ++NodesCounter;
        spdlog::get("node_log")->info("[{}] {}", B, considered.size());
    }

    spdlog::get("time_log")->info("AstarAlternativeHeuristic: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));

    return {};
}

 std::vector<Vertex> AstarAlternativeHeuristic(
        const Graph* graph,  const Vertex& start,
        const Vertex& goal, const short& agent
    ) {
        // Start timer for this function
        spdlog::stopwatch sw;

        std::cout << yellow << "Start A*" << reset << '\n';

        std::stack<PathTaken*> s;
        std::vector<Vertex> considered;
        bool found = false;

        // g and h values for every vertex
        std::vector<float>
            g(boost::num_vertices(*graph), std::numeric_limits<float>::max());
        std::vector<float>
            h(boost::num_vertices(*graph), std::numeric_limits<float>::max());


        // Set g and h for the start node
        g[start] = 0;
        h[start] = distance((*graph)[start].pos, (*graph)[goal].pos);

        // Initial threshold
        float B = h[start];
        int PathCounter = 0;
        int NodesCounter= 0;

        // Try to find a path
        while(!found && B < std::numeric_limits<float>::max()) {

            s.push(new PathTaken(start));
           // std::cout << "Pushed start vertex: " << start << '\n';
            considered = {start};
            float Bnew = std::numeric_limits<float>::max();
            while(!found && s.size() != 0) {

                // Check if all nodes lie on the stack
                if(s.size() == boost::num_vertices(*graph)) {
                    return {};
                }
                PathTaken* ni = s.top();
                if(ni->vertex != goal) {
                    boost::optional<Vertex> nj =
                        nextSuccessor(graph, ni->vertex, goal, agent, ni->expanded);
                    if(nj != boost::none) {
                       // considered.push_back(*nj);

                    	float valg;
                    	valg=g[ni->vertex] + cost(graph, ni->vertex, *nj);
                    	ni->expanded.push_back(*nj);

                    //	std::cout << " nj val "<< *nj <<'\n';
                    //	std::cout << "  g[*nj] val "<<  g[*nj] <<'\n';
                    //	std::cout << "  valg "<<  valg <<'\n';

                    	if(g[*nj]>=valg) {
                    		g[*nj] =std::min(g[*nj],valg );
                        h[*nj] =
                                distance((*graph)[*nj].pos, (*graph)[goal].pos);
                        if((g[*nj] + h[*nj]) > B) {
                            Bnew = std::min(Bnew, (g[*nj] + h[*nj]));
                        //    std::cout << " Bnew "<<Bnew <<'\n';
                        } else {
                            s.push(new PathTaken(*nj, ni));
                         //   std::cout << "Pushed vertex: " << *nj << '\n';
                        }
                    	}
                    } else {
                    	PathTaken* popped= s.top();
                      //  std::cout << "Popped vertex: " << ni->vertex << '\n';
                        s.pop();
                    	delete popped;
                    }
                } else {
                    std::cout << "\nFound a path!\n";
                    found = true;
                    std::vector<Vertex> p;
                     PathTaken* current = ni;
                	std::cout << "Complete path:\n";
                	//printPath(ni);
                	while(current) {
                     //   std::cout << "Popped vertex (path): " << current->vertex << '\n';
                        p.push_back(current->vertex);

                    	PathTaken* popped= current;
                    	current = current->predecessor;
						delete popped;
                    }

                    std::reverse(p.begin(), p.end());




                    spdlog::get("time_log")->info("AstarAlternativeHeuristic: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
                    spdlog::get("time_log")->info("PathCounter: {}", PathCounter);
                    spdlog::get("time_log")->info("NodesCounter: {}", NodesCounter);
                    std::cout << " PathCounter: " << PathCounter << " NodesCounter: " << NodesCounter;
                    return p;
                }
                ++PathCounter;
            }
            B = Bnew;
            ++NodesCounter;
            spdlog::get("node_log")->info("[{}] {}", B, considered.size());

        }

        spdlog::get("time_log")->info("AstarAlternativeHeuristic: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));


        return {};
    }


    std::vector<Vertex> AstarAlternativeHeuristicTest22(
        const Graph* graph, const Vertex& start,
        const Vertex& goal, const short& agent
    ) {
        // Start timer for this function
        spdlog::stopwatch sw;

        std::cout << yellow << "Start A*" << reset << '\n';

        std::stack<PathTaken> s;
        std::vector<Vertex> considered;
        bool found = false;

        // g and h values for every vertex
        std::vector<float>
            g(boost::num_vertices(*graph), std::numeric_limits<float>::max());
        std::vector<float>
            h(boost::num_vertices(*graph), std::numeric_limits<float>::max());

        // Set g and h for the start node
        g[start] = 0;
        h[start] = distance((*graph)[start].pos, (*graph)[goal].pos);

        // Initial threshold
        float B = h[start];
        int PathCounter = 0;
        int NodesCounter= 0;

        // Try to find a path
        while(!found && B < std::numeric_limits<float>::max()) {

            s.push( PathTaken(start));
           // std::cout << "Pushed start vertex: " << start << '\n';
            considered = {start};
            float Bnew = std::numeric_limits<float>::max();
            while(!found && s.size() != 0) {

                // Check if all nodes lie on the stack
                if(s.size() == boost::num_vertices(*graph)) {
                    return {};
                }
                PathTaken ni = s.top();
                if(ni.vertex != goal) {
                    boost::optional<Vertex> nj =
                        nextSuccessor(graph, ni.vertex, goal, agent, considered);
                    if(nj != boost::none) {
                        considered.push_back(*nj);
                        g[*nj] =
                            std::min(g[*nj], g[ni.vertex] + cost(graph, ni.vertex, *nj));
                        h[*nj] =
                                distance((*graph)[*nj].pos, (*graph)[goal].pos);
                        if((g[*nj] + h[*nj]) > B) {
                            Bnew = std::min(Bnew, (g[*nj] + h[*nj]));
                        } else {
                            s.push( PathTaken(*nj, &ni));
                         //   std::cout << "Pushed vertex: " << *nj << '\n';
                        	std::cout << "Pushed pred vertex: " ; ni.print();
                        	std::cout << "Ni nj values: "<<nj<<&ni<< '\n'; ;
                        }
                    } else {
                      //  std::cout << "Popped vertex: " << ni.vertex << '\n';
                    //	std::cout << "Popped pred vertex: " << ni.predecessor << '\n';

                        s.pop();
                    }
                } else {
                    std::cout << "\nFound a path!\n";
                    found = true;
                    std::vector<Vertex> p;
                     PathTaken current = ni;

                	while(current.predecessor) {
                        std::cout << "Popped vertex (path): " << current.vertex << '\n';
                        p.push_back(current.vertex);
                        current = *(current.predecessor);
                    }
                    std::reverse(p.begin(), p.end());

                    std::cout << "Complete path:\n";
                    printPath(&ni);

                    spdlog::get("time_log")->info("AstarAlternativeHeuristic: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));
                    spdlog::get("time_log")->info("PathCounter: {}", PathCounter);
                    spdlog::get("time_log")->info("NodesCounter: {}", NodesCounter);
                    std::cout << " PathCounter: " << PathCounter << " NodesCounter: " << NodesCounter;
                    return p;
                }
                ++PathCounter;
            }
            B = Bnew;
            ++NodesCounter;
            spdlog::get("node_log")->info("[{}] {}", B, considered.size());

        }

        spdlog::get("time_log")->info("AstarAlternativeHeuristic: {}", std::chrono::duration_cast<std::chrono::microseconds>(sw.elapsed()));


        return {};
    }



				{
				std::cout << "object is not parallel to axis" << std::endl;

				// Use convhull_3d to create convex hull for non-parallel object
				std::vector<Geometry::Vec3> points = {smallerTri->p1, smallerTri->p2, smallerTri->p3};

				if (points.size() < 4) {
					std::cerr << "Error: At least 4 points are required to compute a 3D convex hull." << std::endl;
					return {}; // Return an empty vector
				}
				// Convert points to ch_vertex format required by convhull_3d
				std::vector<ch_vertex> vertices(points.size());
				for (size_t i = 0; i < points.size(); ++i) {
					vertices[i].x = static_cast<CH_FLOAT>(points[i].x());
					vertices[i].y = static_cast<CH_FLOAT>(points[i].y());
					vertices[i].z = static_cast<CH_FLOAT>(points[i].z());
				}

				// Print vertices before passing to convhull_3d_build
				std::cout << "Vertices to be used for convex hull computation:" << std::endl;
				for (const auto& vertex : vertices) {
					std::cout << "Vertex: (" << vertex.x << ", " << vertex.y << ", " << vertex.z << ")" << std::endl;
				}

				//std::cout << "float coversion"<<std::endl;
				// Prepare data structures for convex hull computation
				int* faceIndices = nullptr;
				int nFaces = 0;

				std::cout << "before convhull_3d"<<std::endl;
				// Compute the convex hull
				convhull_3d_build(vertices.data(), vertices.size(), &faceIndices, &nFaces);
				std::cout << "after convhull_3d"<<std::endl;
				// Check if faceIndices and nFaces are valid
				if (!faceIndices || nFaces <= 0) {
					std::cerr << "Error: Invalid convex hull faces." << std::endl;
					return {}; // Return an empty vector
				}

				// Extract convex hull vertices and faces
				std::vector<Geometry::Vec3> convexHullVertices(vertices.size());
				for (size_t i = 0; i < vertices.size(); ++i) {
					convexHullVertices[i] = Geometry::Vec3(
						static_cast<float>(vertices[i].x),
						static_cast<float>(vertices[i].y),
						static_cast<float>(vertices[i].z));
				}

				// Here you might want to do something with convexHullVertices and faceIndices
				// For example, print the convex hull vertices and faces
				std::cout << "Convex hull vertices:" << std::endl;
				for (const auto& vertex : convexHullVertices) {
					std::cout << vertex.x() << ", " << vertex.y() << ", " << vertex.z() << std::endl;
				}

				std::cout << "Convex hull faces:" << std::endl;
				for (int i = 0; i < nFaces; ++i) {
					std::cout << faceIndices[3 * i] << ", " << faceIndices[3 * i + 1] << ", " << faceIndices[3 * i + 2] << std::endl;
				}

				// Free allocated memory
				free(faceIndices);
			}

}
